/*
 * Gigaset 307x configuration program
 *
 * This program parses the XML parameter specification file homerpar.xml
 * which Siemens delivers with Talk&Surf.
 *
 * Copyright (c) 2004 by Tilman Schmidt (tilman.schmidt@epost.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <ctype.h>
#include <errno.h>
#include <expat.h>
#include <asm/types.h>
#include <gigaset/interf.h>
#include <gigaset/conf.h>

#define BUFFSIZE 8192
#define MAXKEYS 4

#define stringify(x) str(x)
#define str(x) #x

/* XML attribute names */
const char *attrtab[] = {
    "id",
#define ATTR_ID 0
    "name",
#define ATTR_NAME 1
    "range",
#define ATTR_RANGE 2
    "length",
#define ATTR_LENGTH 3
    "readOnly",
#define ATTR_READONLY 4
    "typeInfo",
#define ATTR_TYPEINFO 5
    NULL
};

/* typeInfo strings */
const char *typetab[] = {
    "Byte",
#define TYPE_BYTE 0
    "Bytearray",
#define TYPE_BYTEARRAY 1
    "Long",
#define TYPE_LONG 2
    "String",
#define TYPE_STRING 3
    NULL
};

/* location of Siemens parameter specification file homerpar.xml */
const char *homerpar_path = "/usr/local/lib/homerpar.xml";	/* default path */
const char *homerpar_envvar = "HOMERPAR";	/* environment variable for alternate path */

/* Gigaset interface device */
const char *if_path = "/dev/ttyG";	/* default path */
const char *if_envvar = "GIGASET";	/* environment variable for alternate path */

/* extensions excluded from processing (0: include, 1: exclude)
 * 0: base
 * 1: any (flag "-e option seen")
 * 11..18: cordless phones
 * 21..22: analog ports
 * 40..48: data devices
 * 91..93: answering machines
 * remaining fields unused
 */
char exclist[100];

/* output destination */
FILE *out = NULL;

/* XML parser state */
enum {
    INIT,
    HX_PARAMS,
    PARAMTREE,
    SKIP,
    CONFIG,
    PARAM
} Hpar_state;
int Depth;

/* Gigaset device state */
static struct gstatus *gs = NULL;
/* Log structure */
struct glog gl;

/* backward compatibility for pre 1.95.7 Expat */
#if !defined(XMLCALL)
#define XMLCALL
#endif

/* string table lookup functions
 * return index in table if found, -1 otherwise
 */
static inline int
strlookup(const char **tab, const char *s)
{
    int i;
    for (i = 0; tab[i] != NULL; i++)
	if (strcmp(s, tab[i]) == 0)
	    return i;
    return -1;
}

static inline int
strilookup(const char **tab, const char *s)
{
    int i;
    for (i = 0; tab[i] != NULL; i++)
	if (strcasecmp(s, tab[i]) == 0)
	    return i;
    return -1;
}

/* Gigaset interface functions */

/* closeif
 * close interface device and return connection to Gigaset base to normal state
 */
static void
closeif(void)
{
    int r;

    /* flush stuck replies */
    if ((r = gigaset_clear(gs)) < 0)
	    gigaset_logerr(&gl, GLOG_ERR, -r, gigaset_errno(gs), "flushing Gigaset responses");

    /* unlock and close */
    if ((r = gigaset_lock(gs, 0)) < 0)
	    gigaset_logerr(&gl, GLOG_ERR, -r, gigaset_errno(gs), "unlocking Gigaset device");

    gigaset_close(gs);
    gs = NULL;
}

/* openif
 * open Gigaset interface device and prepare connection to base for
 * transmission of configuration commands
 * return value:
 *	<0: error
 */
static int
openif(const char *path)
{
    int r, retry;
    int gig_errno;

    /* open gigaset device */
    if ((gs = gigaset_open(&r, &gig_errno, path, &gl, NULL)) == NULL) {
	gigaset_logerr(&gl, GLOG_ERR, r, gig_errno, NULL);
	return -r;
    }

    /* acquire lock */
    retry = 0;
    while ((r = gigaset_lock(gs, 1)) == -ERR_IO && gigaset_errno(gs) == EBUSY && retry < 5) {
	if (gl.verb > 1)
	    fprintf (stderr, "lock busy, retry %d\n", ++retry);
	sleep(1);
    }
    if (r < 0) {
	gigaset_logerr(&gl, GLOG_ERR, -r, gigaset_errno(gs), "locking Gigaset device");
	gigaset_close(gs);
	gs = NULL;
	return r;
    }

    /* switch to Call ID (native) command mode */
    if ((r = gigaset_cidmode(gs)) < 0) {
	gigaset_logerr(&gl, GLOG_ERR, -r, gigaset_errno(gs), "setting CID mode");
	gigaset_lock(gs, 0);
	gigaset_close(gs);
	gs = NULL;
	return r;
    }

    /* schedule close on exit */
    atexit(&closeif);

    return 0;
}

/* fmtbytearray
 * format byte array for XML output
 */

static inline char
hex(unsigned i)
{
    if (i < 10) return '0'+i;
    if (i < 16) return 'A'+i-10;
    return '?';
}

static char *
fmtbytearray(char *s, size_t l)
{
    char *res;
    size_t i, ires;

    /* write as sequence of unsigned decimal values */
    if ((res = malloc(3*l)) == NULL) {
	fprintf(stderr, "Out of memory!\n");
	exit(1);
    }
    ires = 0;
    for (i = 0; i < l; i++) {
	res[ires++] = hex((s[i] >> 4) & 0xf);
	res[ires++] = hex(s[i] & 0xf);
	res[ires++] = ' ';
    }
    res[--ires] = '\0';
    free(s);
    return res;
}

/* doescape
 * quote string, escaping non-printable characters
 */
static char *
doescape(char *s, size_t l)
{
    size_t i, ires, lres;
    char *res;

    //FIXME should use XML entities instead of \ escapes?
    lres = l+3;
    for (i = 0; i < l; i++) {
	if (s[i] >= 0 && s[i] < 0x20) lres += 3;
	else if (s[i] == '"' || s[i] == '\\') lres++;
    }
    if ((res = malloc(lres+1)) == NULL) {
	fprintf(stderr, "Out of memory!\n");
	exit(1);
    }
    ires = 0;
    res[ires++] = '"';
    for (i = 0; i < l; i++) {
	if (s[i] >= 0 && s[i] < 0x20) {
	    res[ires++] = '\\';
	    res[ires++] = '0' + (s[i] / 0100);
	    res[ires++] = '0' + ((s[i] % 0100) / 010);
	    res[ires++] = '0' + (s[i] % 010);
	} else {
	    if (s[i] == '"' || s[i] == '\\') res[ires++] = '\\';
	    res[ires++] = s[i];
	}
    }
    res[ires++] = '"';
    res[ires] = '\0';
    free(s);
    return res;
}

/* getcfgpar
 * read one parameter value from the Gigaset base
 */
static char *
getcfgpar(int id, char *name, int type)
{
    __s32 len;
    __u16 param_id;
    __u8 unit;
    char *buf;

    /* skip unreadable values silently */
    if (id < 0 || id >= 1000000) {
	if (gl.verb > 1)
	    fprintf(stderr, "skipped parameter id %d [%s]\n", id, name);
	return NULL;
    }

    param_id = id % 10000;
    unit = id / 10000;
    if (exclist[unit])		/* excluded */
	return NULL;

    if ((len = gigaset_confread(gs, param_id, unit, type==TYPE_BYTE?3:2 /*FIXME*/, &buf)) < 0) {
	if (len == -ERR_ABORT) {
	    /* value not present: ignore silently */
	    if (gl.verb > 1)
		fprintf(stderr, "no value for parameter %s [id %u], unit %u\n",
			name, param_id, unit);
	} else {
	    if (gl.verb > 0) {
		if ((buf = malloc(80)) != NULL) {
		    snprintf(buf, 80, "reading parameter %.40s [id %u], unit %u",
			     name, param_id, unit);
		}
		gigaset_logerr(&gl, GLOG_ERR, -len, gigaset_errno(gs), buf);
		free(buf);
	    }
	}
	return NULL;
    }

    switch (type) {
	case TYPE_BYTE:
	case TYPE_BYTEARRAY:
	case TYPE_LONG:
	    return fmtbytearray(buf, len);
	case TYPE_STRING:
	    if (buf[len-1]) {
		if (gl.verb > 0)
		    fprintf(stderr, "string parameter %s [id %u], unit %u: not zero terminated\n",
			    name, param_id, unit);
		return fmtbytearray(buf, len);	/* fall back to Bytearray */
	    }
	    return doescape(buf, len-1);
	default:
	    if (gl.verb > 0)
		fprintf(stderr, "parameter %s [id %u], unit %u: unknown type %d\n",
			name, param_id, unit, type);
	    return fmtbytearray(buf, len);	/* default to Bytearray */
    }
}

/* myatoi
 * convert decimal string to integer, terminate program with message on error
 * empty string is acceptable and converted to 0
 */
static int
myatoi(const char *s)
{
    int result = 0;
    int sign = 1;

    switch (*s) {
	case '-':
	    sign = -1;
	case '+':
	    s++;
    }
    while (*s) {
	if (*s < '0' || *s > '9') {
	    fprintf(stderr, "Error: bad integer value '%s' in %s\n", s, homerpar_path);
	    exit(1);
	}
	result = 10 * result + (*s++ - '0');
    }
    return result*sign;
}

/* do_param
 * process a single parameter entry
 *argument:
 *	list of name/value pairs
 */
static void
do_param(const char **attr)
{
    int i;
    char *name = NULL;
    char *range = NULL;
    int length = -1;
    int readOnly = -1;
    int type = -1;
    int id = -2;
    char *value;

    /* parse attributes */
    for (i = 0; attr[i]; i += 2) {
	switch (strlookup(attrtab, attr[i])) {
	    case ATTR_NAME:
		name = strdup(attr[i+1]);
		break;
	    case ATTR_RANGE:
		range = strdup(attr[i+1]);
		break;
	    case ATTR_LENGTH:
		if ((length = myatoi(attr[i+1])) < 0) {
		    fprintf(stderr, "Error: bad length attribute %d for parameter '%s' in %s\n",
			    length, attr[i], homerpar_path);
		    exit(1);
		}
		break;
	    case ATTR_READONLY:
		if ((readOnly = myatoi(attr[i+1])) < 0 || readOnly > 1) {
		    fprintf(stderr, "Error: bad readOnly attribute '%s' for parameter '%s' in %s\n",
			    attr[i+1], attr[i], homerpar_path);
		    exit(1);
		}
		break;
	    case ATTR_TYPEINFO:
		if ((type = strlookup(typetab, attr[i+1])) < 0) {
		    fprintf(stderr, "Error: bad typeInfo attribute '%s' for parameter '%s' in %s\n",
			    attr[i+1], attr[i], homerpar_path);
		    exit(1);
		}
		break;
	    case ATTR_ID:
		if ((id = myatoi(attr[i+1])) < -1) {
		    fprintf(stderr, "Error: bad id attribute '%s' for parameter '%s' in %s\n",
			    attr[i+1], attr[i], homerpar_path);
		    exit(1);
		}
		break;
	    default:
		fprintf(stderr, "Error: unexpected parameter attribute '%s' in %s\n",
			attr[i], homerpar_path);
		exit(1);
	}
    }
    if (name == NULL) {
	fprintf(stderr, "Error: parameter without name attribute in %s\n",
		homerpar_path);
	exit(1);
    }
    if (range == NULL) {
	fprintf(stderr, "Error: parameter '%s' without range attribute in %s\n",
		name, homerpar_path);
	exit(1);
    }
    if (length < 0) {
	fprintf(stderr, "Error: parameter '%s' without length attribute in %s\n",
		name, homerpar_path);
	exit(1);
    }
    if (readOnly < 0) {
	fprintf(stderr, "Error: parameter '%s' without readOnly attribute in %s\n",
		name, homerpar_path);
	exit(1);
    }
    if (type < 0) {
	fprintf(stderr, "Error: parameter '%s' without typeInfo attribute in %s\n",
		name, homerpar_path);
	exit(1);
    }
    if (id < -1) {
	fprintf(stderr, "Error: parameter '%s' without id attribute in %s\n",
		name, homerpar_path);
	exit(1);
    }

    /* retrieve value from base */
    if ((value = getcfgpar(id, name, type)) != NULL) {
	fprintf(out,
	"\t<param name=\"%s\" range=\"%s\" length=\"%d\" readOnly=\"%d\" typeInfo=\"%s\" id=\"%d\">\n"
	"\t\t%s\n"
	"\t</param>\n",
			name, range, length, readOnly, typetab[type], id, value);
	free(value);
    }

    free(name);
    free(range);
}

/* starttag
 * called by XML parser for every element start tag encountered
 */
static void XMLCALL
starttag(void *data, const char *el, const char **attr)
{
    const char *s;
    int i;

    switch (Hpar_state) {
	case INIT:
	    if (strcmp(el, "HxParams") != 0) {
		fprintf(stderr, "Error: unexpected toplevel element '%s' in %s\n",
			el, homerpar_path);
		exit(1);
	    }
	    Hpar_state = HX_PARAMS;
	    fprintf(out, "   <HxParams>\n");
	    return;

	case HX_PARAMS:
	    if (strcmp(el, "paramTree") != 0) {
		fprintf(stderr, "Error: unexpected HxParams element '%s' in %s\n",
			el, homerpar_path);
		exit(1);
	    }
	    Hpar_state = PARAMTREE;
	    fprintf(out, "  <paramTree>\n");
	    return;

	case PARAMTREE:
	    if (strcmp(el, "key") != 0) {
		fprintf(stderr, "Error: unexpected paramTree element '%s' in %s\n",
			el, homerpar_path);
		exit(1);
	    }
	    if (attr[0] == NULL) {
		fprintf(stderr, "Error: nameless key element in %s\n",
			homerpar_path);
		exit(1);
	    }
	    if (strcmp(s = attr[0], "name") != 0 || (s = attr[2]) != NULL) {
		fprintf(stderr, "Error: unexpected key attribute '%s' in %s\n",
			s, homerpar_path);
		exit(1);
	    }
	    if (strcmp(s = attr[1], "config") != 0) {
		if (gl.verb > 2)
		    fprintf(stderr, "skipping non-config key '%s'\n", s);
		Hpar_state = SKIP;
		Depth = 0;
		return;
	    }
	    Hpar_state = CONFIG;
	    Depth = 0;
	    fprintf(out, " <key name=\"config\">\n");
	    return;

	case SKIP:
	    Depth++;
	    return;

	case CONFIG:
	    if (strcmp(el, "key") == 0) {
		if (attr[0] == NULL) {
		    fprintf(stderr, "Error: nameless key element in %s\n",
			    homerpar_path);
		    exit(1);
		}
		if (strcmp(s = attr[0], "name") != 0 || (s = attr[2]) != NULL) {
		    fprintf(stderr, "Error: unexpected key attribute '%s' in %s\n",
			    s, homerpar_path);
		    exit(1);
		}
		s = attr[1];
		if (Depth >= MAXKEYS) {
		    fprintf(stderr, "Error: key nesting depth %d exceeds maximum %d in %s at %s\n",
			    Depth, MAXKEYS, homerpar_path, s);
		    exit(1);
		}
		for (i=0; i<Depth; i++)
		    putc(' ', out);
		fprintf(out, "<key name=\"%s\">\n", s);
		Depth++;
		return;
	    } else if (strcmp(el, "param") == 0) {
		Hpar_state = PARAM;
		do_param(attr);
		return;
	    } else {
		fprintf(stderr, "Error: unexpected element '%s' in %s\n",
			el, homerpar_path);
		exit(1);
	    }

	case PARAM:
	    fprintf(stderr, "Error: unexpected param subelement '%s' in %s\n",
		    el, homerpar_path);
	    exit(1);

	default:
	    fprintf(stderr, "Error: unknown state %d\n", Hpar_state);
	    exit(1);
    }
}

/* endtag
 * called by XML parser for every element end tag encountered
 */
static void XMLCALL
endtag(void *data, const char *el)
{
    int i;

    switch (Hpar_state) {
	case PARAM:
	    /* end tag printed with value if any */
	    Hpar_state = CONFIG;
	    return;
	case CONFIG:
	    if (Depth > 0) {
		Depth--;
		for (i=0; i<Depth; i++)
		    putc(' ', out);
		fprintf(out, "</key>\n");
	    } else {
		fprintf(out, " </key>\n");
		Hpar_state = PARAMTREE;
	    }
	    return;
	case SKIP:
	    if (Depth > 0) {
		Depth--;
	    } else {
		Hpar_state = PARAMTREE;
	    }
	    return;
	case PARAMTREE:
	    fprintf(out, "  </paramTree>\n");
	    Hpar_state = HX_PARAMS;
	    return;
	case HX_PARAMS:
	    fprintf(out, "   </HxParams>\n");
	    Hpar_state = INIT;
	    return;
	case INIT:
	    fprintf(stderr, "Element nesting error at top level in %s\n",
		    homerpar_path);
	    exit(1);
	default:
	    fprintf(stderr, "Error: unknown state %d\n", Hpar_state);
	    exit(1);
    }
}

/* command line options */
const char *optstring = "hVvqd:e:o:x:";
const struct option longopts[] = {
    { "help", no_argument, NULL, 'h' },
    { "device", required_argument, NULL, 'd' },
    { "extension", required_argument, NULL, 'e' },
    { "output", required_argument, NULL, 'o' },
    { "homerpar", required_argument, NULL, 'x' },
    { "version", no_argument, NULL, 'V' },
    { "verbose", no_argument, NULL, 'v' },
    { "quiet", no_argument, NULL, 'q' },
    { NULL, 0, NULL, 0 }
};

/* display usage information
 */
static void
usage(char *name)
{
    fprintf(stderr,
	    "\nUsage: %s [OPTION...]\n"
	    "\nOptions:"
	    "\n  -d, --device=<file>       set gigaset interface device"
	    "\n                            (default: %s)"
	    "\n  -e, --extension=<list>    process only settings for specified extensions"
	    "\n                            (comma-separated numbers, hyphen for range,"
	    "\n                            0 for extension independent settings)"
	    "\n  -o, --output=<file>       set output file (default: standard output)"
	    "\n  -x, --homerpar=<file>     set location of homerpar.xml file"
	    "\n                            (default: %s)"
	    "\n  -h, --help                show this message"
	    "\n  -V, --version             show version number"
	    "\n  -v, --verbose             show more details about operations"
	    "\n  -q, --quiet               show less details about operations"
	    "\n\n",
	    name, if_path, homerpar_path);
}

/* parse list of extension numbers
 * error handling: print message to stderr and exit
 */
void parseextlist(char *s, char *tbl)
{
    int c, v, from, to;

    v = 0;
    if (*s == '-') {		/* exclude */
	s++;
	tbl[1] = v = 1;		/* flag -e option seen */
    } else if (!tbl[1]) {
	/* exclude everything not explicitly included */
	memset(tbl, 1, 100);
    }

    from = -1;
    for (;;) {
	switch (c = *s++) {
	    case '\0':		/* end of string */
		return;

	    case '0':		/* base */
		if (isdigit(*s)) {
		    fprintf(stderr, "bad extension number '%.2s' in option -e\n", s-1);
		    exit(1);
		}
		if (from >= 0) {
		    fprintf(stderr, "bad end of range '0' in option -e\n");
		    exit(1);
		}
		tbl[0] = v;
		from = 0;
		break;

	    case '1':		/* extension */
	    case '2':
	    case '4':
	    case '9':
		to = 10 * (c - '0');
		if (!isdigit(c = *s++)) {
		    fprintf(stderr, "bad extension number '%.2s' in option -e\n", s-2);
		    exit(1);
		}
		if (isdigit(*s)) {
		    fprintf(stderr, "bad extension number '%.3s' in option -e\n", s-2);
		    exit(1);
		}
		to += c - '0';

		if (from < 0) {
		    tbl[to] = v;
		    from = to;
		} else {
		    if (to < from) { c = to; to = from; from = c; }
		    while (from <= to) {
			tbl[from++] = v;
		    }
		    from = -1;
		}
		break;

	    case ',':		/* list separator */
	    case ' ':
		from = -1;
		break;
	    case '-':		/* range separator */
	    case ':':
		if (from < 0) {
		    fprintf(stderr, "missing start of range in option -e\n");
		    exit(1);
		}
		if (!isdigit(*s)) {
		    fprintf(stderr, "missing end of range in option -e\n");
		    exit(1);
		}
		break;

	    default:
		fprintf(stderr, "character '%c' illegal in option -e\n", c);
		exit(1);
	}
    }
}

/* main program
 */
int
main(int argc, char *argv[])
{
    XML_Parser p;
    FILE *homerpar;
    static char buff[BUFFSIZE];
    char *s;
    int len;
    int c;

    /* log settings */
    gl.verb = GLOG_ERR | GLOG_WARN;
    gl.log = stderr;
    gl.logfunc = NULL;
    gl.data = NULL;
    gl.errorstr = NULL;

    /* default settings (may be overridden by environment or command line) */
    out = stdout;		/* output destination */

    /* environment settings (may be overridden by command line) */
    if ((s = getenv(homerpar_envvar)) != NULL)
	homerpar_path = s;	/* parameter spec file */
    if ((s = getenv(if_envvar)) != NULL)
	if_path = s;		/* interface device file */

    /* command line settings */
    while ((c = getopt_long(argc, argv, optstring, longopts, NULL)) >= 0) {
	switch (c) {
	    case 'e':		/* extension selection */
		parseextlist(optarg, exclist);
		break;
	    case 'o':		/* output file */
		if ((out = fopen(optarg, "w")) == NULL) {
		    perror(optarg);
		    exit(1);
		}
		break;
	    case 'd':		/* device */
		if_path = optarg;
		break;
	    case 'x':		/* homerpar.xml file */
		homerpar_path = optarg;
		break;
	    case 'v':		/* verbose */
		if (gl.verb)
			gl.verb |= GLOG_INFO | GLOG_STATUS | GLOG_IO | GLOG_DEBUG | GLOG_DUMP;
		else
			gl.verb = GLOG_ERR | GLOG_WARN;
		break;
	    case 'q':		/* quiet */
		if (gl.verb & GLOG_DEBUG)
			gl.verb = GLOG_ERR | GLOG_WARN;
		else
			gl.verb = 0;
		break;
	    case 'V':		/* version */
		fprintf(stderr,
			"%s: gigaset-frontend release " stringify(VERSION) "\n"
			"Copyright (C) 2011 Tilman Schmidt\n",
			argv[0]);
		exit(0);
	    case 'h':		/* help */
		usage(argv[0]);
		exit(0);
	    case '?':		/* unknown option */
		usage(argv[0]);
		exit(1);
	    default:		/* inconsistency between getopt and switch */
		fprintf(stderr, "Command line parse error: option '%c' not implemented\n", c);
		exit(1);
	}
    }
    if (optind < argc) {
	fprintf(stderr, "%s: unexpected argument -- %s\n", argv[0], argv[optind]);
	usage(argv[0]);
	exit(1);
    }

    /* open Gigaset device */
    if (openif(if_path) < 0) {
	exit(1);
    }

    /* traverse homerpar.xml file */
    if ((homerpar = fopen(homerpar_path, "r")) == NULL) {
	perror(homerpar_path);
	exit(1);
    }
    if ((p = XML_ParserCreate(NULL)) == NULL) {
	fprintf(stderr, "Couldn't allocate memory for parser\n");
	exit(1);
    }
    XML_SetElementHandler(p, starttag, endtag);

    Hpar_state = INIT;
    Depth = 0;
    while (!feof(homerpar)) {
	len = fread(buff, 1, BUFFSIZE, homerpar);
	if (ferror(homerpar)) {
	    perror(homerpar_path);
	    exit(1);
	}

	if (XML_Parse(p, buff, len, 0) == XML_STATUS_ERROR) {
	    fprintf(stderr, "Parse error in %s at line %d:\n%s\n",
		    homerpar_path, XML_GetCurrentLineNumber(p),
		    XML_ErrorString(XML_GetErrorCode(p)));
	    exit(1);
	}
    }
    if (XML_Parse(p, buff, 0, 1) == XML_STATUS_ERROR) {
	fprintf(stderr, "Parse error in %s at EOF:\n%s\n",
		homerpar_path, XML_ErrorString(XML_GetErrorCode(p)));
	exit(1);
    }

    exit(0);
}
