/*
 * Main file of the configuration program for Gigaset compatible devices
 *
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

/* WARNING: This is only test code! Parts of it are just for
 * experimenting, other parts are completely untested!
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <stdint.h>

#include <gigaset/interf.h>
#include <gigaset/phonebook.h>
#include <gigaset/gigdata.h>
#include <gigaset/sms.h>
#include <gigaset/am.h>
#include <gigaset/conf.h>
#include <gigaset/basemisc.h>
#include <gigaset/m10x.h>
#include <gigaset/log.h>

#define stringify(x) str(x)
#define str(x) #x

#define MAXLINE		256
#define MAXFIELDS	6

#define CF_OPEN         0x0001
#define CF_LOCK         0x0002

#define CF_UNIMODEM     0x0004
#define CF_CIDMODE      0x0008
#define CF_PHONEBOOK    0x0010
#define CF_SMSMODE      0x0020
#define CF_MESSAGES     0x0040
#define CF_CFGLOCK      0x0080
#define CF_ABMODE       0x0100

#define CF_CONF         0x0200
#define CF_M10XSET      0x0400
#define CF_M10XGET      0x0800

#define CF_ISDN         0x1000

#define CF_MAX          0x1000

#define CF_ANY          ((CF_MAX<<1)-1)

#define CF_STD          (CF_LOCK|CF_OPEN)
#define CF_UNI          (CF_STD|CF_UNIMODEM)
#define CF_CFG          (CF_STD|CF_CONF)
#define CF_MSET         (CF_CFG|CF_M10XSET)
#define CF_MGET         (CF_CFG|CF_M10XGET)
#define CF_CID          (CF_UNI|CF_CIDMODE)
#define CF_SMS          (CF_CID|CF_SMSMODE)
#define CF_AB           (CF_CID|CF_ABMODE)
#define CF_MSG          (CF_SMS|CF_MESSAGES)
#define CF_PB           (CF_CID|CF_PHONEBOOK)
#define CF_LCK          (CF_CID|CF_CFGLOCK)

#define MVAR_FLOW       0
#define MVAR_MODE       1
#define MVAR_RATE       2
#define MVAR_FORMAT     3
#define MVAR_HSHAKE     4
#define MVAR_TYPE       5
#define MVAR_PARTNER    6


/* For DT_XXXX >=0, DT_XXXX&0xf is the second id for cfgr. */
#define DT_INVALID      -1
#define DT_RAW          (2|1<<4)
/*#define DT_DEC8         (1|2<<4) // not working when writing (pid 80). what is 1 for?*/
#define DT_DEC8         (3|2<<4)
#define DT_STR          (2|3<<4)
#define DT_HEX		(2|4<<4)
#define DT_DEC32	(2|5<<4)
 
#define PF_SINGLE       1
#define PF_OVERWRITE    2
#define PF_ABORT        4
#define PF_LAST         8

#define OPT_MAXNAMES    3

#define LINE_WIDTH 80
#define USAGE_INDENT 4
#define HELP_INDENT 2
#define LEFT_COL_WIDTH 15
#define COL_SEP 2

#define RIGHT_COL_WIDTH (LINE_WIDTH - LEFT_COL_WIDTH - COL_SEP)


#define GLOG_APPDBG (0x0001U << GLOG_LIBBITS)
#define GLOG_APPERR (0x0002U << GLOG_LIBBITS)

#define ERR_PARAM_NUM   (ERR_LIBMAX + 1)
#define ERR_COM_INVALID (ERR_LIBMAX + 2)
#define ERR_PARAM_LAST  (ERR_LIBMAX + 3)
#define ERR_NO_ARG      (ERR_LIBMAX + 4)
#define ERR_ONE_ARG     (ERR_LIBMAX + 5)
#define ERR_NULL_ARG    (ERR_LIBMAX + 6)
#define ERR_FILEOPEN    (ERR_LIBMAX + 7)
#define ERR_FILEREAD    (ERR_LIBMAX + 8)
#define ERR_FILEWRITE   (ERR_LIBMAX + 9)

static const char * const errstr[] = {
	"invalid number of parameters",
	"unknown command",
	"command not at end of command line",
	"no command",
	"invalid combination of commands",
	"command is NULL",
	"can't open file",
	"can't read file",
	"can't write file",
};

#define ERRSTR_MIN (ERR_LIBMAX + 1)
#define ERRSTR_MAX (ERRSTR_MIN + sizeof errstr / sizeof errstr[0] - 1)

#define IS_PARAM_ERROR(num) (num>=ERR_PARAM_NUM && num <= ERR_PARAM_LAST)

struct contr_state
{
	struct glog gl;
	struct gstatus *gs;	/* NULL while !(cs.pflags & CF_OPEN) */
	const char * interface;
	unsigned pflags;
	int abort;
	//int checkflags;
	struct gphonebook * pbfirst, * pblast;
	//size_t listlen;
	//char * list;
	struct m10x_state mstate;
	int error;
	int errnum;
	unsigned exit_flags;
	volatile sig_atomic_t *sigint;

	/* SMS */
	struct gigaset_sm_chain * chains;
	struct gigaset_sm * messages;
	struct gigaset_sm_chain * send_chain;
	struct gigaset_sm * send_messages;

	/* at most one of these can be selected */
	struct gigaset_sm_chain * selected_chain;
	struct gigaset_sm * selected_message;
	int handle_selected;
	uint32_t selected_handle;

	int handle_set;
	uint32_t set_handle;

	/* AB */
	unsigned amindex;
};

struct cmd_def {
	const char *const namelist[OPT_MAXNAMES];
	int numpara;
	unsigned parflags;
	unsigned flags_allow;
	unsigned flags_need;
	void (*func)(struct contr_state*, char**, unsigned long);
	unsigned long data;
	const char *const help_params;
	const char *const help_description;
};

struct param_info
{
	char ** param;
	const struct cmd_def * cmd;
};

static int tolong(const char *s, long *result, int base)
{
	char *e;

	errno = 0;
	*result = strtol(s, &e, base);
	if (errno || e == s)
		return 0;
	
	while (isspace((unsigned char) *e))
		++e;

	if (*e)
		return 0;

	return 1;
}

static inline int tolongrange(const char *s, long *result, long min, long max, int base)
{
	return tolong(s, result, base) && *result >= min && *result <= max;
}

static inline int tointrange(const char *s, int *result, int min, int max, int base)
{
	long tmp;

	if (!tolongrange(s, &tmp, min, max, base))
		return 0;
	*result = tmp;
	return 1;
}

static inline int toint(const char *s, int *result, int base)
{
	return tointrange(s, result, INT_MIN, INT_MAX, base);
}

static int toulong(const char *s, unsigned long *result, int base)
{
	char *e;

	errno = 0;
	*result = strtoul(s, &e, base);
	if (errno || e == s)
		return 0;

	while (isspace((unsigned char) *e))
		++e;

	if (*e)
		return 0;

	return 1;
}

static inline int toulongrange(const char *s, unsigned long *result, unsigned long min, unsigned long max, int base)
{
	return toulong(s, result, base) && *result >= min && *result <= max;
}

static inline int touintrange(const char *s, unsigned *result, unsigned min, unsigned max, int base)
{
	unsigned long tmp;

	if (!toulongrange(s, &tmp, min, max, base))
		return 0;
	*result = tmp;
	return 1;
}

static inline int touint(const char *s, unsigned *result, int base)
{
	return touintrange(s, result, 0, UINT_MAX, base);
}

static void set_error_num(struct contr_state *cs, int num, int err)
{
	if (num >= 0 || cs->error)
		return;

	cs->error = num;
	cs->errnum = err;
	cs->abort = 1;
}

static void set_error_num_msg(struct contr_state *cs, int num, int err, const char *text)
{
	if (num >= 0)
		return;

	gigaset_logerr(&cs->gl, GLOG_APPERR, -num, err, text);
	set_error_num(cs, num, err);
}

static void set_error_msg(struct contr_state *cs, int num, const char *text)
{
	set_error_num_msg(cs, num, gigaset_errno(cs->gs), text);
}

static void cmd_get_string(struct contr_state*cs, char**argv, unsigned long data)
{
	char * s;
	int r;

	r=((int(*)(struct gstatus*, char**))data)(cs->gs,&s);
	if (r<0)
		set_error_msg(cs, r, argv[-1]);
	else {
		puts(s);
		free(s);
	}
}

static void cmd_get_str_idx(struct contr_state *cs, char **argv, unsigned long data)
{
	char *s;
	int r;
	unsigned num;

	if (!touint(argv[0], &num, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	r = ((int(*)(struct gstatus*, unsigned, char**))data)(cs->gs, num, &s);
	if (r < 0)
		set_error_msg(cs, r, argv[-1]);
	else {
		puts(s);
		free(s);
	}
}

static void cmd_set_str_idx(struct contr_state *cs, char **argv, unsigned long data)
{
	int r;
	unsigned num;

	if (!touint(argv[0], &num, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	r=((int(*)(struct gstatus*, unsigned, const char*))data)(cs->gs,num,argv[1]);
	if (r < 0)
		set_error_msg(cs, r, argv[-1]);
}

static void cmd_m10x_getvar(struct contr_state*cs, char**argv, unsigned long data)
{
	switch (data) {
	case MVAR_FLOW:
		printf("%d\n", cs->mstate.flow);
		break;
	case MVAR_MODE:
		printf("%d\n", cs->mstate.mode);
		break;
	case MVAR_RATE:
		printf("%d\n", cs->mstate.baudrate);
		break;
	case MVAR_FORMAT:
		printf("%d\n", cs->mstate.format);
		break;
	case MVAR_HSHAKE:
		printf("%d\n", cs->mstate.handshake);
		break;
	case MVAR_TYPE:
		printf("%d\n", cs->mstate.type);
		break;
	case MVAR_PARTNER:
		printf("%d\n", cs->mstate.partner);
		break;
	default:
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[-1]);
	}
}

static void cmd_m10x_setvar(struct contr_state*cs, char**argv, unsigned long data)
{
	unsigned num;

	if (!touint(argv[0], &num, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	switch (data) {
	case MVAR_FLOW:
		cs->mstate.flow=num;
		break;
	case MVAR_MODE:
		cs->mstate.mode=num;
		break;
	case MVAR_RATE:
		cs->mstate.baudrate=num;
		break;
	case MVAR_FORMAT:
		cs->mstate.format=num;
		break;
	case MVAR_HSHAKE:
		cs->mstate.handshake=num;
		break;
	default:
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[-1]);
	}
}

int parse_ext_char(unsigned char ** ps, unsigned char * dest)
{
	unsigned char c;
	unsigned val=0;
	unsigned hex=0;
	unsigned digits=3;
	if (!ps||!*ps||!**ps) return 0;
	c=*(*ps)++;
	switch(c) {
	case '\\':
		c=**ps;
		if(!c) return 0;
		++*ps;
		switch (c) {
		case 'x':
		case 'X':
			hex=1;
			digits=2;
			/* fall through */
		case '0':
			while(digits) {
				c=**ps;
				if(c>='0'&&c<='7')
					c-='0';
				else if (!hex)
					break;
				else if (c=='8'||c=='9')
					c-='0';
				else if (c>='a' && c<='z')
					c-='a'-10;
				else if (c>='A' && c<='Z')
					c-='A'-10;
				else
					break;
				val<<=hex?4:3;
				val+=c;
				digits--;
				++*ps;
			}
			if((hex&&digits==2)||val>0xff) return 0;
			*dest=val;
			break;
		default:
			*dest=c;
		}
		break;
	case '^':
		c=**ps;
		if(!c) return 0;
		++*ps;
		if (c>=0x40&&c<0x60) {
			*dest=c^0x40;
			return 1;
		}
		if (c>=0x60&&c<0x80) {
			*dest=c^0x60;
			return 1;
		}
		return 0;
	default:
		*dest=c;
	}
	return 1;
}

//FIXME additional "don't care" column for CF_xxxx (here: CID-mode etc.)
//FIXME test for lock state in driver?
static void cmd_bchars(struct contr_state*cs, char**argv, unsigned long data)
{
	unsigned char buf[6]={0};
	unsigned char*pos=buf+5;
	unsigned char*s=argv[0];
	int r;

	for (;*s;) {
		if(!parse_ext_char(&s,pos)) {
			set_error_num_msg(cs, -ERR_PARAMINVALID, 0, s);
			return;
		}
		if (pos==buf) break;
		--pos;
	}
	if (*s) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, s);
		return;
	}
	r=gigaset_brkchars(cs->gs, buf);
	if (r<0)
		set_error_msg(cs, r, "brkchars");
}

static void cmd_verbosity(struct contr_state *cs, char **argv, unsigned long data)
{
	cs->gl.verb = 0;
	switch (data) {
	case 3:
		cs->gl.verb |= GLOG_APPDBG | GLOG_INFO | GLOG_STATUS | GLOG_DEBUG | GLOG_DUMP;
	case 2:
		cs->gl.verb |= GLOG_IO;
	case 1:
		cs->gl.verb |= GLOG_APPERR | GLOG_ERR| GLOG_WARN;
	case 0:
		;
	}
}

static void cmd_sleep(struct contr_state*cs, char**argv, unsigned long data)
{
	unsigned num;

	if (!touintrange(argv[0], &num, 0, 2147, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	gigaset_usleep_unint((int_fast32_t)num * 1000000);
}

static void cmd_mode(struct contr_state*cs, char**argv, unsigned long data)
{
	if      (!strcmp(argv[0],"config"  )) cs->exit_flags=CF_LOCK|CF_CONF;
	else if (!strcmp(argv[0],"cid"     )) cs->exit_flags=CF_LOCK|CF_CIDMODE;
	else if (!strcmp(argv[0],"unimodem")) cs->exit_flags=CF_LOCK|CF_UNIMODEM;
	else if (!strcmp(argv[0],"isdn"    )) cs->exit_flags=CF_ISDN;
	else if (!strcmp(argv[0],"locked"  )) cs->exit_flags=CF_LOCK|(cs->pflags&(CF_CONF|CF_CIDMODE|CF_UNIMODEM));
	else
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
}

static void cmd_dev(struct contr_state*cs, char**argv, unsigned long data)
{
	cs->interface=argv[0];
}

static void cmd_mreset(struct contr_state*cs, char**argv, unsigned long data)
{
	int r;
	r=m10x_reset(cs->gs);
	if (r<0) {
		set_error_msg(cs, r, "M10x reset");
		return;
	}
	cs->pflags=0;
	r=gigaset_lock(cs->gs,0);
	if (r<0)
		set_error_msg(cs, r, "lock");
}

static void cmd_msetpartner(struct contr_state *cs, char **argv, unsigned long data)
{
	unsigned num;
	int r;

	if (!touint(argv[0], &num, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	r = m10x_setpartner(cs->gs, num);
	set_error_msg(cs, r, "M10x setpartner");
}

static void cmd_mlistnames(struct contr_state *cs, char **argv, unsigned long data)
{
	unsigned num;
	int r;
	int i;
	char **list;

	if (!touintrange(argv[0], &num, 0, 1, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	r = m10x_adapterlist(cs->gs, num, &list);
	if (r < 0) {
		set_error_msg(cs, r, "M10x adapterlist");
		return;
	}

	for (i = 0; i < r; ++i) {
		printf("%d: %s\n", i, list[i]);
		free(list[i]);
	}
	free(list);
}

static void cmd_intlist(struct contr_state *cs, char **argv, unsigned long data)
{
	unsigned num;
	int r;
	int i;
	unsigned *list;

	if (!touintrange(argv[0], &num, 0, 1, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	r = ((int(*)(struct gstatus*, int, unsigned**))data)(cs->gs, num, &list);
	if (r < 0) {
		set_error_msg(cs, r, "int list");
		return;
	}

	for (i = 0; i < r; ++i)
		printf("%d: %u\n", i, list[i]);

	free(list);
}

static void cmd_mlisten(struct contr_state*cs, char**argv, unsigned long data)
{
	int r;
	r=m10x_listen(cs->gs);
	if (r<0)
		set_error_msg(cs, r, "M10x listen");
}

static void cmd_msettype(struct contr_state*cs, char**argv, unsigned long data)
{
	int r;
	unsigned num;

	if (!touintrange(argv[0], &num, 0, 1, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	r=m10x_settype(cs->gs,num);
	if (r<0)
		set_error_msg(cs, r, "M10x settype");
}

static void cmd_mregister(struct contr_state*cs, char**argv, unsigned long data)
{
	int r;
	unsigned index;

	if (!touint(argv[0], &index, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	r=m10x_register(cs->gs,argv[0],index);
	if (r < 0)
		set_error_msg(cs, r, "registering M10x");
}

static void cmd_moccupyconn(struct contr_state *cs, char **argv, unsigned long data)
{
	int r;
	r = m10x_occupyconn(cs->gs);
	if (r < 0)
		set_error_msg(cs, r, "occupying connection");
}

static void cmd_mreleaseconn(struct contr_state *cs, char **argv, unsigned long data)
{
	int r;
	r = m10x_releaseconn(cs->gs);
	if (r < 0)
		set_error_msg(cs, r, "releasing connection");
}

static void cmd_msetautorel(struct contr_state *cs, char **argv, unsigned long data)
{
	int r;
	unsigned secs;

	if (!touint(argv[0], &secs, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	r = m10x_setautorel(cs->gs, secs);
	if (r < 0)
		set_error_msg(cs, r, "setting auto release timer");
}

/* process a file of phonebook entries */
static void cmd_pbfile(struct contr_state *cs, char **argv, unsigned long data)
{
	FILE *f;
	struct gphonebook *pbcur;
	char *s;
	char *buf = NULL;
	size_t n;
	char *field[MAXFIELDS];
	int nfields;

	gigaset_log(&cs->gl, GLOG_APPDBG, "PBFILE %s %s\n", argv[0], argv[1]);
	if (!strcmp(argv[0],"-"))
		f=stdin;
	else {
		f=fopen(argv[0], "r");
		if (!f) {
			set_error_num_msg(cs, -ERR_FILEOPEN, errno, argv[0]);
			return;
		}
	}

	for(;;) {
		/* read next line */
		buf = malloc(MAXLINE);
		if (!buf) {
			set_error_num_msg(cs, -ERR_MEM, 0, "reading phonebook file");
			break;
		}
		if (!fgets(buf, MAXLINE, f)) {
			/* EOF or error */
			if (ferror(f))
				set_error_num_msg(cs, -ERR_FILEREAD, errno, argv[0]);
			break;
		}
		n=strlen(buf);
		if (!n||buf[n-1]!='\n') {
			gigaset_log(&cs->gl, GLOG_APPERR, "line too long or missing newline!\n");
			set_error_num(cs, -ERR_DATA, 0);
			break;
		}
		buf[n-1]='\0';

		/* split on tabs */
		for (s = buf, nfields = 1; nfields <= MAXFIELDS; nfields++) {
			field[nfields-1]=s;
			s=strchr(s,'\t');
			if (!s) break;
			*s++='\0';
		}

		if (nfields < 2) {
			gigaset_log(&cs->gl, GLOG_APPERR, "tab not found!\n");
			set_error_num(cs, -ERR_DATA, 0);
			break;
		}

		/* build entry */
		pbcur = calloc(1, sizeof *pbcur);
		if (!pbcur) {
			set_error_num_msg(cs, -ERR_MEM, 0, "reading phonebook file");
			break;
		}
		pbcur->name = field[0];
		pbcur->number = field[1];
		if (nfields >= 3) {
			pbcur->melody = strtol(field[2], &s, 10);
			if (*s || pbcur->melody < 0 || pbcur->melody > 10) {
				gigaset_log(&cs->gl, GLOG_APPERR, "bad melody number %s!\n", field[2]);
				set_error_num(cs, -ERR_DATA, 0);
				break;
			}
		} else {
			pbcur->melody = -1;
		}
		if (nfields >= 6) {
			pbcur->alert = strtol(field[3], &s, 10);
			if (*s || pbcur->alert < 0 || pbcur->alert > 1) {
				gigaset_log(&cs->gl, GLOG_APPERR, "bad alert flag %s!\n", field[3]);
				set_error_num(cs, -ERR_DATA, 0);
				break;
			}
			pbcur->hour = strtol(field[4], &s, 10);
			if (!*s) {
				/* hhmm */
				pbcur->minute = pbcur->hour % 100;
				pbcur->hour /= 100;
			} else if (*s++ == ':') {
				/* hh:mm */
				pbcur->minute = strtol(s, &s, 10);
			}
			if (*s || pbcur->hour < 0 || pbcur->hour > 23 || pbcur->minute < 0 || pbcur->minute > 59) {
				gigaset_log(&cs->gl, GLOG_APPERR, "bad alert time %s!\n", field[4]);
				set_error_num(cs, -ERR_DATA, 0);
				break;
			}

			pbcur->year = strtol(field[5], &s, 10);
			switch (*s) {
			case '\0':
				/* [yy]mmdd */
				pbcur->day = pbcur->year % 100;
				pbcur->year /= 100;
				pbcur->month = pbcur->year % 100;
				pbcur->year /= 100;
				break;
			case '-':
				s++;
				pbcur->month = strtol(s, &s, 10);
				if (!*s) {
					/* mm-dd */
					pbcur->day = pbcur->month;
					pbcur->month = pbcur->year;
					pbcur->year = 0;
				} else if (*s++ == '-')
					/* yy-mm-dd */
					pbcur->day = strtol(s, &s, 10);
				break;
			case '.':
				/* dd.mm.[yy] */
				pbcur->day = pbcur->year;
				s++;
				pbcur->month = strtol(s, &s, 10);
				if (!*s)
					pbcur->year = 0;
				else if (*s++ == '.')
					pbcur->year = strtol(s, &s, 10);
			}
			if (*s || pbcur->year < 0 || pbcur->year > 99 || pbcur->month < 1 || pbcur->month > 12 || pbcur->day < 1 || pbcur->day > 31) {
				gigaset_log(&cs->gl, GLOG_APPERR, "bad alert date %s!\n", field[5]);
				set_error_num(cs, -ERR_DATA, 0);
				break;
			}
			if (nfields > 6) {
				gigaset_log(&cs->gl, GLOG_WARN, "%d extra fields ignored!\n", nfields-6);
			}
		} else {
			pbcur->alert = -1;
			if (nfields > 3) {
				gigaset_log(&cs->gl, GLOG_WARN, "incomplete alert (%d fields) ignored!\n", nfields-3);
			}
		}

		/* put onto chain */
		if (cs->pblast)
			cs->pblast->next=pbcur;
		else
			cs->pbfirst=pbcur;
		cs->pblast=pbcur;
		pbcur->next=NULL;
		pbcur->pnr=argv[1];
		pbcur->buffer=buf;
		buf=NULL;
	}
	if (f!=stdin) fclose(f);
	free(buf);
}

/* queue one phonebook entry for transmission to handset */
static void cmd_pbent(struct contr_state *cs, char **argv, unsigned long data)
{
	struct gphonebook *pbcur;

	gigaset_log(&cs->gl, GLOG_APPDBG, "PBENT %s %s %s\n", argv[0], argv[1], argv[2]);
	pbcur = malloc (sizeof *pbcur);
	if (!pbcur) {
		set_error_num_msg(cs, -ERR_MEM, 0, "building phonebook entry");
		return;
	}
	if (cs->pblast)
		cs->pblast->next=pbcur;
	else
		cs->pbfirst=pbcur;
	cs->pblast=pbcur;
	pbcur->next=NULL;
	pbcur->name=argv[0];
	pbcur->number=argv[1];
	pbcur->pnr=argv[2];
	pbcur->buffer=NULL;
}

static void cmd_pbdel(struct contr_state *cs, char **argv, unsigned long data)
{
	ssize_t s;

	s = gigaset_delpb(cs->gs, argv[0]);
	if (s < 0) {
		set_error_msg(cs, s, "deleting phonebook entries");
		return;
	}
}

static void cmd_pbdump(struct contr_state *cs, char **argv, unsigned long data)
{
	ssize_t s;
	struct gphonebook *list;
	struct gphonebook *entry, *next;
	int method;

	if (!tointrange(argv[1], &method, 0, 2, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[1]);
		return;
	}

	if (method == PB_TRANSFER_OLD)
		printf("Please start manual phonebook transfer at handset %s\n",
		       argv[0]);

	s = gigaset_readpbarray(cs->gs, &list, argv[0], method);
	if (s < 0) {
		set_error_msg(cs, s, "getting phonebook entries");
		return;
	}

	printf("Entries: %ld\n", (long)s);
	printf("Name\tNumber\tMelody\tAlert\thh:mm\tyy-mm-dd\n");
	entry = list;
	while (entry) {
		printf("%s\t%s",
		       entry->name ? entry->name : "",
		       entry->number ? entry->number : "");
		if (entry->melody >= 0) {
			printf("\t%.2i\t%.2i\t%.2i:%.2i\t%.2i-%.2i-%.2i",
			       entry->melody, entry->alert,
			       entry->hour, entry->minute,
			       entry->year, entry->month, entry->day);
		}
		printf("\n");
		next = entry->next;
		free(entry->buffer);
		free(entry);
		entry = next;
	}
}

static void cmd_dial(struct contr_state *cs, char **argv, unsigned long data)
{
	unsigned type;
	int ret;

	if (!touintrange(argv[0], &type, 0, 4, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	ret = gigaset_dial(cs->gs, type, argv[1], argv[2]);
	if (ret < 0)
		set_error_msg(cs, ret, "dial");
}

static void cmd_set_cfx(struct contr_state *cs, char **argv, unsigned long data)
{
	unsigned mode, proc, service;
	int ret;

	if (!touintrange(argv[1], &mode, 0, 1, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[1]);
		return;
	}
	if (!touintrange(argv[2], &proc, 0, 2, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[2]);
		return;
	}
	if (!touintrange(argv[3], &service, 0, 38, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[3]);
		return;
	}

	ret = gigaset_cfxset(cs->gs, argv[0], mode, proc, service, argv[4]);
	if (ret < 0)
		set_error_msg(cs, ret, "setcfx");
}

static void cmd_smshandle(struct contr_state*cs, char**argv, unsigned long data)
{
	uint32_t handle;

	if (gigaset_32fromdec(&handle, argv[0])<0) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	cs->selected_message=NULL;
	cs->selected_chain=NULL;
	cs->handle_selected=1;
	cs->selected_handle=handle;
}

static struct gigaset_sm * findmsg (struct contr_state*cs,char*name)
{
	uint32_t handle;
	struct gigaset_sm * msg;

	if (name[0]=='-') {
		if (cs->send_messages&&gigaset_32fromdec(&handle, name+1)>=0) {
			if (cs->send_chain) {
				if(handle&&--handle<cs->send_chain->msgcount)
					return cs->send_chain->msg[handle];
			} else {
				if(handle==1)
					return cs->send_messages;
			}
		}
	} else {
		if (cs->messages&&gigaset_32fromdec(&handle, name)>=0) {
			msg=cs->messages;
			do {
				if (msg->handle==handle) return msg;
				msg=msg->next;
			} while (msg!=cs->messages);
		}
	}
	return NULL; /*FIXME: msg*/
}

int createmsg(struct contr_state*cs, struct gigaset_sm * msg, const char * number, unsigned flags)
{
	int r;

	cs->handle_set=0;
	r=gigaset_smsnew(cs->gs, &msg->handle);
	if(r<0) return r;
	cs->set_handle=msg->handle;
	cs->handle_set=1;

	r = gigaset_smsputmessage(cs->gs, msg); //FIXME delete if <0 ?
	if(r<0) return r;

	if(*number) {
		r=gigaset_smsnumber(cs->gs,number);
		if(r<0) return r;
	}


	if(flags&SMS_STORE) {
		r=gigaset_smsstore(cs->gs);
		if(r<0) return r;
	}

	if(flags&SMS_SEND) {
		r=gigaset_smssend(cs->gs);
		if(r<0) return r;
	}

	return 0;
}

#ifdef _WIN32
/* copy from dietlibc */
static char *strsep(char **stringp, const char *delim) {
  register char *tmp=*stringp;
  register char *tmp2=tmp;
  register const char *tmp3;
  if (!*stringp) return 0;
  for (tmp2=tmp; *tmp2; ++tmp2) {
    for (tmp3=delim; *tmp3; ++tmp3)
      if (*tmp2==*tmp3) {	/* delimiter found */
	*tmp2=0;
	*stringp=tmp2+1;
	return tmp;
      }
  }
  *stringp=0;
  return tmp;
}
#endif

static void cmd_smscreate(struct contr_state*cs, char**argv, unsigned long data)
{
	unsigned flags, unknown1, dcs, vp;
	int chain_id;
	size_t len;
	size_t bufsize;
	char * buf;
	unsigned creationflags;
	unsigned i;
	int chain;
	int r;
	FILE * f;
	char * token, * s;

	gigaset_smsfreemessages(&cs->send_messages, &cs->send_chain);
	cs->selected_chain=NULL;
	cs->selected_message=NULL;
	cs->handle_selected=0;

	if (!touintrange(argv[0], &flags, 0, 255, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	if (!touintrange(argv[1], &unknown1, 0, 255, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[1]);
		return;
	}
	if (!touintrange(argv[2], &dcs, 0, 255, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[2]);
		return;
	}
	if (!touintrange(argv[3], &vp, 0, 255, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[3]);
		return;
	}

	if (!tointrange(argv[4], &chain_id, -1, 255, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[4]);
		return;
	}

	if (strlen(argv[5]) > SMS_LEN_NUMBER) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[5]);
		return;
	}

	s=argv[7];
	creationflags=0;
	while ((token=strsep(&s,","))) {
		if (!*token)
			continue;
		if (!strcmp(token,"send"))
			creationflags|=SMS_SEND;
		else if (!strcmp(token,"store"))
			creationflags|=SMS_STORE;
		else {
			set_error_num_msg(cs, -ERR_PARAMINVALID, 0, token);
			return;
		}
	}

	bufsize=255*(160-7)*3+1; /* worst case: every character needs 3 chars (only UCS2 supported) */
	buf=malloc(bufsize);
	if (!buf) {
		set_error_num_msg(cs, -ERR_MEM, 0, "smscreate");
		return;
	}
	if(!strcmp(argv[6],"-"))
		f=stdin;
	else {
		f = fopen(argv[6], "r");
		if (f == NULL) {
			set_error_num_msg(cs, -ERR_FILEOPEN, errno, argv[6]);
			free(buf);
			return;
		}
	}
	len=fread(buf,1,bufsize,f);
	if(ferror(f)) {
		set_error_num_msg(cs, -ERR_FILEREAD, errno, argv[6]);
		free(buf);
		if (f!=stdin) fclose(f);
		return;
	}
	if(len == bufsize) {
		set_error_num_msg(cs, -ERR_LENGTH, 0, argv[6]);
		free(buf);
		if (f!=stdin) fclose(f);
		return;
	}
	if (f!=stdin) fclose(f);

	if (chain_id>=0)
		chain=1;
	else {
		chain=0;
		chain_id=0;
	}

	r=gigaset_smsfromutf8 (&cs->send_messages, &cs->send_chain,
	                       flags, unknown1, dcs, vp,
	                       chain, chain_id, buf, len);
	free(buf);
	if (r<0) {
		set_error_msg(cs, r, "converting from UTF8");
		return;
	}
	if (r == 0) {
		/* single message */
		if (!cs->send_messages) {
			r = -ERR_PARAMINVALID;	//FIXME
			goto creerr;
		}
		r=createmsg(cs,cs->send_messages, argv[5], creationflags);
		if (r < 0)
			goto creerr;
		cs->selected_message = cs->send_messages;
		cs->selected_chain = NULL;
	} else {
		if (!cs->send_chain) {
			r = -ERR_PARAMINVALID;	//FIXME
			goto creerr;
		}
		for (i=0;i<cs->send_chain->msgcount;++i) {
			r=createmsg(cs,cs->send_chain->msg[i], argv[5], creationflags);
			if (r < 0)
				goto creerr; //FIXME delete msgs <i
		}
		cs->selected_chain = cs->send_chain;
		cs->selected_message = NULL;
	}
	return;	/* success */

creerr:
	//FIXME delete?
	set_error_msg(cs, r, "creating message");
	gigaset_smsfreemessages(&cs->send_messages, &cs->send_chain);
}

static void cmd_smsselect(struct contr_state*cs, char**argv, unsigned long data)
{
	struct gigaset_sm * msg;
	msg=findmsg(cs,argv[0]);
	if (!msg) {
		set_error_num_msg(cs, -ERR_NOENT, 0, "smsselect");
		return;
	}
	cs->selected_message=msg;
	cs->selected_chain=NULL;
	cs->handle_selected=0;
}

static void cmd_smsselectmsg(struct contr_state*cs, char**argv, unsigned long data)
{
	struct gigaset_sm * msg;
	msg=findmsg(cs,argv[0]);
	if (!msg) {
		set_error_num_msg(cs, -ERR_NOENT, 0, "smsselectmsg");
		return;
	}
	if (msg->chain) {
		gigaset_log(&cs->gl, GLOG_APPDBG, "selected chain (%s)\n", argv[0]);
		cs->selected_message=NULL;
		cs->selected_chain=msg->chain;
	} else {
		gigaset_log(&cs->gl, GLOG_APPDBG, "selected message (%s)\n", argv[0]);
		cs->selected_message=msg;
		cs->selected_chain=NULL;
	}
	cs->handle_selected=0;
}

static int selectmsg(struct contr_state*cs,uint32_t handle)
{
	int r;
	if(cs->handle_set&&handle==cs->set_handle) return 0;
	cs->handle_set=0;
	gigaset_log(&cs->gl, GLOG_APPDBG,"--> select handle %lu\n", (unsigned long) handle);
	r=gigaset_smsselect(cs->gs,handle);
	if (r<0) {
		set_error_msg(cs, r, "selecting msg");
		return r;
	}
	cs->set_handle=handle;
	cs->handle_set=1;
	return 0;
}

static void printmsg(struct contr_state *cs)
{
	ssize_t s;
	unsigned char * buf, *buf2, *buf3;
	uint16_t len;
	uint16_t septets;

	s = gigaset_smsinfo(cs->gs, (char**)&buf);
	if (s < 0) {
		set_error_msg(cs, s, "getting msg info");
		return;
	}
	if (s<15) {
		printf ("info: invalid length (%d)\n", (int)s);
	} else {
		printf("Date: %s%.2s-%.2s-%.2s %.2s:%.2s\n",
		       (buf[8]<'8')? "20":"19", buf+8, buf+6, buf+4, buf+2, buf+0);
		printf("Number: %.*s\n", s>35?20:(int)s-15, buf+15);
		if (strncmp(buf+11, "FFFF", 4))
			printf("Error: %.4s\n", buf+11);
	}
	free(buf);
	s = gigaset_smsgettext (cs->gs, (char**)&buf);
	if (s < 0) {
		set_error_msg(cs, s, "getting msg text");
		return;
	}

	fputs("\n", stdout);
	len=(uint16_t)buf[0]<<8|buf[1];
	septets=(uint16_t)buf[2]<<8|buf[3];
	if (septets-septets/8+4<=len&&len+4<=(size_t)s) {
		buf2=malloc(septets);
		if (!buf2) {
			gigaset_log(&cs->gl, GLOG_APPERR, "buf2=NULL\n");
		} else {
			gigaset_unpackseptetbuf(buf2,buf+8,septets,0);
			buf3=gigaset_septets_to_utf8 (buf2, septets, NULL);
			free(buf2);
			if (!buf3) {
				gigaset_log(&cs->gl, GLOG_APPERR, "buf3=NULL\n");
			} else {
				fputs (buf3, stdout);
				fputs ("\n", stdout);
				free(buf3);
			}
		}
	} else {
		gigaset_log(&cs->gl, GLOG_APPERR, "invalid sizes (total %u, message %u, septets %u\n",
		            (unsigned)s, (unsigned)len, (unsigned)septets);
	}
	free(buf);
	fputs("\n\n", stdout);
}

static int iterate_smsprint(struct contr_state *cs, unsigned long data)
{
	printmsg(cs);
	return 0;
}

static int iterate_smslib(struct contr_state *cs, unsigned long data)
{
	return ((int(*)(struct gstatus *gs))data)(cs->gs);
}

static int iterate_smsnumber(struct contr_state *cs, unsigned long data)
{
	return gigaset_smsnumber(cs->gs, (char *) data);
}

static void iterate(struct contr_state *cs, int (*func)(struct contr_state*, unsigned long), unsigned long data)
{
	int r;
	unsigned i;
	struct gigaset_sm *msg;
	uint32_t handle;

	if (cs->selected_chain) {
		for (i = 0; i < cs->selected_chain->msgcount; ++i) {
			msg = cs->selected_chain->msg[i];
			if (!msg)
				continue;
			if (selectmsg(cs,msg->handle) < 0)
				continue;
			r = func(cs, data);
			if (r < 0)
				set_error_msg(cs, r, "smsfunction");
		}
	} else {
		if (cs->selected_message)
			handle = cs->selected_message->handle;
		else if (cs->handle_selected)
			handle = cs->selected_handle;
		else {
			set_error_num_msg(cs, -ERR_NOENT, 0, "smsfunction");
			return;
		}
		if (selectmsg(cs, handle) < 0)
			return;
		r = func(cs, data);
		if (r < 0)
			set_error_msg(cs, r, "smsfunction");
	}
}

static void cmd_smsfunction(struct contr_state*cs, char**argv, unsigned long data)
{
	iterate(cs, iterate_smslib, data);
}

static void cmd_smsnumber(struct contr_state*cs, char**argv, unsigned long data)
{
	iterate(cs, iterate_smsnumber, (unsigned long) argv[0]);
}

static void cmd_smsget(struct contr_state*cs, char**argv, unsigned long data)
{
	iterate(cs, iterate_smsprint, 0);
}

static void cmd_smsreset(struct contr_state*cs, char**argv, unsigned long data)
{
	gigaset_smsfreemessages (&cs->messages, &cs->chains);
	gigaset_smsfreemessages (&cs->send_messages, &cs->send_chain);
	cs->selected_message=NULL;
	cs->selected_chain=NULL;
	cs->handle_selected=0;
	cs->handle_set=0;
	set_error_msg(cs, gigaset_smsreset(cs->gs), "smsreset");
}

static void cmd_smslevel(struct contr_state*cs, char**argv, unsigned long data)
{
	int r;
	r=gigaset_smslevel(cs->gs);
	if (r<0) {
		set_error_msg(cs, r, "smslevel");
		return;
	}
	printf ("%d\n", r);
}

static void cmd_smsdump(struct contr_state*cs, char**argv, unsigned long data)
{
	size_t listlen;
	char * list;
	size_t pos;
	uint32_t handle;
	uint16_t flags;
	ssize_t s;

	s=gigaset_smsgetlist (cs->gs, &list);
	if (s<0) return;
	listlen=s;
	printf (" %lu entries:\n", (unsigned long) (listlen/6));
	for (pos=0;pos+6<=listlen;pos+=6) {
		handle = gigaset_from_be32(list+pos);
		flags  = gigaset_from_be16(list+pos+4);
		if (!handle&&!flags) {
			printf (" end marker after %lu entries.\n", (unsigned long) (pos/6));
			break;
		}
		printf("Handle: %-10lu\n", (unsigned long) handle);
		printf("Flags: %s%s%s%s%s%s%s%s\n",
		       (flags & 0x10) ? "In " : "Out ",
		       (flags & 0x08) ? "New " : "",
		       (flags & 0x04) ? "Store " : "",
		       (flags & 0x01) ? "Send " : "",
		       (flags & 0x02) ? "Sent " : "",
		       (flags & 0x20) ? "Error " : "",
		       (flags & 0x40) ? "Delete " : "",
		       (flags & 0x80) ? "Tmp " : "");
		if (selectmsg(cs, handle) < 0)
			continue;
		printmsg(cs);
	}
}

static const char *bitlist[8][3] = {
	{"[TTX] send",   "no",  "yes"},
	{"[ITX] sent",   "no",  "yes"},
	{"[STR] store",  "no",  "yes"},
	{"[NW ] new",    "no",  "yes"},
	{"[INL] list",   "out", "in"},
	{"[ERR] error",  "no",  "yes"},
	{"[DEL] delete", "no",  "yes"},
	{"[TMP] -",      "0",   "1"}
};

static void print_header(struct gigaset_sm * msg)
{
	unsigned i,mask;

	printf ("%-10lu %04lx\n", (unsigned long) msg->handle, (unsigned long) msg->status);
	printf ("change counter: %u\nMSN index: %u\n",
	        (msg->status >> SMS_CHANGE_SHIFT) & SMS_CHANGE_MASK,
	        (msg->status >> SMS_MSN_SHIFT) & SMS_MSN_MASK);
	for (i=0,mask=1;i<8;++i,mask<<=1)
		printf ("%-14s <%s>\n", bitlist[i][0], bitlist[i][msg->status&mask?2:1]);

	//FIXME more (incl. udh)
}

static void cmd_smsdumpext(struct contr_state*cs, char**argv, unsigned long data)
{
	struct gigaset_sm * msg;
	struct gigaset_sm_chain * c;
	unsigned i;
	ssize_t s;
	char * buf;

	msg=cs->messages;
	if (!msg) return;
	do {
		if (msg->chain&&msg->id!=1) {
			msg=msg->next;
			continue;
		}

		c=msg->chain;
		if (!c) {
			puts ("single message:");
			print_header (msg);
		} else {
			printf ("concatenated message: %u/%u parts missing\n",
			        c->missing, (unsigned) c->msgcount);
			for (i=0;i<c->msgcount;++i)
				if(c->msg[i]) {
					printf ("part %u:\n", i);
					print_header (c->msg[i]);
				}
		}

		s=gigaset_sms2utf8 (msg, 1, &buf);
		if (s>=0) {
			puts(buf);
			free(buf);
		}
		putchar('\n');

		msg=msg->next;
	} while (msg!=cs->messages);
}

static void cmd_smslist(struct contr_state*cs, char**argv, unsigned long data)
{
	ssize_t s;
	size_t listlen;
	char * list;
	size_t pos;
	uint32_t handle;
	uint16_t flags;

	s=gigaset_smsgetlist (cs->gs, &list);
	if (s<0) return;
	listlen=s;
	printf (" %lu entries:\n", (unsigned long) (listlen/6));
	for (pos=0;pos+6<=listlen;pos+=6) {
		handle = gigaset_from_be32(list+pos);
		flags  = gigaset_from_be16(list+pos+4);
		if (!handle&&!flags) {
			printf (" end marker after %lu entries.\n", (unsigned long) (pos/6));
			break;
		}
		printf ("%-10lu %04lx\n", (unsigned long) handle, (unsigned long) flags);
	}
}

static void cmd_amlist(struct contr_state*cs, char**argv, unsigned long data)
{
	size_t listlen;
	char * list;
	size_t pos;
	uint16_t handle;
	uint8_t status;
	uint8_t amindex;
	ssize_t s;
	unsigned char *buf;

	s=gigaset_amgetlist (cs->gs, &list);
	if (s<0) {
		set_error_msg(cs, s, "getting msg list");
		return;
	}

	listlen=s;
	for (pos=0;pos+4<=listlen;pos+=4) {
		handle = gigaset_from_be16(list+pos);
		status = list[pos+2];
		amindex = list[pos+3];

		if(status<1||status>9) continue;

		printf("AB%d #%u", amindex,  (unsigned short) handle);

		s=gigaset_aminfo (cs->gs, handle, (char**)&buf);

		if (s<0) {
			fputs("\n", stdout);
			set_error_msg(cs, s, "getting msg info");
			continue;
		}
		if (s>=8) {
			printf(" %04u-%02u-%02u %02u:%02u",
				(buf[4]>=100) ? 9999 : 2000+buf[4],
				(buf[3]>12)   ? 99   : buf[3], 
				(buf[2]>31)   ? 99   : buf[2],
				(buf[1]>24)   ? 99   : buf[1],
				(buf[0]>60)   ? 99   : buf[0]);
			if(buf[6] != 0xFD) printf(" for MSN%d", buf[6]);
			if(buf[7] != 0) printf(" from '%s'", buf+7);
		}
		free (buf);

		fputs(": ", stdout);
		switch(status) {
		case 2: fputs("Incoming message (NEW)", stdout); break;
		case 3: fputs("Incoming message", stdout); break;
		case 4: fputs("Local info (NEW)", stdout); break;
		case 5: fputs("Local info", stdout); break;
		case 6: fputs("Greeting #1", stdout); break;
		case 7: fputs("Greeting #2", stdout); break;
		case 8: fputs("Info", stdout); break;
		case 9: fputs("Goodbye", stdout); break;
		default:
		case 1: fputs("Deleted or unknown", stdout);
		};

		fputs("\n", stdout);
	};
	free(list);
}

static inline FILE *open_file(const char *name, const char *mode, int *use_std, FILE *std)
{
	*use_std = name[0] == '-' && !name[1];
	return *use_std ? std : fopen(name, mode);
}

static inline int close_file(FILE *f, int use_std)
{
	return use_std ? 0 : fclose(f);
}

static ssize_t read_file(FILE *f, char **buf)
{
	size_t buf_size = 512;
	size_t cur_size = 0;
	size_t num;
	char *tmp;

	*buf = malloc(buf_size);
	if (!*buf)
		return -ERR_MEM;

	for (;;) {
		num = fread(*buf + cur_size, 1, buf_size - cur_size, f);
		if (num != buf_size - cur_size) {
			if (ferror(f)) {
				free(*buf);
				return -ERR_FILEREAD;
			}
			return cur_size + num;
		}
		cur_size = buf_size;
		if (cur_size >= 0x40000000U) { //FIXME SSIZE_MAX/2 ?
			free(*buf);
			return -ERR_LENGTH;
		}
		buf_size *= 2;
		tmp = realloc(*buf, buf_size);
		if (!tmp) {
			free(*buf);
			return -ERR_MEM;
		}
		*buf = tmp;
	}
}

static void cmd_amget(struct contr_state*cs, char**argv, unsigned long data)
{
	FILE *f;
	ssize_t s;
	size_t written;
	char *buf;
	unsigned handle;
	int to_stdout;

	if (!touintrange(argv[0], &handle, 0, 0xffff, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	f = open_file(argv[1], "wb", &to_stdout, stdout);
	if (f == NULL) {
		set_error_num_msg(cs, -ERR_FILEOPEN, errno, NULL);
		return;
	}

	s = gigaset_amgetmsg(cs->gs, handle, &buf);
	if (s < 0) {
		set_error_msg(cs, s, "getting msg audio");
	} else {
		written = fwrite(buf, 1, s, f);
		if (written != (size_t) s) {
			set_error_num_msg(cs, -ERR_FILEWRITE, errno, NULL);
		}
		free(buf);
	}

	if (close_file(f, to_stdout) != 0) {
		set_error_num_msg(cs, -ERR_FILEWRITE, errno, NULL);
	}
}

static void cmd_amput(struct contr_state*cs, char**argv, unsigned long data)
{
	FILE *f;
	ssize_t s;
	ssize_t r;
	char *buf;
	int from_stdin;

	f = open_file(argv[0], "rb", &from_stdin, stdin);
	if (f == NULL) {
		set_error_num_msg(cs, -ERR_FILEOPEN, errno, NULL);
		return;
	}

	s = read_file(f, &buf);
	if (s < 0) {
		set_error_num_msg(cs, s, errno, NULL);
		close_file(f, from_stdin);
		return;
	}

	close_file(f, from_stdin);

	r = gigaset_amputmsg(cs->gs, buf, s);
	free(buf);

	if (r < 0) {
		set_error_msg(cs, r, "putting msg audio");
		return;
	}

	cs->pflags &= ~CF_ABMODE;
}

static void cmd_amdel(struct contr_state*cs, char**argv, unsigned long data)
{
	int r;
	unsigned handle;

	if (!touintrange(argv[0], &handle, 0, 0xffff, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}

	r = gigaset_amdelmsg(cs->gs, handle);
	if (r < 0)
		set_error_msg(cs, r, "deleting msg");
}

static void cmd_amselect(struct contr_state*cs, char**argv, unsigned long data)
{
	unsigned new_amindex;

	if (!touintrange(argv[0], &new_amindex, 0, 3, 0))
		set_error_msg(cs, -ERR_PARAMINVALID, "selecting am");
	else
		cs->amindex = new_amindex;
}

static void cmd_amdelall(struct contr_state*cs, char**argv, unsigned long data)
{
	int r;
	unsigned i, m, types;

	if (!touintrange(argv[0], &types, 0, 0xffff, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	
	for(m=32768, i=16; i>0; m>>=1, i--) {
		if(types & m) {
			r = gigaset_amdeltyped(cs->gs, i);
			if(r<0) {
				set_error_msg(cs, r, "deleting all of type");
				return;
			};
		};
	};
}

static int todatatype (const char * s)
{
	if (!strcmp(s,"raw"))
		return DT_RAW;
	else if (!strcmp(s,"str"))
		return DT_STR;
	else if (!strcmp(s,"dec8"))
		return DT_DEC8;
	else if (!strcmp(s,"dec32"))
		return DT_DEC32;
	else if (!strcmp(s,"hex"))
		return DT_HEX;

	return DT_INVALID;
}

static void printdata(struct contr_state *cs, const char *buf, size_t len, int type)
{
	switch (type)
	{
	case DT_RAW:
		if (!len) break;
		if (fwrite (buf, len, 1, stdout) != len)
			set_error_num_msg(cs, -ERR_FILEWRITE, errno, NULL);
		break;
	case DT_DEC8:
		if (len!=1&&(len!=4||buf[1]!=0x3f||buf[2]!=4||buf[3])) {
			set_error_num_msg(cs, -ERR_DATA, 0, NULL);
			break;
		}
		if (printf("%u\n", (unsigned char)*buf)<0)
			set_error_num_msg(cs, -ERR_FILEWRITE, errno, NULL);
		break;
	case DT_STR:
		if (!len||buf[len-1]||strlen(buf)!=len-1) {
			set_error_num_msg(cs, -ERR_DATA, 0, NULL);
			break;
		}
		if (puts(buf)==EOF)
			set_error_num_msg(cs, -ERR_FILEWRITE, errno, NULL);
		break;
	case DT_HEX:
		if (!len) break;
		for (; len > 1; --len, ++buf)
			if (printf("%02X ", (unsigned)(unsigned char)*buf) < 0) {
				set_error_num_msg(cs, -ERR_FILEWRITE, errno, NULL);
				return;
			}
		if (printf("%02X\n", (unsigned)(unsigned char)*buf) < 0)
			set_error_num_msg(cs, -ERR_FILEWRITE, errno, NULL);
		break;
	case DT_DEC32:
		if (!len) break;
		if (len==4) 
			printf("Dec: %u, Hex: %x\n", (unsigned) gigaset_from_le32(buf), 
				(unsigned) gigaset_from_le32(buf));
		break;
	default:
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, NULL);
	}
}

static inline int fromhex(char c)
{
	return  c >= '0' && c <= '9' ? c - '0'     :
	       (c >= 'a' && c <= 'f' ? c - 'a' + 10:
	       (c >= 'A' && c <= 'F' ? c - 'A' + 10: -1));
}

static ssize_t scandata(char **buf, const char *s, int type)
{
	unsigned num;
	size_t l;
	unsigned long d;
	int digit;
	unsigned char *dest;
	int allow_prefix;

	switch (type) {
	case DT_DEC8:
		if (!touintrange(s, &num, 0, 255, 0))
			return -ERR_DATA;
		*buf = malloc(1); /* :-/ */
		if (!*buf)
			return -ERR_MEM;
		**buf = (unsigned char) num;
		return 1;
	case DT_STR:
		l = strlen(s) + 1;
		*buf = malloc(l);
		if (!*buf)
			return -ERR_MEM;
		memcpy(*buf, s, l);
		return l;
	case DT_DEC32:
		if (!toulong(s, &d, 0))
			return -ERR_DATA;
		*buf = malloc(4);
		if (!buf)
			return -ERR_MEM;
		gigaset_to_le32(*buf, d);
		return 4;
	case DT_HEX:
		l = strlen(s);
		*buf = malloc((l+1)/2);
		if (!*buf)
			return -ERR_MEM;
		dest = *buf;

		allow_prefix = 1;
		for (;;) {
			if (isspace((unsigned char)*s)) {
				allow_prefix = 1;
				++s;
			}
			while (isspace((unsigned char)*s))
				++s;
			if (!*s)
				break;
			if (allow_prefix && *s == '0' && (s[1] == 'x' || s[1] == 'X'))
				s+=2;
			digit = fromhex(*s);
			if (digit < 0) {
				free(*buf);
				return -ERR_DATA;
			}
			++s;
			*dest = digit;
			if (*s && !isspace(*s)) {
				digit = fromhex(*s);
				if (digit < 0) {
					free(*buf);
					return -ERR_DATA;
				}
				++s;
				*dest = *dest * 16 + digit;
			}
			++dest;
			allow_prefix = 0;
		}
		return dest - (unsigned char *)*buf;
	case DT_RAW:
	default:
		return -1;
	}
}

static void cmd_cfgread(struct contr_state *cs, char **argv, unsigned long data)
{
	unsigned id;
	unsigned tid;
	int32_t num;
	int datatype;
	char *s;

	if (!touintrange(argv[0], &id, 0, 0xffff, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	if (!touintrange(argv[1], &tid, 0, 99, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[1]);
		return;
	}
	datatype = todatatype(argv[2]);
	if (datatype == DT_INVALID) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[2]);
		return;
	}

	num = gigaset_confread(cs->gs, id, tid, datatype & 0xf, &s);
	if (num < 0) {
		set_error_msg(cs, num, "confread");
		return;
	}

	printdata(cs, s, num, datatype);
	free (s);
}

static void cmd_cfgwrite(struct contr_state*cs, char**argv, unsigned long data)
{
	unsigned id;
	unsigned tid;
	int r;
	int datatype;
	char * s;
	ssize_t l;

	if (!touintrange(argv[0], &id, 0, 0xffff, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[0]);
		return;
	}
	if (!touintrange(argv[1], &tid, 0, 99, 0)) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[1]);
		return;
	}
	datatype = todatatype(argv[2]);
	if (datatype == DT_INVALID) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[2]);
		return;
	}
	l = scandata(&s, argv[3], datatype);
	if (l < 0) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[3]);
		return;
	}
	if ((size_t)l > 0xffffU) {
		set_error_num_msg(cs, -ERR_PARAMINVALID, 0, argv[3]);
		free(s);
		return;
	}
	r = gigaset_confwrite(cs->gs, id, tid, datatype & 0xf, s, l);
	if (r < 0)
		set_error_msg(cs, r, "confwrite");
	free(s);
}

static void cmd_logdel(struct contr_state *cs, char **argv, unsigned long data)
{
	int r;

	r = gigaset_logdel(cs->gs);
	if (r < 0) {
		set_error_msg(cs, r, "deleting log");
		return;
	}
}

static void cmd_logstatus(struct contr_state *cs, char **argv, unsigned long data)
{
	int r;
	unsigned id, num;

	r = gigaset_logstatus(cs->gs, &id, &num);
	if (r < 0) {
		set_error_msg(cs, r, "logstatus");
		return;
	}

	printf("%u records, newest ID: %u\n", num, id);

}

/* print a log entry element decoded as operation code
 */
static struct op_type {
	int logtype;
	unsigned opcode;
	const char *opname;
} optypes[] = {
	/* answering machine */
	{  8,      4, "deleted" },
	{  8, 0x3f08, "playedMsg" },
	{  8, 0x3f0c, "playedInfo" },
	{  8, 0x3f28, "renewedMsg" },
	{  8, 0x3f2c, "renewedInfo" },
	/* SMS */
	{ 12, 0, "new" },
	{ 12, 1, "read" },
	{ 12, 2, "deleted" },
	{ 12, 3, "stored" },
	{ 12, 4, "send" },
	{ 12, 5, "stored/send" },
	{ 12, 6, "sent" },
	/* end of table */
	{  0, 0, NULL }
};

static int print_opcode(unsigned opcode, int type, FILE *out)
{
	struct op_type *pop;

	/* look up in table of known opcodes */
	for (pop = optypes; pop->opname; pop++) {
		if (pop->logtype == type && pop->opcode == opcode) {
			if (fprintf(out, "%s", pop->opname) < 0)
				return -ERR_WRITE;
			return 0;
		}
	}

	/* fallback for unknown values: numeric value */
	if (fprintf(out, "%X", opcode) < 0)
		return -ERR_WRITE;

	return 0;
}

/* print unsigned integer decomposed in hours:minutes:seconds
 */
static int print_duration(unsigned value, FILE *out)
{
	unsigned secs, mins;

	secs = value%60;
	value /= 60;
	if (value >= 60) {
		mins = value%60;
		value /= 60;
		if (value >= 24) {	/* does this ever happen? */
			if (fprintf(out, "%ud", value/24) < 0)
				return -ERR_WRITE;
			value %= 24;
		}
		if (fprintf(out, "%u:%02u:%02u", value, mins, secs) < 0)
			return -ERR_WRITE;
	} else {
		if (fprintf(out, "%u:%02u", value, secs) < 0)
			return -ERR_WRITE;
	}
	return 0;
}

/* format unsigned integer with scaling
 * result is silently truncated if bufsize insufficient
 */
static void do_scale(char *buf, size_t bufsize, unsigned value, int scale)
{
	int i, r;

	if (scale == 0) {
		/* no scaling */
		snprintf(buf, bufsize, "%d", value);
	} else if (scale > 0) {
		/* append scale zeros */
		snprintf(buf, bufsize, "%d_%.*d", value, scale, 0);
	} else /* scale < 0 */ {
		/* insert dot scale chars from end */
		r = snprintf(buf, bufsize, ".%.*d", -scale+1, value);
		r -= - scale + 1;
		if (r > 0) {
			if (r > bufsize - 1)
				r = bufsize - 1;
			for (i = 0; i < r; i++)
				buf[i] = buf[i+1];
			if (r < bufsize - 1)
				buf[r] = '.';
		}
	}
}

/* print_logentry
 * write formatted journal entry to output stream
 * Parameters:
 *	recbuf	buffer containing entry
 *	out	destination output stream
 * Return value:
 *	0: success
 *	<0: error code
 */
static const char *logentnames[] = {
	"00",	/*  0: -unused- */
	"<+",	/*  1: incoming, accepted */
	">+",	/*  2: outgoing, accepted */
	"<-",	/*  3: incoming, missed */
	">-",	/*  4: outgoing, failed */
	"AM",	/*  5: answering machine message */
	"AI",	/*  6: answering machine info */
	"AR",	/*  7: answering machine recording */
	"AC",	/*  8: answering machine state change */
	"MW",	/*  9: message waiting indication received from Centrex */
	"S<",	/* 10: incoming SM */
	"S>",	/* 11: outgoing SM */
	"SC",	/* 12: SMS state change */
};

static const char *elnames[] = {
	"00",		/*  0: generic unknown	*/
	"fromNumber",	/*  1: Caller Number	*/
	"fromName",	/*  2: Caller Name	*/
	"toNumber",	/*  3: Called Number	*/
	"toName",	/*  4: Called Name	*/
	"intNumber",	/*  5: Internal Number	*/
	"intName",	/*  6: Internal Name	*/
	"connID",	/*  7: Connection ID	*/
	"smsHandle",	/*  8: Short Message Handle	*/
	"newState",	/*  9: New Status	*/
	"amHandle",	/* 10: Answering Machine Handle	*/
	"chargeInfo",	/* 11: Charging Information	*/
	"callDuration",	/* 12: Call Duration [s]	*/
	"recDuration",	/* 13: Recording Duration [s]	*/
	"ctrlNumber",	/* 14: Controlling Number	*/
	"mgsCount",	/* 15: Message Count	*/
	"rcvNumber",	/* 16: Receiving Number	*/
	"pfxLength"	/* 17: Prefix Length	*/
};

static int print_logentry(const struct glogrec *recbuf, FILE *out)
{
	int iel, eltype, r;
	char fmtbuf[12];	/* sufficient for 2^32-1 and a dot */
	const char *s;
	const struct glogel *pel;

	/* header */
	if (recbuf->type < sizeof logentnames/sizeof logentnames[0]) {
		s = logentnames[recbuf->type];
	} else {
		snprintf(fmtbuf, sizeof fmtbuf, "%02X", recbuf->type);
		s = fmtbuf;
	}
	if (fprintf(out, "%6u  %2s  %02u.%02u.%02u %02u:%02u:%02u ",
		    recbuf->recid, s,
		    recbuf->day, recbuf->month, recbuf->year,
		    recbuf->hour, recbuf->min, recbuf->sec) < 0)
		return -ERR_WRITE;

	/* data elements */
	for (iel = 0; iel < recbuf->nels; iel++) {
		pel = &recbuf->el[iel];
		eltype = pel->type & ~ELTYPE_FORMAT;
		if (eltype < sizeof elnames/sizeof elnames[0]) {
			if (fprintf(out, " %s=", elnames[eltype]) < 0)
				return -ERR_WRITE;
		} else {
			if (fprintf(out, " %02X=", eltype) < 0)
				return -ERR_WRITE;
		}
		switch (pel->type & ELTYPE_FORMAT) {
		case ELTYPE_CBUF:
			if (fprintf(out, ":") < 0)
				return -ERR_WRITE;
			s = pel->str;
			for (r = pel->num; r > 0; r--)
				if (fprintf(out, "%02X:", (unsigned char) *s++) < 0)
					return -ERR_WRITE;
			break;
		case ELTYPE_STRING:
			if (fprintf(out, "'%s'", pel->str) < 0)
				return -ERR_WRITE;
			break;
		case ELTYPE_NUMBER:
			switch (eltype) {
			case 9:
				if (print_opcode(pel->num, recbuf->type, out) < 0)
					return -ERR_WRITE;
				break;
			case 12:
				if (print_duration(pel->num, out) < 0)
					return -ERR_WRITE;
				break;
			default:
				if (fprintf(out, "%u", pel->num) < 0)
					return -ERR_WRITE;
			}
			break;
		case ELTYPE_CHRG:
			if (pel->str) {
				/* currency */
				do_scale(fmtbuf, sizeof fmtbuf,
					 pel->num, pel->scale);
				if (fprintf(out, "'%s %s'",
					    fmtbuf, pel->str) < 0)
					return -ERR_WRITE;
			} else {
				/* units */
				if (fprintf(out, "%u", pel->num) < 0)
					return -ERR_WRITE;
			}
			break;
		default:
			fprintf(out, "bad type field %d\n", pel->type);
			return -ERR_DATA;
		}
	}
	if (fputc('\n', out) < 0)
		return -ERR_WRITE;
	return 0;
}

/* print_logheader
 * write header line for print_logentry() to output stream
 * Parameters:
 *	gs	status structure
 *	id	ID of entry to retrieve
 *	out	destination output stream
 * Return value:
 *	0: success
 *	<0: error
 */
static int print_logheader(struct gstatus *gs, FILE *out)
{
	if (fputs("\n    ID Type Date     Time      Info\n", stdout) < 0)
		return -ERR_WRITE;
	return 0;
}

static void cmd_logdump(struct contr_state *cs, char **argv, unsigned long data)
{
	int r;
	uint32_t id, lastid, numid;
	long nextid;
	char *idstr, *endid;
	struct glogrec *prec;

	/* process list of IDs */
	idstr = argv[0];
	if (*idstr == '-') {
		/* prime lastid for start with oldest existing */
		r = gigaset_logstatus(cs->gs, &lastid, &numid);
		if (r < 0) {
			set_error_msg(cs, r, "logstatus");
			return;
		}
		lastid -= numid;
	}

	print_logheader(cs->gs, stdout);

	while (*idstr) {
		/* get range of IDs to retrieve */
		if (*idstr == '.') {
			/* most recent */
			r = gigaset_logstatus(cs->gs, &id, NULL);
			if (r < 0) {
				set_error_msg(cs, r, "logstatus");
				return;
			}
			lastid = id;
			endid = idstr+1;
		} else if (*idstr == '-' && (*(idstr+1) == '.' || *(idstr+1) == 0)) {
			/* range from previous to most recent */
			id = lastid+1;
			r = gigaset_logstatus(cs->gs, &lastid, NULL);
			if (r < 0) {
				set_error_msg(cs, r, "logstatus");
				return;
			}
			endid = idstr+1;
			if (*endid) endid++;
		} else {
			/* number */
			nextid = strtol(idstr, &endid, 10);
			if (endid == idstr) {
				set_error_num_msg(cs, -ERR_PARAMINVALID, 0, idstr);
				return;
			}
			if (nextid < 0) {
				/* '-' => range */
				id = lastid+1;
				lastid = -nextid;
			} else {
				id = lastid = nextid;
			}
		}

		while (id <= lastid) {
			/* retrieve entry */
			r = gigaset_logread(cs->gs, id, &prec);
			if (r < 0) {
				if (prec)
					free(prec);
				set_error_msg(cs, r, "logread");
				return;
			}

			/* warn if wrong id field */
			if (prec->recid != id) {
				fprintf(stderr,
					"ID mismatch: requested %u, got %u\n",
					id, prec->recid);
			}

			/* format entry */
			r = print_logentry(prec, stdout);
			free(prec);
			if (r < 0) {
				set_error_msg(cs, r, "print logentry");
				return;
			}

			id++;
		}

		idstr = endid;
		if (*idstr == ',') idstr++; /* list separator */
	}
}

static void cmd_progversion(struct contr_state *cs, char **argv, unsigned long data)
{
	fprintf(stderr,
		"gigacontr: gigaset-frontend release " stringify(VERSION) "\n"
		"Copyright (C) 2011 Tilman Schmidt\n");
}

static struct cmd_def commands[];

static void cmd_usage(struct contr_state *cs, char **argv, unsigned long data)
{
	struct cmd_def *cmd;
	char line[LINE_WIDTH+1] = "Usage: gigacontr";
	char cur_string[LINE_WIDTH];
	unsigned linepos, strpos;
	unsigned i;
	size_t len;

	linepos = strlen(line);

	for (cmd = commands; cmd->namelist[0]; ++cmd) {
		if (!cmd->help_description)
			continue;
		cur_string[0] = '[';
		strpos = 1;
		for (i = 0; i < OPT_MAXNAMES && cmd->namelist[i]; ++i) {
			len = strlen(cmd->namelist[i]);
			if (strpos + len + 1 > LINE_WIDTH-USAGE_INDENT-1) {
				gigaset_log(&cs->gl, GLOG_APPERR, "error: string too long\n");
				strpos = 0;
				break;
			}
			memcpy(cur_string + strpos, cmd->namelist[i], len);
			strpos += len;
			cur_string[strpos++] = '|';
		}
		if (!strpos)
			continue;
		--strpos;
		if (cmd->help_params && *cmd->help_params) {
			cur_string[strpos++] = ' ';
			len = strlen(cmd->help_params);
			if (strpos + len > LINE_WIDTH-USAGE_INDENT-1) {
				gigaset_log(&cs->gl, GLOG_APPERR, "error: string too long\n");
				continue;
			}
			memcpy(cur_string + strpos, cmd->help_params, len);
			strpos += len;
		}
		cur_string[strpos++] = ']';

		if (linepos + 1 + strpos > LINE_WIDTH) {
			line[linepos] = 0;
			puts(line);
			memset(line, ' ', USAGE_INDENT - 1);
			linepos = USAGE_INDENT - 1;
		}
		line[linepos++] = ' ';
		memcpy(line + linepos, cur_string, strpos);
		linepos += strpos;
	}
	line[linepos] = 0;
	puts(line);
}

static void cmd_help(struct contr_state*cs, char**argv, unsigned long data)
{
	struct cmd_def *cmd;
	char line[LINE_WIDTH+1];
	unsigned linepos;
	unsigned i;
	size_t len;

	cmd_usage(cs, argv, data);

	putchar('\n');

	for (cmd = commands; cmd->namelist[0]; ++cmd) {
		if (!cmd->help_description)
			continue;
		linepos = 0;
		for (i = 0; i < OPT_MAXNAMES && cmd->namelist[i]; ++i) {
			len = strlen(cmd->namelist[i]);
			if (i) {
				line[linepos++] = ',';
				if (linepos + len + 1 + (i+1 < OPT_MAXNAMES && cmd->namelist[i+1]) > LEFT_COL_WIDTH) {
					line[linepos] = 0;
					puts(line);
					linepos = HELP_INDENT;
					memset(line, ' ', HELP_INDENT);
				} else {
					line[linepos++] = ' ';
				}
			}
			memcpy(line + linepos, cmd->namelist[i], len);
			linepos += len;
		}
		if (!cmd->help_description[0]) {
			line[linepos] = 0;
			puts(line);
			continue;
		}
		if (linepos > LEFT_COL_WIDTH) {
			line[linepos] = 0;
			puts(line);
			linepos = 0;
		}

		memset(line + linepos, ' ', LEFT_COL_WIDTH + COL_SEP - linepos);
		linepos = LEFT_COL_WIDTH + COL_SEP;

		len = strlen(cmd->help_description);
		if (linepos + len > LINE_WIDTH) {
			gigaset_log(&cs->gl, GLOG_APPERR, "error: string too long\n");
			len = LINE_WIDTH - linepos;
		}
		memcpy(line + linepos, cmd->help_description, len);
		linepos += len;
		line[linepos] = 0;
		puts(line);
	}

	putchar('\n');

	puts ("see man page for details");
}

static struct cmd_def commands[] = {
	/* names, argc, pflags, modes allowed, modes needed, ... */
	{{"--help","-h","-?"}, 0, PF_OVERWRITE|PF_ABORT, 0,0, cmd_help, 0, NULL,
	                                             "show this usage message"},
	{{"--usage"},        0, PF_OVERWRITE|PF_ABORT, 0,0, cmd_usage, 0},
	{{"--version","-V"}, 0, PF_OVERWRITE|PF_ABORT, 0,0, cmd_progversion, 0, NULL,
	                                             "show program version"},
	{{"-q","--quiet"},   0, 0, CF_ANY,  0,       cmd_verbosity,    0, NULL,
	                                             "supress error messages"},
	{{"--show-errors"},  0, 0, CF_ANY,  0,       cmd_verbosity,    1, NULL,
	                                             "reset log level to default"},
	{{"-v","--verbose"}, 0, 0, CF_ANY,  0,       cmd_verbosity,    2, NULL,
	                                             "show device i/o"},
	{{"-vv","--debug"},  0, 0, CF_ANY,  0,       cmd_verbosity,    3, NULL,
	                                             "show debug messages"},
	{{"--dev"},          1, 0, 0,       0,       cmd_dev,          0, "FILE",
	                                             "use the given device node"},
	{{"--mode"},         1, PF_LAST, CF_ANY, 0,  cmd_mode,         0, "MODE",
	                                             "set operation mode"},
	{{"--sleep"},        1, 0, CF_ANY,  0,       cmd_sleep,        0, "SECONDS",
	                                             "pause for the given time"},
	{{"--pbfile"},       2, 0, CF_PB,   CF_PB,   cmd_pbfile,       0, "FILE ID",
	                                             "send phonebook entries to phone"},
	{{"--pbent"},        3, 0, CF_PB,   CF_PB,   cmd_pbent,        0, "NAME NUMBER ID",
	                                             "send phonebook entry to phone"},
	{{"--pbdump"},       2, 0, CF_CID,  CF_CID,  cmd_pbdump,       0, "ID MODE",
	                                             "receive phonebook entries from phone" },
	{{"--pbdel"},        1, 0, CF_CID,  CF_CID,  cmd_pbdel,        0, "ID",
	                                             "delete all phonebook entries from phone" },
	{{"--smslist"},      0, 0, CF_MSG,  CF_SMS,  cmd_smslist,      0, NULL,
	                                             "list SMs"},
	{{"--smslevel"},     0, 0, CF_MSG,  CF_SMS,  cmd_smslevel,     0, NULL,
	                                             "show SMS memory usage"},
	{{"--smsreset"},     0, 0, CF_SMS,  CF_SMS,  cmd_smsreset,     0, NULL,
	                                             "delete all SMs"},
	{{"--smsdump"},      0, 0, CF_MSG,  CF_SMS,  cmd_smsdump,      0, NULL,
	                                             "dump all messages/message parts"},
	{{"--smsdumpext"},   0, 0, CF_MSG,  CF_MSG,  cmd_smsdumpext,   0, NULL,
	                                             "dump all SMs"},
	{{"--smsselect"},    1, 0, CF_MSG,  CF_MSG,  cmd_smsselect,    0, "ID",
	                                             "select SM"},
	{{"--smsselectmsg"}, 1, 0, CF_MSG,  CF_MSG,  cmd_smsselectmsg, 0, "ID",
	                                             "select SM or concatenated message"},
	{{"--smshandle"},    1, 0, CF_MSG,  CF_SMS,  cmd_smshandle,    0, "ID",
	                                             "select SM (low level)"},
	{{"--smsget"},       0, 0, CF_MSG,  CF_SMS,  cmd_smsget,       0, NULL,
	                                             "dump selected message/message part"},
	{{"--smscreate"},    8, 0, CF_MSG,  CF_SMS,  cmd_smscreate,    0, "FLAGS UNKNOWN DCS VP CHAIN NUMBER FILE CRFLAGS",
	                                             "create SM"},
	{{"--smsnumber"},    1, 0, CF_MSG,  CF_SMS,  cmd_smsnumber,    0, "NUMBER",
	                                             "set number of SM"},
	{{"--smsstore"},     0, 0, CF_MSG,  CF_SMS,  cmd_smsfunction,  (unsigned long)&gigaset_smsstore, NULL,
	                                             "set store bit"},
	{{"--smsold"},       0, 0, CF_MSG,  CF_SMS,  cmd_smsfunction,  (unsigned long)&gigaset_smsold, NULL,
	                                             "set old bit"},
	{{"--smssend"},      0, 0, CF_MSG,  CF_SMS,  cmd_smsfunction,  (unsigned long)&gigaset_smssend, NULL,
	                                             "set send bit"},
	{{"--smsdel"},       0, 0, CF_MSG,  CF_SMS,  cmd_smsfunction,  (unsigned long)&gigaset_smsdel, NULL,
	                                             "set delete bit"},
	{{"--amlist"},       0, 0, CF_AB,   CF_AB,   cmd_amlist,       0, NULL,
	                                             "list all audio messages in phone memory"},
	{{"--amget"},        2, 0, CF_AB,   CF_AB,   cmd_amget,        0, "ID FILE",
	                                             "archive message in file"},
	{{"--amput"},        1, 0, CF_AB,   CF_AB,   cmd_amput,        0, "FILE",
	                                             "transfer message from file to phone memory"},
	{{"--amdel"},        1, 0, CF_AB,   CF_AB,   cmd_amdel,        0, "ID",
	                                             "delete single message"},
	{{"--amdelall"},     1, 0, CF_AB,   CF_AB,   cmd_amdelall,     0, "TYPES",
	                                             "delete several messages"},
	{{"--amselect"},     1, 0, CF_CID,  CF_CID,  cmd_amselect,     0, "ID",
	                                             "select answering machine 1, 2 or 3"},
	{{"--cfgread"},      3, 0, CF_CID,  CF_CID,  cmd_cfgread,      0, "PID ID TYPE",
	                                             "read configuration entry"},
	{{"--cfgwrite"},     4, 0, CF_LCK,  CF_LCK,  cmd_cfgwrite,     0, "PID ID TYPE VALUE",
	                                             "write configuration entry"},
	{{"--dial"},         3, 0, CF_CID,  CF_CID,  cmd_dial,         0, "TYPE NUMBER ID",
	                                             "make phone dial"},
	{{"--logdel"},       0, 0, CF_CID,  CF_CID,  cmd_logdel,       0, NULL,
	                                             "flush journal log buffer (delete all entries)"},
	{{"--logstatus"},    0, 0, CF_CID,  CF_CID,  cmd_logstatus,    0, NULL,
	                                             "get journal log status"},
	{{"--logdump"},      1, 0, CF_CID,  CF_CID,  cmd_logdump,      0, "IDLIST",
	                                             "retrieve and display journal log entries"},
	{{"--msnget"},       1, 0, CF_CID,  CF_CID,  cmd_get_str_idx,  (unsigned long)&gigaset_msnget, "INDEX",
	                                             "get MSN"},
	{{"--msnset"},       2, 0, CF_CID,  CF_CID,  cmd_set_str_idx,  (unsigned long)&gigaset_msnset, "INDEX NUMBER",
	                                             "set MSN"},
	{{"--cfxset"},       5, 0, CF_LCK,  CF_LCK,  cmd_set_cfx,      0, "MSN MODE PROC SERVICE TO-MSN",
	                                             "set MSN"},
	{{"--vendor"},       0, 0, CF_UNI,  CF_UNI,  cmd_get_string,   (unsigned long)&gigaset_vendor, NULL,
	                                             "get vendor string"},
	{{"--product"},      0, 0, CF_UNI,  CF_UNI,  cmd_get_string,   (unsigned long)&gigaset_product, NULL,
	                                             "get product name"},
	{{"--fwver"},        0, 0, CF_UNI,  CF_UNI,  cmd_get_string,   (unsigned long)&gigaset_version, NULL,
	                                             "get firmware version"},
	{{"--device"},       0, 0, CF_CID,  CF_CID,  cmd_get_string,   (unsigned long)&gigaset_device, NULL,
	                                             "get device name"},
	{{"--mvendor"},      1, 0, CF_CFG,  CF_CFG,  cmd_get_str_idx,  (unsigned long)&m10x_vendor, "ADAPTER",
	                                             "get vendor string"},
	{{"--mfwver"},       1, 0, CF_CFG,  CF_CFG,  cmd_get_str_idx,  (unsigned long)&m10x_version, "ADAPTER",
	                                             "get firmware version"},
	{{"--mhwver"},       1, 0, CF_CFG,  CF_CFG,  cmd_get_str_idx,  (unsigned long)&m10x_versionhw, "ADAPTER",
	                                             "get hardware version"},
	{{"--mdevid"},       1, 0, CF_CFG,  CF_CFG,  cmd_get_str_idx,  (unsigned long)&m10x_devid, "ADAPTER",
	                                             "get device id"},
	{{"--mgetname"},     1, 0, CF_CFG,  CF_CFG,  cmd_get_str_idx,  (unsigned long)&m10x_getname, "ADAPTER",
	                                             "get adapter name"},
	{{"--mconnstate"},   0, 0, CF_CFG,  CF_CFG,  cmd_get_string,   (unsigned long)&m10x_connstate, NULL,
	                                             "get connection state"},
	{{"--moccupy"},      0, 0, CF_CFG,  CF_CFG,  cmd_moccupyconn,  0, NULL,
	                                             "occupy cordless connection"},
	{{"--mrelease"},     0, 0, CF_CFG,  CF_CFG,  cmd_mreleaseconn, 0, NULL,
	                                             "release cordless connection"},
	{{"--mfield"},       0, 0, CF_CFG,  CF_CFG,  cmd_get_string,   (unsigned long)&m10x_field, NULL,
	                                             "get signal strength"},
	{{"--mquality"},     0, 0, CF_CFG,  CF_CFG,  cmd_get_string,   (unsigned long)&m10x_quality, NULL,
	                                             "get signal quality"},
	{{"--mgetautorel"},  1, 0, CF_CFG,  CF_CFG,  cmd_intlist,      (unsigned long)&m10x_listautorel, "ADAPTER",
	                                             "get auto release timers"},
	{{"--msetautorel"},  1, 0, CF_CFG,  CF_CFG,  cmd_msetautorel,  0, "SECONDS",
	                                             "set auto release timer"},
	{{"--mreset"},       0, 0, CF_CFG,  CF_CFG,  cmd_mreset,       0, NULL,
	                                             "reset adapter configuration"},
	{{"--mgettype"},     0, 0, CF_MGET, CF_MGET, cmd_m10x_getvar,  MVAR_TYPE, NULL,
	                                             "get adapter type"},
	{{"--msettype"},     1, 0, CF_CFG,  CF_CFG,  cmd_msettype,     0, "TYPE",
	                                             "set adapter type"},
	{{"--mlisten"},      0, 0, CF_CFG,  CF_CFG,  cmd_mlisten,      0, NULL,
	                                             "wait for PP to register"},
	{{"--mregister"},    2, 0, CF_CFG,  CF_CFG,  cmd_mregister,    0, "PIN INDEX",
	                                             "register to FP"},
	{{"--mlistnames"},   1, 0, CF_CFG,  CF_CFG,  cmd_mlistnames, 0, "ADAPTER",
	                                             "get list of adapter names"},
	{{"--mlistconn"},    1, 0, CF_CFG,  CF_CFG,  cmd_intlist,      (unsigned long)&m10x_listconns, "ADAPTER",
	                                             "get associated connections"},
	{{"--mgetpartner"},  0, 0, CF_MGET, CF_MGET, cmd_m10x_getvar,  MVAR_PARTNER, NULL,
	                                             "get index of partner"},
	{{"--msetpartner"},  1, 0, CF_CFG,  CF_CFG,  cmd_msetpartner,     0, "INDEX",
	                                             "set active partner"},
	{{"--mgetmode"},     0, 0, CF_MGET, CF_MGET, cmd_m10x_getvar,  MVAR_MODE, NULL,
	                                             "get operation mode"},
	{{"--msetmode"},     1, 0, CF_MSET, CF_MSET, cmd_m10x_setvar,  MVAR_MODE, "INDEX",
	                                             "set operation mode"},
	{{"--mgetrate"},     0, 0, CF_MGET, CF_MGET, cmd_m10x_getvar,  MVAR_RATE, NULL,
	                                             "get baud rate"},
	{{"--msetrate"},     1, 0, CF_MSET, CF_MSET, cmd_m10x_setvar,  MVAR_RATE, "INDEX",
	                                             "set baud rate"},
	{{"--mgetformat"},   0, 0, CF_MGET, CF_MGET, cmd_m10x_getvar,  MVAR_FORMAT, NULL,
	                                             "get data format"},
	{{"--msetformat"},   1, 0, CF_MSET, CF_MSET, cmd_m10x_setvar,  MVAR_FORMAT, "INDEX",
	                                             "set data format"},
	{{"--mgethshake"},   0, 0, CF_MGET, CF_MGET, cmd_m10x_getvar,  MVAR_HSHAKE, NULL,
	                                             "get handshake setting"},
	{{"--msethshake"},   1, 0, CF_MSET, CF_MSET, cmd_m10x_setvar,  MVAR_HSHAKE, "INDEX",
	                                             "set local handshake"},
	{{"--mgetflow"},     0, 0, CF_MGET, CF_MGET, cmd_m10x_getvar,  MVAR_FLOW, NULL,
	                                             "get flow control setting"},
	{{"--msetflow"},     1, 0, CF_MSET, CF_MSET, cmd_m10x_setvar,  MVAR_FLOW, "INDEX",
	                                             "set flow control"},
	{{"--bchars"},       1, 0, CF_CID,  CF_STD,  cmd_bchars,       0, "BRKCHARS",
                                                     "set break characters"},
	{{NULL}}
};


static int scanparams (struct param_info ** p, int argc, char ** argv, unsigned *pnum)
{
	int k;
	const char * const * name;
	int found;
	unsigned num=0;
	int i=0;
	int j;
	char * c;
	int single=0;
	const struct cmd_def * cmd;
	struct param_info * pp, *cp;
	int ret;

	*p=NULL;
	pp=malloc(argc*sizeof *pp);
	if (!pp) return -ERR_MEM;

	cp=pp;
	for(i=1;i<argc;++i,++cp) {
		++num;
		c=argv[i];
		if(!c) goto error;
		found=0;
		for(cmd=commands;*(name=cmd->namelist);++cmd) {
			for(k=0; k<OPT_MAXNAMES && *name; ++k,++name) {
				if(!strcmp(c,*name)) {
					if(cmd->parflags&PF_OVERWRITE) {
						num=1;
						cp=pp;
					}
					if (cmd->parflags&PF_SINGLE) single=1;
					cp->param=argv+i+1;
					cp->cmd=cmd;
					for (j=i+1;j<=i+cmd->numpara;++j)
						if (j>=argc||!argv[j]) {
							ret=-ERR_PARAM_NUM;
							goto error;
						}
					if ((cmd->parflags&PF_LAST)&&i+cmd->numpara!=argc-1) {
						ret=-ERR_PARAM_LAST;
						goto error;
					}
					i+=cmd->numpara;
					found=1;
					break;
				}
			}
			if (found) break;
		}
		if(!*name) {
			ret=-ERR_COM_INVALID;
			goto error;
		}
		if(cmd->parflags&PF_ABORT) {
			++cp;
			break;
		}
	}
#if 0
	if (!num) {
		ret = -ERR_NO_ARG;
		goto error;
	}
#endif
	if (single && num != 1) {
		ret = -ERR_ONE_ARG;
		goto error;
	}
	cp->cmd=NULL;
	*p=pp;
	*pnum=num;
	return 0;

error:
	free(pp);
	*pnum=i;
	return ret;
}

static const char * get_interface (const char ** interface)
{
	if (!interface)
		return NULL;
	if (!*interface)
		*interface=gigaset_default_interface();

	return *interface;
}

/* clear a state flag, processing queued operations for that state */
static int clearflag (struct contr_state * cs, unsigned flag)
{
	struct gphonebook * pbcur;
	int r;//FIXME

	switch (flag)
	{
	case CF_OPEN:
		gigaset_close(cs->gs);
		cs->gs = NULL;
		break;
	case CF_LOCK:
		if (cs->abort) gigaset_usleep_unint(2000000); /*FIXME*/
		r=gigaset_clear (cs->gs);
		if (r<0)
			set_error_msg(cs, r, "clearing CF_LOCK");
		else
			gigaset_log (&cs->gl, GLOG_APPDBG, "%d strings deleted\n", r);
		/*sleep (1);*/
#if 0
		r=gigaset_reset(gs);
		if (r<0)
			set_error_msg(cs, r, "resetting");
		r=gigaset_clear (gs);
		if (r<0)
			set_error_msg(cs, r, "clearing");
		else
			gigaset_log (&cs->gl, GLOG_APPERR, "%d strings deleted\n", r);
#endif
		r=gigaset_lock(cs->gs,0);
		if (r<0)
			set_error_msg(cs, r, "unlocking");
		break;
	case CF_CIDMODE:
		r=gigaset_unimodem(cs->gs);
		if (r<0)
			set_error_msg(cs, r, "entering unimodem mode");
		break;
	case CF_CONF:
		//FIXME
		r=gigaset_clear (cs->gs);
		if (r<0)
			set_error_msg(cs, r, "clearing CF_CONF");
		else
			gigaset_log (&cs->gl, GLOG_APPDBG, "%d strings deleted\n", r);
		r=gigaset_quitconfig(cs->gs);
		if (r<0)
			set_error_msg(cs, r, "leaving config mode");
		else
			gigaset_usleep_unint(1000000); /* we must wait for a while, until commands can be sent */
			/* FIXME genaue Zeit?*/
		break;
	case CF_PHONEBOOK:
		if (!cs->abort) {
			r = gigaset_sendpbarray(cs->gs, cs->pbfirst);
			if (r < 0)
				set_error_msg(cs, r, "sending phonebook entries");
		}
		while(cs->pbfirst)
		{
			pbcur=cs->pbfirst;
			cs->pbfirst=cs->pbfirst->next;
			free (pbcur->buffer);
			free (pbcur);
		}
		cs->pblast=NULL;
		break;
	case CF_MESSAGES:
		if (cs->selected_chain) {
			if (cs->selected_chain!=cs->send_chain)
				cs->selected_chain=NULL;
		} else if (cs->selected_message) {
			if (cs->selected_message!=cs->send_messages&&
			    cs->selected_message->chain!=cs->send_chain)
				cs->selected_message=NULL;
		}
		gigaset_smsfreemessages (&cs->messages, &cs->chains);
		break;
	case CF_SMSMODE:
		gigaset_smsfreemessages (&cs->send_messages, &cs->send_chain);
		cs->handle_selected=0;
		cs->handle_set=0;
		//cs->haveselection=0;
		//free(cs->list);
		//cs->list=NULL;
		//r=gigaset_clear (cs->gs);
		//if (r<0)
		//	set_error_msg(cs, r, "clearing CF_SMSMODE");
		//else
		//	gigaset_log (&cs->gl, GLOG_APPDBG, "%d strings deleted\n", r);
		//if (!(cs->gs.watchstate&0x20))
		//{
			r=gigaset_smsend(cs->gs);
			if (r<0)
				set_error_msg(cs, r, "leaving SMS mode");
		//}
		break;
	case CF_ABMODE:
		r=gigaset_amend(cs->gs);
		if (r < 0)
			set_error_msg(cs, r, "leaving AB mode");
		break;
	case CF_CFGLOCK:
		gigaset_usleep_unint(1000000);
		r=gigaset_clear (cs->gs);
		if (r<0)
			set_error_msg(cs, r, "clearing CF_CFGLOCK");
		else
			gigaset_log (&cs->gl, GLOG_APPDBG, "%d strings deleted\n", r);

		r=gigaset_confunlock(cs->gs);
		if (r < 0)
			set_error_msg(cs, r, "unlocking cfg");
		break;
	case CF_M10XSET:
		r=m10x_setstate (cs->gs, &cs->mstate);
		if (r < 0)
			set_error_msg(cs, r, "setting M10x state");
		break;
	}
	cs->pflags&=~flag;
	return 0;
}

static int setflag (struct contr_state * cs, unsigned flag)
{
	int r, gig_errno;//FIXME
	const char * path;
	int checkflags=0;
	int retry;

	switch (flag)
	{
	case CF_OPEN:
		path=get_interface (&cs->interface);
		if (path)
		{
			cs->gs = gigaset_open(&r, &gig_errno, path, &cs->gl, cs->sigint);
			if (!cs->gs)
				set_error_num_msg(cs, -ERR_OPEN, gig_errno, path);
		}
		else
			set_error_num_msg(cs, -ERR_UNIMPLEMENTED, 0, "internal");
		break;
	case CF_LOCK:

#if 0
		r=gigaset_config(gs,2);
		if (r < 0) {
			set_error_msg(cs, r, "querying config state");
			break;
		}
		gigaset_log(&cs->gl, GLOG_APPDBG, "config state: %d\n", r);

		if (r)
		{
			pflags|=CF_CONF;
			checkflags=1;
		}

		r=gigaset_lock(gs,-1);
		if (r < 0) {
			set_error_msg(cs, r, "querying lock state");
			break;
		}
		gigaset_log(&cs->gl, GLOG_APPDBG, "lock state: %d\n", r);
#endif
		//gigaset_lock(gs,0); // FIXME only when lock fails and called with some --force option
		retry = 0;
		r=gigaset_lock(cs->gs,1);
		if (r < 0) {
			set_error_msg(cs, r, "setting lock");
			break;
		}

		cs->pflags&=~(CF_CONF|CF_CIDMODE|CF_UNIMODEM);
		switch (gigaset_curmode(cs->gs)) {
		case GMODE_CONFIG:
			r=gigaset_initconfig(cs->gs);
			if (r < 0) {
				set_error_msg(cs, r, "getting config prefix");
				break;
			}
			cs->pflags|=CF_CONF;
			break;
		case GMODE_CID:
			/* not always reliable - set again to be sure */
			/* --v-- fall through --v-- */
		case GMODE_UNIMODEM:
			cs->pflags|=CF_UNIMODEM;
			break;
		}
		if (cs->abort)
			break;

		checkflags=1;
		break;
	case CF_UNIMODEM:
		r=gigaset_unimodem(cs->gs);
		if (r<0)
			set_error_msg(cs, r, "entering unimodem mode");
		break;
	case CF_CIDMODE:
		r=gigaset_cidmode(cs->gs);
		if (r<0)
			set_error_msg(cs, r, "entering CID mode");
		break;
	case CF_CONF:
		r=gigaset_initconfig(cs->gs);
		if (r<0)
			set_error_msg(cs, r, "entering config mode");
		break;
	case CF_SMSMODE:
		//cs->gs.watchstate&=~0x38UL;
		cs->send_messages=NULL;
		cs->send_chain=NULL;
		cs->messages=NULL;
		cs->chains=NULL;
		cs->handle_set=0;
		cs->handle_selected=0;
		cs->selected_message=NULL;
		cs->selected_chain=NULL;
		r=gigaset_smsstart(cs->gs);
		if (r < 0)
			set_error_msg(cs, r, "preparing SMS mode");
		break;
	case CF_ABMODE:
		r=gigaset_amstart(cs->gs, cs->amindex);
		if (r < 0)
			set_error_msg(cs, r, "preparing AB mode");
		break;
	case CF_MESSAGES:
		r=gigaset_smscontinue(cs->gs);
		if (r<0) {
			set_error_msg(cs, r, "entering SMS mode");
			break;
		}
		gigaset_log (&cs->gl, GLOG_APPDBG, "--> get messages\n");
		r=gigaset_smsgetmessages (cs->gs, &cs->messages, &cs->chains);
		if (r<0)
			set_error_msg(cs, r, "getting messages");
		break;
	case CF_M10XSET:
		cs->mstate.mode=-1;
		cs->mstate.baudrate=-1;
		cs->mstate.format=-1;
		cs->mstate.handshake=-1;
		cs->mstate.flow=-1;
		cs->mstate.type=-1;
		cs->mstate.partner=-1;
		break;
	case CF_M10XGET:
		r=m10x_getstate (cs->gs, &cs->mstate);
		if (r < 0)
			set_error_msg(cs, r, "getting M10x state");
		break;
	case CF_CFGLOCK:
		cs->abort=1;
		r=gigaset_confinitlock(cs->gs);
		if (r < 0)
			set_error_msg(cs, r, "locking cfg");
		else
			cs->abort = 0;
		break;
	}
	if (!cs->abort)
		cs->pflags|=flag;
	return checkflags;
}


static int keepflag (struct contr_state * cs, unsigned flag)
{
	int r;//FIXME

	switch (flag)
	{
	case CF_LOCK:
		r = gigaset_clear(cs->gs);
		if (r < 0)
			set_error_msg(cs, r, "clearing CF_LOCK");
		break;
	case CF_SMSMODE:
#if 0
		if (cs->gs.watchstate&0x18)
		{
			gigaset_log(&cs->gl, GLOG_APPERR, "sms abort (state 0x%02x)\n", (unsigned)cs->gs.watchstate&0x18);
			cs->abort=1;
			break;
		}
#endif
		r=gigaset_smscontinue(cs->gs);
		if (r<0)
			set_error_msg(cs, r, "maintaining SMS mode");
		break;
	case CF_ABMODE:
		r=gigaset_amcontinue(cs->gs);
		if (r<0)
			set_error_msg(cs, r, "maintaining AB mode");
		break;
	case CF_CFGLOCK:
		cs->abort=1;
		r=gigaset_confkeeplock(cs->gs);
		if (r<0) break;
		cs->abort=0;
		break;
	}
	return 0;
}

static void handleflags (struct contr_state * cs, unsigned flags_allow, unsigned flags_need, int leave)
{
	unsigned cflags,iflags;
	int testflags;
	unsigned flags;

	do {
		testflags=0;
		flags=(cs->pflags&flags_allow)|flags_need;

		/* find flags to clear */
		cflags=cs->pflags&~flags;
		for (iflags=CF_MAX;iflags;iflags>>=1)
			if (iflags&cflags) {
				testflags=clearflag (cs, iflags);
				if (testflags) break;
			}

		if (testflags) continue;
		if (leave||cs->abort) break;

		/* find flags to set */
		cflags=flags&~cs->pflags;
		for (iflags=1;iflags<=CF_MAX;iflags<<=1)
			if (iflags&cflags) {
				testflags=setflag (cs, iflags);
				if (cs->abort||testflags) break;
			}

		if (cs->abort) break;
		if (testflags) continue;

		/* find active flags */
		cflags=cs->pflags;
		for (iflags=1;iflags<=CF_MAX;iflags<<=1)
			if (iflags&cflags) {
				testflags=keepflag(cs, iflags);
				if (cs->abort||testflags) break;
			}

		if (cs->abort) break;
	} while (testflags);
}

const char *errorstr(int error)
{
	if (error >= ERRSTR_MIN && error <= ERRSTR_MAX)
		return errstr[error - ERRSTR_MIN];

	return NULL;
}

volatile static sig_atomic_t sigint = 0;

static void sigint_handler(int sig)
{
	signal(SIGINT, sigint_handler);
	sigint = 1;
}

int main (int argc, char ** argv)
		/*FIXME which TIMEOUT values?
		  FIXME error not set at many places!
		  FIXME retvals of printf/puts not checked*/
{
	int r;
	struct param_info * p, * pp;
	struct contr_state cs;
	unsigned pnum;

	if (signal(SIGINT, sigint_handler) == SIG_ERR) {
		perror ("signal");
		return EXIT_FAILURE;
	}

	/* initialize cs structure */
	cs.gl.verb = GLOG_APPERR | GLOG_ERR | GLOG_WARN;
	cs.gl.log = stderr;
	cs.gl.logfunc = NULL;
	cs.gl.data = NULL;
	cs.gl.errorstr = errorstr;

	cs.gs = NULL;
	cs.abort=0;
	cs.pflags=0;
	cs.interface=NULL;
	cs.pbfirst=cs.pblast=NULL;
	//cs.list=NULL;
	cs.error=0;
	cs.exit_flags=0;
	cs.messages=NULL;
	cs.chains=NULL;
	cs.handle_selected=0;
	cs.handle_set=0;
	cs.selected_chain=NULL;
	cs.selected_message=NULL;
	cs.chains=NULL;
	cs.messages=NULL;
	cs.send_messages=NULL;
	cs.send_chain=NULL;
	cs.amindex=0;
	cs.sigint = &sigint;

	r=scanparams (&p, argc, argv, &pnum);
	if (r<0) {
		cs.error=r;/*FIXME*/
		if (IS_PARAM_ERROR(-r))
			gigaset_logerr(&cs.gl, GLOG_APPERR, -r, 0, argv[pnum]);
		else
			gigaset_logerr(&cs.gl, GLOG_APPERR, -r, 0, NULL);
		p=NULL;
	} else if (pnum==0) {
		gigaset_log(&cs.gl, GLOG_APPERR, "You should enter at least one option. Try 'gigacontr --usage' for valid options\n");
		cs.error=-1;/*FIXME*/
	} else {
		for (pp=p;pp->cmd;++pp) {
			handleflags (&cs, pp->cmd->flags_allow, pp->cmd->flags_need, 0);
			if (cs.abort) break;
			pp->cmd->func(&cs, pp->param, pp->cmd->data);
			if (cs.abort) break;
		}
		if (cs.exit_flags) {
			cs.abort=0;
			handleflags (&cs, cs.exit_flags|CF_OPEN|CF_LOCK, cs.exit_flags|CF_OPEN|CF_LOCK, 0);
			cs.abort=0;
			handleflags (&cs, cs.exit_flags, cs.exit_flags, 1);
		} else {
			cs.abort=0;
			handleflags (&cs, CF_CIDMODE|CF_UNIMODEM, 0, 1);
		}

	}
	free(p);

	return cs.error?EXIT_FAILURE:EXIT_SUCCESS; //FIXME
}
