/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#ifndef GIGASET_DATA_H
#define GIGASET_DATA_H

#include <stdint.h>
#include "interf.h"

size_t gigaset_unpackseptetbuf(char * dest, const char *src, size_t septets, size_t skipoctets);
size_t gigaset_packseptetbuf(char * dest, const char *src, size_t octets, size_t skipoctets);
ssize_t gigaset_stuffbuf(char * dest, const char *src,  size_t l, char esc, char add, size_t bufsize);
ssize_t gigaset_unstuffbuf(char * dest, const char *src,  size_t l, char esc, size_t bufsize);
ssize_t gigaset_bytestuff(char ** dest, const char * src, size_t l, char esc, char add, size_t headersize, size_t tail);
ssize_t gigaset_unstuff(char ** dest, const char * src, size_t l, char esc);
uint16_t gigaset_checksum(const char * buf, size_t l);
void gigaset_hexchecksum (char * dest, const char * buf, size_t l);
int gigaset_verifychecksum(const char * buf, size_t l, uint16_t CRC);
void gigaset_tohex4 (char * dest, uint16_t val);
void gigaset_todec2 (char * dest, unsigned char val);
int32_t gigaset_fromhex4 (const char * src);
int16_t gigaset_fromhex2 (const char * src);
int8_t gigaset_fromhex1 (char src);
int gigaset_32fromdec (uint32_t * dest, const char * src); /* 0 terminated */
void gigaset_32todec (char*dest, uint32_t val); /* 0 terminated */
size_t gigaset_stuffedsize(const char *src, size_t l, char esc, char add);
ssize_t gigaset_unstuffedsize(const char *src, size_t l, char esc);


static inline uint16_t gigaset_from_be16 (const unsigned char * buf)
{
	return (uint8_t)buf[1]|(uint16_t)buf[0] << 8;
}

static inline uint32_t gigaset_from_be32 (const unsigned char * buf)
{
	return (uint8_t)buf[3]
		|(uint32_t)buf[2] << 8
		|(uint32_t)buf[1] << 16
		|(uint32_t)buf[0] << 24;
}

static inline void gigaset_to_be16 (char * buf, uint16_t v)
{
	buf[0]=v>>8;
	buf[1]=v&0xff;
}

static inline void gigaset_to_be32 (char * buf, uint32_t v)
{
	buf[3]=v&0xff;v>>=8;
	buf[2]=v&0xff;v>>=8;
	buf[1]=v&0xff;v>>=8;
	buf[0]=v;
}

static inline uint16_t gigaset_from_le16 (const unsigned char * buf)
{
	return (uint8_t)buf[0]|(uint16_t)buf[1] << 8;
}

static inline uint32_t gigaset_from_le32 (const unsigned char * buf)
{
	return (uint8_t)buf[0]
		|(uint32_t)buf[1] << 8
		|(uint32_t)buf[2] << 16
		|(uint32_t)buf[3] << 24;
}

static inline void gigaset_to_le16 (unsigned char * buf, uint16_t v)
{
	buf[1]=v>>8;
	buf[0]=v&0xff;
}

static inline void gigaset_to_le32 (unsigned char * buf, uint32_t v)
{
	buf[0]=v&0xff;v>>=8;
	buf[1]=v&0xff;v>>=8;
	buf[2]=v&0xff;v>>=8;
	buf[3]=v;
}

static inline unsigned gigaset_bcd8toi(unsigned char bcd)
{
	return ((unsigned) bcd & 0xf) +
	       ((unsigned) bcd >> 4 & 0xf) * 10;
}

static inline unsigned char gigaset_itobcd8 (int bcd)
{
	if (bcd<0 || bcd>99)
		return '\0';
	return ((unsigned char)(bcd/10) << 4) + (unsigned char)(bcd%10);
}

static inline int gigaset_fromdec4(const char *src, unsigned *ret)
{
	if (src[0] < '0' || src [0] > '9' ||
	    src[1] < '0' || src [1] > '9' ||
	    src[2] < '0' || src [2] > '9' ||
	    src[3] < '0' || src [3] > '9')
		return 0;
	*ret = 10 * (10 * (10 * (
		(unsigned)(src[0] - '0')) +
		(unsigned)(src[1] - '0')) +
		(unsigned)(src[2] - '0')) +
		(unsigned)(src[3] - '0');
	return 1;
}

static inline int gigaset_fromdec2(const char *src, unsigned *ret)
{
	if (src[0] < '0' || src [0] > '9' ||
	    src[1] < '0' || src [1] > '9')
		return 0;
	*ret = 10 * (unsigned)(src[0] - '0') + (unsigned)(src[1] - '0');
	return 1;
}

static inline int gigaset_fromdec1(char c, unsigned *ret)
{
	if (c < '0' || c > '9')
		return 0;
	*ret = c - '0';
	return 1;
}

#define GIG_SMS_ESC_DEPTH 2
#define GIG_UTF8_MAX_BYTES 3

unsigned char * gigaset_utf8_to_septets (const unsigned char * src, size_t srclen, size_t * ressize);
ssize_t gigaset_utf8_to_septetbuf (const unsigned char * src, size_t srclen, unsigned char * dest, size_t destlen,
                                   unsigned * esccount, char * escapedchar, size_t * septets);
unsigned char * gigaset_septets_to_utf8 (const unsigned char * src, size_t srclen, size_t * ressize);
size_t gigaset_septets_to_utf8buf (const unsigned char * src, size_t srclen, unsigned char * dest, size_t destlen, unsigned * esc);

size_t gigaset_UCS2_to_utf8buf(const unsigned char *src, size_t srclen, unsigned char *dest, size_t destlen, unsigned *incomplete, unsigned char *prev);

int gigaset_valid_sms_char(uint16_t c_ucs2);

#endif //GIGASET_DATA_H
