/*
 * This part of the configuration program handles the "low level"
 * communication with the device driver.
 *
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#ifndef GIGASET_INTERF_H
#define GIGASET_INTERF_H

#define ERR_READ                1
#define ERR_WRITE               2
#define ERR_OPEN                3
#define ERR_CLOSE               4
#define ERR_IO                  5
#define ERR_MEM                 6
#define ERR_NORESPONSE          7
#define ERR_RESPINVALID         8
#define ERR_PARAMINVALID        9
#define ERR_CONNECT             10
#define ERR_LOCK                11
#define ERR_MSGUNEXPECTED       12
#define ERR_CONFIG              13
#define ERR_LENGTH              14
#define ERR_DATA                15
#define ERR_ABORT               16
#define ERR_CRC                 17
#define ERR_PROTO               18
#define ERR_TIMEOUT             19
#define ERR_NOENT               20
#define ERR_VERSION             21
#define ERR_UNIMPLEMENTED       22
#define ERR_IO_UNKNOWN          23 /* no errno */
#define ERR_SIGINT              24
#define ERR_MODE                25

#define ERR_LIBMAX              63
#define ERR_MAX                 255

#define CID_MAX    65535
#define CID_MAXLEN 5

#define ERR_WITH_ERRNO(n) ((n)>=ERR_READ&&(n)<=ERR_IO)
#define GET_ERRNO(gs,n) (ERR_WITH_ERRNO(n)?(gs)->error:0)

#include <stdlib.h>
#include <stdint.h>
#include <limits.h>
#include <stdio.h>
#include <signal.h>

#ifdef __GNUC__
#define attribute_format(type,fstr,first) __attribute__((format(type,fstr,first)))
#else
#define attribute_format(type,fstr,first)
#endif

#if INT_MAX < 2147483647L || CHAR_BIT != 8
#error "invalid integer types"
#endif

#ifdef _WIN32
#include <basetsd.h>
typedef SSIZE_T ssize_t;
#ifndef SSIZE_MAX
#define SSIZE_MAX ((1UL<<(sizeof(ssize_t)*8-1))-1)
#endif
#endif

#define GIG_INFTY INT_MAX

#define GLOG_ERR      0x0001
#define GLOG_WARN     0x0002
#define GLOG_INFO     0x0004
#define GLOG_STATUS   0x0008
#define GLOG_IO       0x0010
#define GLOG_DEBUG    0x0020
#define GLOG_DUMP     0x0040

#define GLOG_LIBBITS  16

#define GMODE_UNKNOWN  0
#define GMODE_CONFIG   1
#define GMODE_UNIMODEM 2
#define GMODE_CID      3

#define GIG_TYPE_INTERNAL 0
#define GIG_TYPE_EXTERNAL 1
#define GIG_TYPE_RECALL   2
#define GIG_TYPE_DOOR     3
#define GIG_TYPE_CENTREX  4

struct glog {
	FILE *log;
	unsigned verb;
	void *data;
	void (*logfunc) (void *data, unsigned level, const char *str);
	const char *(*errorstr) (int error);
};

struct gigstrlist {
	struct gigstrlist *next;
	char *str;
};

struct gstatus;

const char *gigaset_default_interface(void);

int gigaset_usleep_unint(int_fast32_t usecs);
int_fast32_t gigaset_usleep(int_fast32_t usecs);

struct gstatus *gigaset_open(int *gig_err, int *sys_err, const char *filename, struct glog *gl,
                             volatile sig_atomic_t *sigint);
int gigaset_close(struct gstatus *gs);
int gigaset_clear(struct gstatus *gs);
int gigaset_brkchars(struct gstatus *gs, const unsigned char buf[6]);
int gigaset_lock(struct gstatus *gs, int l);
//int gigaset_connected(struct gstatus * gs);

int gigaset_reset (struct gstatus * gs);
int gigaset_initconfig (struct gstatus * gs);
int gigaset_quitconfig (struct gstatus * gs);
int gigaset_unimodem (struct gstatus * gs);
int gigaset_cidmode (struct gstatus * gs);

void gigaset_logerr(struct glog *gl, unsigned level, int gig_err, int sys_err, const char *text);
void gigaset_logbuf(struct glog *gl, unsigned level, const char *text, const unsigned char *data, size_t len);
void gigaset_logstr(struct glog *gl, unsigned level, const char *text, const char *data);
void gigaset_log(struct glog *gl, unsigned level, const char *format, ...)
                 attribute_format (printf, 3, 4);

void gigaset_mayinterrupt(struct gstatus *gs, int status);
int gigaset_errno(const struct gstatus *gs);
int gigaset_curmode(struct gstatus *gs);

int gigaset_enumerate_devices(struct gigstrlist **list);
void gigaset_freestrlist(struct gigstrlist **list);

#endif // GIGASET_INTERF_H
