/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#ifndef GIGASET_SMS_H
#define GIGASET_SMS_H

#include <stdint.h>
#include "interf.h"

#define SMS_LEN_NUMBER         19

/* first octet of the SM */
#define SMS_TP_MTI             0x03U
#define SMS_TP_MTI_DELIVER     0x00U
#define SMS_TP_MTI_SUBMIT      0x01U
#define SMS_TP_MMS             0x04U  /* DELIVER */
#define SMS_TP_RD              0x04U  /* SUBMIT */
#define SMS_TP_VPF             0x18U
#define SMS_TP_VPF_SEMIOCTETS  0x18U
#define SMS_TP_VPF_INT         0x10U
#define SMS_TP_RP              0x80U
#define SMS_TP_UDHI            0x40U
#define SMS_TP_SRR             0x20U  /* SUBMIT */
#define SMS_TP_SRI             0x20U  /* DELIVER */

/* TP-DCS */
#define SMS_DCS_TYPE                 0xf0U
#define SMS_DCS_TYPE_NONSTD          0x40U /* values below SMS_DCS_TYPE_NONSTD => standard type below*/
#define SMS_DCS_TYPE_MWI_DISCARDMSG  0xC0U
#define SMS_DCS_TYPE_MWI             0xD0U
#define SMS_DCS_TYPE_MWI_UCS2        0xE0U
#define SMS_DCS_TYPE_F0              0xF0U

#define SMS_DCS_MSGCLASS       0x03U /* for type F0 and standard type */
#define SMS_DCS_ENCODING       0x04U /* for type F0 */
#define SMS_DCS_EXTENCODING    0x0CU /* for standard type */

#define SMS_DCS_7BIT           0x00U
#define SMS_DCS_8BIT           0x04U
#define SMS_DCS_UCS2           0x08U

#define SMS_DCS_HAVE_MSGCLASS  0x10U /* for standard type */
#define SMS_DCS_COMPRESSED     0x20U /* for standard type */

#define SMS_DCS_MWI_SENSE      0x08U
#define SMS_DCS_MWI_CLEAR      0x00U
#define SMS_DCS_MWI_SET        0x08U

#define SMS_DCS_MWI_TYPE       0x03U
#define SMS_DCS_MWI_VOICE      0x00U
#define SMS_DCS_MWI_FAX        0x01U
#define SMS_DCS_MWI_EMAIL      0x02U
#define SMS_DCS_MWI_OTHER      0x03U

/* SM Status */
#define SMS_SEND         0x0001
#define SMS_SENT         0x0002
#define SMS_STORE        0x0004
#define SMS_NEW          0x0008
#define SMS_INBOX        0x0010
#define SMS_ERROR        0x0020
#define SMS_DELETE       0x0040
#define SMS_TEMP         0x0080
#define SMS_MSN_SHIFT    8
#define SMS_MSN_MASK     0x000f
#define SMS_CHANGE_SHIFT 12
#define SMS_CHANGE_MASK  0x000f


struct gigaset_sm {
	struct gigaset_sm *prev, * next;
	struct gigaset_sm_chain * chain;

	/* AT^SSMLIST */
	uint32_t handle;
	uint16_t status; /* SMS_xxxx */

	/* AT^SSMINFO */
	char number[20];
	uint16_t error;
	int year, month, day, hour, minute;
	int MSN;

	/* AT^SSM{GET,PUT}TXT */
	unsigned char data[140];
	uint16_t size;
	uint16_t characters;
	uint8_t flags; /* SMS_TP_xxxx */
	uint8_t dcs; /* TP-DCS: SMS_DCS_xxxx */
	uint8_t unknown1; /* unknown byte; might hold TP-PID */
	uint8_t vp; /* TP-VP */

	/* from TP-UDH */
	uint8_t udhl; /* user data header length+1; 0 if TP-UDHI is not set */
	uint8_t id; /* id in chain */

	unsigned udc; /* number of user data information elements */
	unsigned char*udv[(140-1)/2]; /* pointers to the information elements */
};

struct gigaset_sm_chain {
	struct gigaset_sm_chain * prev, * next;
	struct gigaset_sm *msg[0xff]; /* messages 0x001..0x100 */

	char number[SMS_LEN_NUMBER+1];
	uint8_t chain_id;
	uint8_t msgcount;
	unsigned missing;
};

int gigaset_smsstart (struct gstatus *gs);
int gigaset_smscontinue (struct gstatus *gs);
int gigaset_smsend (struct gstatus * gs);
ssize_t gigaset_smsgetlist (struct gstatus * gs, char ** buffer);
ssize_t gigaset_smsgettext  (struct gstatus * gs, char ** buffer);
int gigaset_smsputtext  (struct gstatus * gs, const char * buffer, size_t len);
int gigaset_smsdel (struct gstatus * gs);
int gigaset_smsold (struct gstatus * gs);
#if 0
/* does not work as expected */
int gigaset_smsdelhandle(struct gstatus * gs, uint32_t num);
int gigaset_smsoldhandle(struct gstatus * gs, uint32_t num);
#endif
int gigaset_smssend (struct gstatus * gs);
int gigaset_smsstore (struct gstatus * gs);
int gigaset_smsreset (struct gstatus * gs);
int gigaset_smsnew  (struct gstatus * gs, uint32_t * num);
int gigaset_smslevel  (struct gstatus * gs);
int gigaset_smsselect (struct gstatus * gs, uint32_t num);
int gigaset_smsnumber (struct gstatus * gs, const char * src);
ssize_t gigaset_smsinfo (struct gstatus * gs, char ** buf);


/* high level funtions */

int gigaset_smsputmessage(struct gstatus *gs, const struct gigaset_sm *msg);

int gigaset_smsgetmessages (struct gstatus * gs, struct gigaset_sm ** messages,
                            struct gigaset_sm_chain ** chains);
void gigaset_smsfreemessages (struct gigaset_sm ** messages,
                              struct gigaset_sm_chain ** chains);

ssize_t gigaset_sms2utf8 (struct gigaset_sm * msg, int chain, char ** buffer);

int gigaset_smsfromutf8 (struct gigaset_sm ** retmsg, struct gigaset_sm_chain ** retchain,
                         uint8_t flags, uint8_t unknown1, uint8_t dcs, uint8_t vp,
                         int chain, uint8_t chain_id, const char * buffer, size_t len);

#endif /* GIGASET_SMS_H */
