/*
 * This part of the configuration program handles the block transfer
 *
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

/* Many thanks to Paul Bolle, for helping me with this. */

#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <gigaset/interf.h>
#include "interf-private.h"
#include <gigaset/block.h>
#include <gigaset/gigdata.h>

#define MAX_SIZE 110U    /* 0xffff did not work because the AT commands get a bit too long. */

/* decode received packet
 * Parameters (in)
 *   src:       (byte stuffed) input string
 * Parameters (out)
 *   type:      0: preamble
 *              1: data packet
 *              2: last data packet
 *              3: ack/nack/...
 *   seq:       sequence number
 *   datatype:  0x10: sms upload (?)
 *              ...
 *   buffer:    output buffer (malloc()ed, only valid if no error occured,
 *              the previous value is ignored)
 *   size.      number of bytes
 * Returns:
 *   <0:        errorcode
 *   0:         error
 */
static int gigaset_blockdecode (uint8_t * type, size_t *size, uint8_t * seq, uint8_t * datatype, char ** buffer, const char * src)
{
	char * buf;
	ssize_t s;

	s=gigaset_unstuff (&buf, src, strlen(src), 0x7d);

	if (s<0) return s;
	if (s<7) {
		free (buf);
		return -1;
	}
	s-=7;

	*datatype=buf[6];
	*seq=buf[5];
	*type=buf[4];
	if ((size_t)s>0xffffU||gigaset_from_be16 (buf+2)!=(size_t)s) {
		free (buf);
		return -1;
	}
	*size=s;

	if(!gigaset_verifychecksum(buf+2,s+5,gigaset_from_be16 (buf+0))) {
		free (buf);
		return -ERR_CRC;
	}

	memmove (buf, buf+7, s);
	*buffer=buf;
	return 0;
}

/* send packet
 * Parameters (in)
 *   type:      0: preamble
 *              1: data packet
 *              2: last data packet
 *              3: ack/nack/...
 *   seq:       sequence number
 *   datatype:  0x10: sms upload (?)
 *              ...
 *   data:      input buffer
 *   size.      number of bytes
 * Returns:
 *   <0:        errorcode
 *   0:         error
 */
static int gigaset_blocktransmit (struct gstatus * gs, uint8_t type, size_t size, uint8_t seq, uint8_t datatype, const char * data)
{
	char * buf, * cmd;
	ssize_t s;
	int r;

	if (size > MAX_SIZE) return -ERR_LENGTH;

	buf=malloc(size+7);
	if (!buf) return -ERR_MEM;

	memcpy (buf+7, data, size);
	buf[6]=datatype;
	buf[5]=seq;
	buf[4]=type;
	gigaset_to_be16 (buf+2,size);
	gigaset_to_be16 (buf+0,gigaset_checksum(buf+2, size+5));

	s=gigaset_bytestuff (&cmd, buf, size+7, 0x7d, 0x7e, 9, 1); /*0x7e okay?*/
	free(buf);

	if (s<0) return s;

	memcpy (cmd, "AT^SBDAT=", 9);
	cmd[9+s]='\0';

	r=gigaset_sendcommand (gs, cmd);
	free(cmd);

	return r;
}

static int gigaset_blockabort (struct gstatus * gs, uint8_t datatype)
{
	char cmd [14]="AT^SBABORT=";

	if (datatype<0x20||datatype==0x7d||datatype==0x7e/*??*/)
	{
		cmd[11]=0x7d;
		cmd[12]=datatype^0x20;
		cmd[13]=0;
	}
	else
	{
		cmd[11]=datatype;
		cmd[12]=0;
	}
	return gigaset_sendcommand (gs, cmd);
}

static struct resp_filter *add_abort_filter(struct gstatus *gs, uint8_t datatype)
{
	char filterstring[11] = "ZBABORT=\0\0\0";
	unsigned filterlen;

	if (datatype < 0x20 || datatype == 0x7d || datatype == 0x7e/*??*/) {
		filterstring[8] = 0x7d;
		filterstring[9] = datatype ^ 0x20;
		filterlen = 10;
	} else {
		filterstring[8] = datatype;
		filterlen = 9;
	}
	return gigaset_addfilter(gs, FILTER_ZBABORT, filterlen+1, filterstring, 0, NULL);
}

/* only window size == 1 supported in the following functions */
/* send data
 * Parameters (in)
 *   datatype:  0x10: sms upload (?)
 *              ...
 *   data:      input buffer
 *   size.      number of bytes
 * Returns:
 *   <0:        errorcode
 *   0:         error
 */
int gigaset_blocksend (struct gstatus * gs, size_t size, uint8_t datatype, const char * data)
{
	uint8_t phase=0; /* sending preamble */
	uint8_t seq=0;
	char preamble[7];
	uint16_t chunksize;
	ssize_t r;
	char * s;
	char * curblock;
	size_t cursize, bytesleft;
	size_t sendlen=0;
	uint8_t curseq, curtype, curdatatype;
	const char * pos = data;
	int badblocks=0;
	int skipwrite=0;
	int tout=10;
	int error;
	struct resp_filter *f;
	int doabort = 1;
	int second_try = 0;

	if (!size) return 0;
	if (size>0xffffffffUL) return -ERR_PARAMINVALID;

	f = add_abort_filter(gs, datatype);
	if (!f)
		return -ERR_MEM;

	bytesleft=size;
	chunksize = size > MAX_SIZE ? MAX_SIZE : size;

	for (;;)
	{
		if (!skipwrite)
		{
			if (phase==0)
			{
				gigaset_to_be32(preamble+0,size);
				gigaset_to_be16(preamble+4,chunksize);
				preamble[6]=1;
				r=gigaset_blocktransmit (gs, phase, 7, seq, datatype, preamble);
			}
			else
				r=gigaset_blocktransmit (gs, phase, sendlen, seq, datatype, pos);
			if (r<0) break;
		}
		skipwrite=0;

		r = gigaset_expectresponse(gs, "ZBDAT=", 6, &s, tout);
		tout=30; /* 1s is not enough, although suggested by the specs */
		if (r<0) {
			if (r == -FILTER_ZBABORT) {
				doabort = 0;
				r = -ERR_ABORT;
			}
			free(s);
			break;
		}

		r=gigaset_blockdecode (&curtype, &cursize, &curseq, &curdatatype, &curblock, s+6);
		free(s);
		gigaset_log(gs->gl, GLOG_DEBUG, "  data: r==%d/size==%d/datatype==0x%02x/ type==%d/seq==%d\n", (int)r,(int)cursize,curdatatype,curtype,curseq);
		error=1;
		if (r==-ERR_CRC)
			;
		else if (r<0||curdatatype!=datatype)
			skipwrite=1;
		else if (curseq!=seq||curtype!=3)
			;
		else if (phase==0&&cursize==8&&(unsigned char)*curblock==0xff) { /* invalid preamble */
			if (second_try || (chunksize = gigaset_from_be16(curblock+5)) > MAX_SIZE) {
				badblocks = 10;
			} else {
				second_try = 1;
			}
		} else if (cursize!=1)
			;
		else if ((unsigned char)*curblock==0xfe) /* "abort" */
			badblocks=10;
		else if (*curblock==3) /* "busy" */
		{
			skipwrite=1;
			tout=300;
		}
		else if (*curblock==1) /* ACK */
			error=0;

		free(curblock);
		if (error)
		{
			gigaset_log(gs->gl, GLOG_DEBUG, "  badblocks: (%d)\n", badblocks+1);
			if (++badblocks>10)
			{
				r=-ERR_CRC;
				break;
			}
		}
		else if (phase==2) { /* last block sent */
			r = 0;
			break;
		} else
		{
			badblocks=0;
			seq++;
			pos+=sendlen;
			bytesleft-=sendlen;
			if (bytesleft>chunksize)
			{
				phase=1;
				sendlen=chunksize;
			}
			else
			{
				phase=2;
				sendlen=bytesleft;
			}
		}
	}

	gigaset_delfilter(gs, f);

	if (r >= 0)
		return 0;

	if (doabort)
		gigaset_blockabort(gs, datatype);

	return r;
}

/* receive data
 * Parameters (in)
 *   datatype:  0x10: sms upload (?)
 *              ...
 * Parameters (out)
 *   buffer:    output buffer (malloc()ed, only valid if no error occured,
 *              the previous value is ignored)
 * Returns:
 *   <0:        errorcode
 *   >=0:       number of bytes
 */
ssize_t gigaset_blockreceive (struct gstatus * gs, uint8_t datatype, char ** buffer)
{
	ssize_t r=0;
	uint8_t phase=0; /* wait for preamble */
	uint8_t seq=0;
	char * s;
	int tout=50; // FIXME: initial timeout guessed
	int doabort=1;
	uint32_t totsize = 0;
	char * curblock;
	char * pos = 0;
	size_t cursize, bytesleft = 0;
	uint8_t curseq, curtype, curdatatype;
	char * dest=NULL;
	int badblocks=0;
	char response;
	int done=0;
	struct resp_filter *f;

	f = add_abort_filter(gs, datatype);
	if (!f)
		return -ERR_MEM;

	while (!done)
	{
		gigaset_log(gs->gl, GLOG_DEBUG, "<----loop:\n");
		gigaset_log(gs->gl, GLOG_DEBUG, "  status: datatype==0x%02x/phase==%d/seq==%d\n", datatype,phase,seq);
		r = gigaset_expectresponse(gs, "ZBDAT=", 6, &s, tout);
		tout=30; /* 3 secs */
		if (r<0) {
			if (r == -FILTER_ZBABORT) {
				doabort = 0;
				r = -ERR_ABORT;
			}
			free(s);
			break;
		}
		gigaset_logstr (gs->gl, GLOG_DEBUG, "  response: ", s);
		r=gigaset_blockdecode (&curtype, &cursize, &curseq, &curdatatype, &curblock, s+6);
		free(s);
		gigaset_log(gs->gl, GLOG_DEBUG, "  data: r==%d/size==%d/datatype==0x%02x/ type==%d/seq==%d\n", (int)r,(int)cursize,curdatatype,curtype,curseq);
		if (r==-ERR_CRC /*&& curdatatype==datatype*/||  /* why can we trust curdatatype? :-/ */
			(r==0&&curdatatype==datatype&&(curseq!=seq||(unsigned)curtype>2||phase!=(curtype?1:0))))
		{
			/* CRC error => NAK */
			if (r>=0) free(curblock);
			gigaset_log(gs->gl, GLOG_DEBUG, "  badblocks: (%d)\n", badblocks+1);
			if (++badblocks>10)
			{
				r=-ERR_CRC;
				break;
			}

			response=0; /* NAK */
			curseq=seq;
			curdatatype=datatype;
		}
		else
		{
			if (r<0||curdatatype!=datatype)
			{
				if (r>=0) free (curblock);
				continue; //FIXME abort? message?
			}
			if (phase==0)
			{
				/* preamble */
				if (curblock[6]!=1) /* window size */
				{
					r=-ERR_PROTO;
					free (curblock);
					break;
				}
				/* maximum chunk size from curblock+4,5 ignored */
				totsize=gigaset_from_be32(curblock+0);
				if (totsize>~(size_t)0||!(dest=malloc(totsize)))
				{
					r=-ERR_MEM;
					free (curblock);
					break;
				}
				bytesleft=totsize;
				++phase;
				pos=dest;
			}
			else if (bytesleft>=cursize)
			{
				/* data */
				memcpy (pos, curblock, cursize);
				pos+=cursize;
				bytesleft-=cursize;
				if (curtype==2) done=1;
			}
			else
			{
				/* got too much data */
				r=-1;
				free (curblock);
				break;
			}
			badblocks=0;
			response=1; /* ACK */
			++seq;
			free (curblock);
		}
		gigaset_log(gs->gl, GLOG_DEBUG, "---->resp %d\n", response);
		r=gigaset_blocktransmit (gs, 3, 1, curseq, curdatatype, &response); /* (N)ACK */
		if (r<0) break;
	}

	gigaset_delfilter(gs, f);

	if (r >= 0) {
		*buffer = dest;
		return totsize;
	}

	free(dest);

	if (doabort)
		gigaset_blockabort(gs, datatype);

	return r;
}
