/*
 * This part of the configuration program handles the "low level"
 * communication with the device driver.
 *
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
//#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <time.h>
#include <glob.h>

#include <gigaset/interf.h>
#include "interf-private.h"
#ifndef NOIOCTL
#include <linux/gigaset_dev.h>
#endif

// -> gigaset.h
#define M_UNKNOWN       0
#define M_CONFIG        1
#define M_UNIMODEM      2
#define M_CID           3

#define GIGASET_BUFLEN 1024

struct tbuf {
	struct tbuf *next;
	char buf[GIGASET_BUFLEN];
};

struct hwstate {
	int fd_dev;
	unsigned bufpos,startpos;
	size_t numbuf;
	struct tbuf *buf, *lastbuf;
#ifndef NOIOCTL
	int direct;
#endif
};

struct gstatus *gigaset_open(int *gig_err, int *sys_err, const char *filename, struct glog *gl,
                             volatile sig_atomic_t *sigint)
{
	int r=0;
	struct termios termios;
#ifndef NOIOCTL
	unsigned version[4];
#endif
	struct gstatus *gs;

	*gig_err = 0;
	*sys_err = 0;

	gs = malloc(sizeof *gs);
	if (!gs) {
		*gig_err = ERR_MEM;
		return NULL;
	}

	gs->gl = gl;
	gs->error = 0;

	gs->hws = malloc(sizeof *gs->hws);
	if (!gs->hws) {
		free(gs);
		*gig_err = ERR_MEM;
		return NULL;
	}

	gigaset_log(gs->gl, GLOG_DEBUG, "open character device\n");
	gs->hws->fd_dev=open(filename,O_RDWR|O_NONBLOCK);//O_ASYNC
	if (gs->hws->fd_dev<0) {
		*sys_err = errno;
		free(gs->hws);
		free(gs);
		*gig_err = ERR_OPEN;
		return NULL;
	}
	r=tcgetattr(gs->hws->fd_dev, &termios);
	if (r >= 0) {
		cfmakeraw(&termios);
		r=tcsetattr(gs->hws->fd_dev, TCSANOW, &termios);
	}
	if (r<0)
	{
		*sys_err = errno;
		close(gs->hws->fd_dev);
		free(gs->hws);
		free(gs);
		*gig_err = ERR_IO;
		return NULL;
	}
#ifdef NOIOCTL
	r = 0;
	gs->locked = 0;
	gigaset_log(gs->gl, GLOG_DEBUG, "direct mode\n");
#else
	version[0] = GIGVER_COMPAT;
	r = ioctl(gs->hws->fd_dev, GIGASET_VERSION, version);
	if (r < 0) {
		if (errno != EINVAL) {
			*sys_err = errno;
			r = -ERR_IO;
		} else {
			r = 0;
			gs->hws->direct = 1;
			gs->locked = 0;
			gigaset_log(gs->gl, GLOG_DEBUG, "direct mode\n");
		}
	} else {
		gigaset_log(gs->gl, GLOG_DEBUG, "driver compatible to version %u.%u.%u.%u\n",
	            version[0], version[1], version[2], version[3]);

		if (version[0] != 0 || version[1] != 4 || version[2] != 0 || version[3] > 1) {
			r = -ERR_VERSION;
		} else {
			gs->hws->direct = 0;
			gs->locked = -1;
		}
	}
#endif

	if (r < 0) {
		close(gs->hws->fd_dev);
		free(gs->hws);
		free(gs);
		*gig_err = -r;
		return NULL;
	}

	gs->sigint = sigint;
	gs->may_interrupt = 0;
	gs->have_ver_fwbase = 0;
	gs->prefix=NULL;
	gs->response=NULL;
	gs->hws->lastbuf=gs->hws->buf=NULL;
	gs->respqueue = gs->resplast = NULL;
	gs->hws->bufpos=0;
	gs->hws->startpos=0;
	gs->hws->numbuf=0;
	//gs->connected=-1;
	gs->config=0;
	gs->unimodem=0;
	gs->cidmode=0;
	gs->last_timeout = 0;
	gs->filter = gs->lastfilter = NULL;
	for (r=0; r<FLT_NUM; r++) gs->temp_filter[r] = NULL;
	if (!gigaset_addfilter(gs, FILTER_ERROR, 5+1, "ERROR", 0, NULL)) {
		gigaset_close(gs);
		*gig_err = -ERR_MEM;
		return NULL;
	}
	//FIXME ZMWI, ZABINFO, ... (+interf-win.c)

	return gs;
}

int gigaset_close(struct gstatus *gs)
{
	//FIXME error checking
	struct tbuf *cb;
	struct resp_filter *f;
	struct resp_buf *resp;

	close(gs->hws->fd_dev);
	free(gs->prefix);
	if (gs->response) {
		free(gs->response->str);
		free(gs->response);
	}

	while (gs->respqueue) {
		resp = gs->respqueue;
		gs->respqueue = resp->next;
		free(resp->str);
		free(resp);
	}

	while (gs->hws->buf) {
		cb = gs->hws->buf;
		gs->hws->buf = cb->next;
		free(cb);
	}

	while (gs->filter) {
		f = gs->filter;
		gs->filter = f->next;
		if (f->len) {
			f->str[f->len-1] = '\0';
			gigaset_logstr(gs->gl, GLOG_DEBUG, "free filter: ", f->str);
		}
		free (f);
	}

	free(gs->hws);
	free(gs);

	return 0;
}

static inline void * geteol(const void * s, size_t n)
{
	const unsigned char * t = s;
	while (n--) {
		if (*t=='\n'||*t=='\r')
			return (void *)t;
		++t;
	}
	return NULL;
}

int gigaset_read(struct gstatus *gs, int timeout)
{
	char *eol, *dest, *buf;
	size_t n;
	ssize_t d;
	struct tbuf *cb;

	if (timeout < 0)
		timeout = gs->last_timeout;
	gs->last_timeout = 0;

	if (gs->response) {
		free(gs->response->str);
		free(gs->response);
		gs->response = NULL;
	}

	if (!gs->hws->lastbuf)
		eol = NULL;
	else {
		buf = gs->hws->lastbuf->buf;
		n = gs->hws->bufpos;
		if (gs->hws->buf == gs->hws->lastbuf) {
			buf += gs->hws->startpos;
			n -= gs->hws->startpos;
		}
		eol = geteol(buf, n);
	}
	while (!eol) {
		if (!gs->hws->buf) {
			gs->hws->buf = gs->hws->lastbuf = malloc(sizeof (*gs->hws->buf));
			if (!gs->hws->buf)
				return -ERR_MEM;
			gs->hws->buf->next = NULL;
			gs->hws->bufpos = 0;
			gs->hws->startpos = 0;
			gs->hws->numbuf = 1;
		} else if (gs->hws->bufpos == GIGASET_BUFLEN) {
			cb = malloc(sizeof (*cb));
			if (!cb)
				return -ERR_MEM;
			cb->next = NULL;
			++gs->hws->numbuf;
			gs->hws->lastbuf->next = cb;
			gs->hws->lastbuf = cb;
			gs->hws->bufpos = 0;
		}

		n = GIGASET_BUFLEN - gs->hws->bufpos;
		for (;;) {
			//FIXME use select
			d = read(gs->hws->fd_dev, gs->hws->lastbuf->buf + gs->hws->bufpos, n);
			if (d < 0) {
				gs->error = errno;
				if (gs->error != EAGAIN && gs->error != EINTR)
					return -ERR_READ;
			}
			if (d > 0) break;

			if (gs->may_interrupt && gs->sigint && *gs->sigint)
				return -ERR_SIGINT;

			if (!timeout)
				return -ERR_TIMEOUT;
			if (timeout < GIG_INFTY)
				--timeout;

			gigaset_usleep_unint(100000);
		}

		eol = geteol(gs->hws->lastbuf->buf + gs->hws->bufpos, d);
		gs->hws->bufpos += d;
	}
	gs->response = malloc (sizeof *gs->response);
	if (!gs->response)
		return -ERR_MEM;
	n = (eol - gs->hws->lastbuf->buf) + (gs->hws->numbuf - 1) * (size_t)GIGASET_BUFLEN - gs->hws->startpos + 1;
	gs->response->str = malloc(n + 1); /* integer overflow not possible */
	if (!gs->response->str) {
		free(gs->response);
		gs->response = NULL;
		return -ERR_MEM;
	}
	gs->response->str[n] = '\0';
	gs->response->len = n;

	dest = gs->response->str;

	while (gs->hws->lastbuf != gs->hws->buf) {
		n = GIGASET_BUFLEN - gs->hws->startpos;
		memcpy(dest, gs->hws->buf->buf + gs->hws->startpos, n);
		dest += n;
		cb = gs->hws->buf;
		gs->hws->buf = cb->next;
		gs->hws->startpos = 0;
		--gs->hws->numbuf;
		free(cb);
	}

	n = (eol - gs->hws->buf->buf) - gs->hws->startpos + 1;
	memcpy (dest, gs->hws->buf->buf + gs->hws->startpos, n);
	gs->hws->startpos += n;
	gs->last_timeout = timeout;

	gigaset_logstr(gs->gl, GLOG_IO, "received: ", gs->response->str);

	return 0;
}

/* FIXME length as parameter? */
int gigaset_write (struct gstatus * gs, const char * s)
{
	size_t n;
	ssize_t d;

	gigaset_logstr (gs->gl, GLOG_IO, "writing: ", s);

	n=strlen(s);
	while (n)
	{
		d=write(gs->hws->fd_dev, s, n);
		if (d<0) {
			gs->error=errno;
			return -ERR_WRITE;
		}
		n-=d;
		s+=d;
		// d==0...
	}
	return 0;
}

int gigaset_clear(struct gstatus *gs)
{
	int r;
	struct tbuf *cb;
	struct resp_buf *resp;

	gigaset_log(gs->gl, GLOG_DEBUG, "flushing buffers\n");
	if (gs->gl->verb >= 2) {
		for (;;) {
			r = gigaset_read(gs, 0);
			if (r < 0) {
				if (r == -ERR_TIMEOUT)
					r = 0;
				break;
			}
		}
	} else {
		r = tcflush(gs->hws->fd_dev, TCIFLUSH);
		if (r != 0) {
			gs->error = errno;
			r = -ERR_READ;
		}
	}

	while (gs->hws->buf) {
		cb = gs->hws->buf;
		gs->hws->buf = cb->next;
		free(cb);
	}
	gs->hws->lastbuf = NULL;
	gs->hws->bufpos = gs->hws->startpos = 0;
	gs->hws->numbuf = 0;

	while (gs->respqueue) {
		resp = gs->respqueue;
		gs->respqueue = resp->next;
		free(resp->str);
		free(resp);
	}
	gs->resplast = NULL;

	return r;
}

int gigaset_brkchars(struct gstatus * gs, const unsigned char buf[6])
{
	gigaset_logbuf (gs->gl, GLOG_DEBUG, "brkchars: ", buf, 6);
#ifdef NOIOCTL
	return -ERR_UNIMPLEMENTED;
#else
	if (gs->hws->direct)
		return -ERR_UNIMPLEMENTED;

	if (ioctl(gs->hws->fd_dev,GIGASET_BRKCHARS,buf)<0)
	{
		gs->error=errno;
		return -ERR_IO;
	}
	return 0;
#endif
}

int gigaset_lock(struct gstatus * gs, int l)
//l:0==unlock,1==lock,<0==locked?
//ret:<0:err;l<0:  0:unlocked; 1:locked
//           l==1: M_xxxx before locking
{
	int r;

	if(l>1) return -ERR_PARAMINVALID;

	//if(gs->connected!=1) return -ERR_CONNECT;
	//if(l==0&&gs->config==1) return -ERR_CONFIG;
#ifdef NOIOCTL
	if (l < 0)
		return gs->locked;
	r = M_UNKNOWN;
#else
	if (gs->hws->direct) {
		if (l < 0)
			return gs->locked;
		r = M_UNKNOWN;
	} else {
		r=l;
		if (ioctl(gs->hws->fd_dev,GIGASET_REDIR,&r)<0)
		{
			gs->error=errno;
			return -ERR_IO;
		}

		if (l<0)
		{
			gs->locked=r;
			return r;
		}
	}
#endif

	gs->locked=l;
	if (l) {
		gigaset_log (gs->gl, GLOG_DEBUG, "LOCK: status=%d\n", r);
		switch (r) {
		case M_CONFIG:
			gs->config=1;
			gs->cidmode=0;  //FIXME one variable
			gs->unimodem=0;
			break;
		case M_CID:
			gs->config=0;
			gs->cidmode=1;
			gs->unimodem=0;
			break;
		case M_UNIMODEM:
			gs->config=0;
			gs->cidmode=0;
			gs->unimodem=1;
			break;
		default:
			gs->config=0;
			gs->cidmode=0;
			gs->unimodem=0;
			//gigaset_lock (gs, 0);//FIXME
			//return -1;
		}
	}

#ifndef NOIOCTL
	if (gs->hws->direct)
		return 0;

	if (l && !gs->have_ver_fwbase) {
		gs->ver_fwbase[0] = GIGVER_FWBASE;
		r = ioctl(gs->hws->fd_dev, GIGASET_VERSION, gs->ver_fwbase);
		if (r >= 0) {
			gs->have_ver_fwbase = 1;
			gigaset_log(gs->gl, GLOG_DEBUG, "base: firmware version %u.%u.%u.%u\n",
			            gs->ver_fwbase[0], gs->ver_fwbase[1],
			            gs->ver_fwbase[2], gs->ver_fwbase[3]);
		} else if (errno == ENOENT) {
			gs->have_ver_fwbase = 0;
		} else {
			gs->error = errno;
			r = 0;
			ioctl(gs->hws->fd_dev, GIGASET_REDIR, &r);
			return -ERR_IO;
		}
	}
#endif
	return 0;
}

struct modemctrl {
	int state;
	long delay;
};

static const struct modemctrl ctrlseq[] = {
	{ TIOCM_DTR,           200000 },
	{ 0,                   200000 },
	{ TIOCM_RTS,           100000 },
	{ 0,                   100000 },
	{ TIOCM_RTS,           100000 },
	{ 0,                   100000 },
	{ TIOCM_RTS,           100000 },
	{ 0,                   100000 },
	{ TIOCM_RTS,           100000 },
	{ 0,                   100000 },
	{ TIOCM_RTS,           100000 },
	{ 0,                   100000 },
	{ TIOCM_RTS|TIOCM_DTR, 800000 },
	{ 0,                   -1}
};

int gigaset_config(struct gstatus * gs,int l)
//ret:<0:err
{
	int r;
	int state,oldstate;
	const struct modemctrl * cur;

	if(l>1||l<0) return -ERR_PARAMINVALID;
	//if(gs->connected!=1) return -ERR_CONNECT;//FIXME
	if(gs->locked!=1) return -ERR_LOCK;

	if(l==0) {
		gs->config=0;
		return 0;
	}

#if 1
	r=ioctl(gs->hws->fd_dev,TIOCMGET,&oldstate);
	if (r<0) {
		gs->error=errno;
		return -ERR_IO;
	}
	state=oldstate;
	gigaset_log (gs->gl, GLOG_DEBUG, "TIOCMGET: %02x\n", (unsigned)oldstate);
	for(cur=ctrlseq;cur->delay>=0;++cur) {
		state&=~(TIOCM_RTS|TIOCM_DTR);
		state|=cur->state;
		r=ioctl(gs->hws->fd_dev,TIOCMSET,&state);
		if (r>=0) {
			r=ioctl(gs->hws->fd_dev,TIOCMGET,&state);
			if (r>=0)
				gigaset_log (gs->gl, GLOG_DEBUG, "TIOCMGET: %02x\n", (unsigned)state);
		}
		if (r<0) {
			gs->error=errno;
			state=3; /* FIXME good value? */
			ioctl(gs->hws->fd_dev,TIOCMSET,&oldstate);
			return -ERR_IO;
		}
		if (cur->delay)
			gigaset_usleep_unint(cur->delay);
	}
#else
	if (gs->hws->direct)
		return -ERR_UNIMPLEMENTED;
	state=1;
	if (ioctl(gs->hws->fd_dev,GIGASET_CONFIG,&state)<0) {
		gs->error=errno;
		return -ERR_IO;
	}
#endif

	/* get rid of old replies stuck in the M10x's buffer */
	tcflush(gs->hws->fd_dev, TCIFLUSH); /* don't care if this fails */

	gs->config=1;
	gs->unimodem=0;
	gs->cidmode=0;
	//prefix??
	return 0;
}

int_fast32_t gigaset_usleep(int_fast32_t usecs)
{
	struct timespec req, rem;
	int ret;

	if (usecs < 0)
		return 0;

	req.tv_nsec = (usecs%1000000) * 1000;
	req.tv_sec = usecs/1000000;
	ret = nanosleep(&req, &rem);
	if (ret != -1)
		return 0;

	if (errno != EINTR)
		return -1;

	return (int_fast32_t)rem.tv_sec*1000000 + (int_fast32_t)rem.tv_nsec/1000;
}

int gigaset_usleep_unint(int_fast32_t usecs)
{
	struct timespec req, rem;
	int ret;

	if (usecs < 0)
		return 0;

	req.tv_nsec = (usecs%1000000) * 1000;
	req.tv_sec = usecs/1000000;
	for (;;) {
		ret = nanosleep(&req, &rem);
		if (ret != -1)
			return 0;
		if (errno != EINTR)
			return -1;
		req = rem;
	}
}

const char *gigaset_default_interface(void)
{
#if 0
	struct stat buf;

	if (stat("/dev/ttyG", &buf) != -1)
#endif
		return "/dev/ttyG";
}

#ifndef NOIOCTL
int gigaset_enumerate_devices(struct gigstrlist **list)
{
	glob_t globbuf;
	size_t i;
	char *str;
	struct gigstrlist *cur;
	size_t size;
	int ret;

	*list = NULL;

	ret = glob("/dev/ttyG*", 0, NULL, &globbuf);
	switch (ret) {
	case 0:
		break;
	case GLOB_NOMATCH:
		return 0;
	case GLOB_NOSPACE:
		return -ERR_MEM;
	default:
		return -ERR_IO_UNKNOWN;
	}

	for (i = 0; i < globbuf.gl_pathc; ++i) {
		size = strlen(globbuf.gl_pathv[i]) + 1;
		if (!size)
			continue; //FIXME

		str = malloc(size);
		if (!str)
			continue; //FIXME

		cur = malloc(sizeof *cur);
		if (!cur) {
			free (str);
			continue;
		}
		memcpy(str, globbuf.gl_pathv[i], size);
		cur->next = *list;
		cur->str = str;
		*list = cur;
	}

	globfree (&globbuf);

	return 0;
}
#else
int gigaset_enumerate_devices(struct gigstrlist **list)
{
	*list = NULL;
	return 0;
}
#endif
