/*
 * This part of the configuration program handles the "low level"
 * communication with the device driver.
 *
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <limits.h>

#include <gigaset/interf.h>
#include "interf-private.h"

static int getnum(unsigned *dest, char **src, char delim)
{
	unsigned val = 0;
	unsigned dig;

	if (**src == delim || !**src)
		return -ERR_DATA;

	do {
		if (**src < '0' || **src > '9')
			return -ERR_DATA;
		dig = **src - '0';
		++*src;
		if (val > UINT_MAX / 10)
			return -ERR_DATA;
		val *= 10;
		if (val > UINT_MAX - dig)
			return -ERR_DATA;
		val += dig;
	} while (**src && **src != delim);

	if (**src)
		++*src;
	*dest = val;
	return 0;
}

static int getver(struct gstatus *gs)
{
	ssize_t r;
	char *buf;
	char *pos;

	if (gs->have_ver_fwbase)
		return 0;

	r = gigaset_sendcommand(gs, "AT+GMR");
	if (r < 0)
		return r;
	r = gigaset_getresponse(gs, &buf, 30);
	if (r < 0)
		return r;
	r = gigaset_expectresponse(gs, "OK", 2+1, NULL, 0);
	if (r >= 0) {
		pos = buf;
		r = getnum(gs->ver_fwbase+0, &pos, '.');
		if (r >= 0) r = getnum(gs->ver_fwbase+1, &pos, '.');
		if (r >= 0) r = getnum(gs->ver_fwbase+2, &pos, '.');
		if (r >= 0) r = getnum(gs->ver_fwbase+3, &pos, 0);
		if (r >= 0) {
			gs->have_ver_fwbase = 1;
			r = 0;
		}
	}
	free(buf);
	return r;
}

int gigaset_unimodem (struct gstatus * gs)
{
	ssize_t r;

	if (gs->config) return -1; //FIXME

	/* set mode even if we think it is set already
	 * (sometimes the device flips without our knowing)
	 */
	if (gs->unimodem || gs->cidmode) {
		gs->cidmode=0;
		r=gigaset_sendcommand(gs,"AT^SGCI=0");
		if (r<0) return r; //FIXME reset?
		r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
		if (r<0) return r; //FIXME reset?
		gs->unimodem=1;
		r = getver(gs);
		if (r < 0)
			return r;
		return 0;
	}

	/* FIXME test also for dle-mode, config mode, and connection in unimodem mode (+++) */
	r=gigaset_sendcommand(gs,"ATZ");
	if (r<0) return r;
	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 100);
	if (r<0) return r;
	gs->unimodem=1;
	r = getver(gs);
	if (r < 0)
		return r;
	return 0;
}

int gigaset_cidmode (struct gstatus * gs)
{
	ssize_t r;

	if (gs->config) return -1; //FIXME

	/* set mode even if we think it is set already
	 * (sometimes the device flips without our knowing)
	 */
	if (gs->unimodem || gs->cidmode) {
		r = getver(gs);
		if (r < 0)
			return r;
		gs->unimodem=0;
		r=gigaset_sendcommand(gs,"AT^SGCI=1");
		if (r<0) return r; //FIXME reset?
		r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
		if (r<0) return r; //FIXME reset?
		gs->cidmode=1;
		return 0;
	}

	/* FIXME test also for dle-mode, config mode, and connection in unimodem mode (+++) */
	r=gigaset_sendcommand(gs,"ATZ");
	if (r<0) return r;
	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 100);
	if (r<0) return r;
	r = getver(gs);
	if (r < 0)
		return r;
	r=gigaset_sendcommand(gs,"AT^SGCI=1");
	if (r<0) return r;
	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r<0) return r;
	gs->cidmode=1;
	return 0;
}

int gigaset_sendcommand (struct gstatus * gs, const char * s)
{
	char * t;
	size_t n,len;
	int r;

	//if (gs->connected!=1) return -ERR_CONNECT;
	if (gs->locked!=1) return -ERR_LOCK;

	len=strlen(s);
	n=len+2;
	t=malloc(n);
	if(!t) return -ERR_MEM;
	memcpy (t,s,len);
	t[n-2]='\r';
	t[n-1]='\0';
	r=gigaset_write(gs,t);
	free (t);
	if (r<0) return r;
	return 0;
}

ssize_t gigaset_getresponse(struct gstatus *gs, char **p, int timeout)
{
	size_t n;
	int r;

	if (gs->locked != 1)
		return -ERR_LOCK;

	if (!p)
		return -ERR_PARAMINVALID;

	*p = NULL;

	for (;;) {
		r = gigaset_read(gs, timeout);
		timeout = -1;
		if (r < 0)
			return r;
		if (gs->response->len != 1)
			break;
		gigaset_log(gs->gl, GLOG_DEBUG, "skipped response\n");
	}

	*p = gs->response->str;
	n = gs->response->len - 1;
	(*p)[n] = 0; /* remove end of line character */

	free(gs->response);
	gs->response = NULL;

	return n;
}

static void free_resp (struct gstatus *gs, struct resp_buf *resp, int free_string)
{
	if (resp->next)
		resp->next->prev = resp->prev;
	else
		gs->resplast = resp->prev;
	if (resp->prev)
		resp->prev->next = resp->next;
	else
		gs->respqueue = resp->next;
	if (free_string)
		free(resp->str);
	free(resp);
}

static int apply_filters(struct gstatus *gs, struct resp_buf *resp, char *cidptr, unsigned cidlen)
{
	struct resp_filter *f;
	size_t len;

	len = resp->len;
	if (cidlen) {
		if (resp->len < cidlen)
			return -1;
		len -= cidlen;
		if (memcmp(cidptr, resp->str + len, cidlen))
			return -1;
		resp->str[len] = 0;
	}

	for (f = gs->filter; f; f = f->next)
		if (f->len <= len + 1 && !memcmp(f->str, resp->str, f->len))
			break;

	if (!f) {
		if (cidlen)
			resp->str[len] = ';';
		return -1;
	}

	f->status = 1;

	if (f->retval <= 0) {
		free_resp(gs, resp, 1);
		return -1;
	}

	resp->len = len;
	return f->retval;
}

/* read from device until one of the expected responses turns up
 * responses are expected if they start with either a string previously added
 * filter, or the filter string passed to this function (if any)
 * Parameters:
 *	gs	status structure
 *	p	pointer to a pointer variable which will be set to a malloc'd
 *		copy of the actually received response or NULL on return
 *		(may be NULL)
 *	timeout	maximum time to wait, in 1/10 sec.
 *	cid	if >0, only consider responses with this call ID
 *	str	expected normal response (NULL: only previously added filters)
 *	len	number of characters in str
 *	prefix	M10x configuration prefix (NULL: base response)
 *	preflen	number of characters in prefix
 * Return value:
 *	length of response (>0) if response specified in str received
 *	-(filter return value of response) if other expected response received
 *	error codes: -ERR_TIMEOUT, -ERR_SIGINT, -ERR_MEM, -ERR_READ
 */
static ssize_t gigaset_getfilteredresponse(struct gstatus *gs, char **p, int timeout, int cid, const char *str, size_t len, const char *prefix, size_t preflen)
{
	struct resp_filter *tempfilter = NULL;
	int match;
	char cidbuf[CID_MAXLEN+2];
	unsigned cidlen = 0;
	char *cidptr = NULL;
	ssize_t retval;
	struct resp_buf *resp, *next;
	int r;

	if (p) *p = NULL;

	if (gs->locked != 1)
		return -ERR_LOCK;
	if (cid > CID_MAX)
		return -ERR_PARAMINVALID;
	if (str) {
		gigaset_logstr(gs->gl, GLOG_DEBUG, "expecting: ", str);
		tempfilter = gigaset_addfilter(gs, FILTER_CURRENT, len, str, preflen, prefix);
		if (!tempfilter)
			return -ERR_MEM;
	}
	else
		gigaset_log(gs->gl, GLOG_DEBUG, "expecting response\n");
	if (cid > 0) {
		cidptr = cidbuf + sizeof cidbuf;
		*--cidptr = 0;
		cidlen = 1;
		do {
			*--cidptr = '0' + cid % 10;
			++cidlen;
			cid /= 10;
		} while (cid);
		*--cidptr = ';';
	}

	match = -1;
	for (resp = gs->respqueue; resp; ) {
		next = resp->next;
		match = apply_filters(gs, resp, cidptr, cidlen);
		if (match >= 0)
			break;
		resp = next;
	}

	while (match < 0) {
		for (;;) {
			r = gigaset_read(gs, timeout);
			timeout = -1;
			if (r < 0) {
				retval = r;
				goto exit;
			}
			if (gs->response->len != 1)
				break;
			gigaset_log(gs->gl, GLOG_DEBUG, "skipped response\n");
		}
		resp = gs->response;
		gs->response = NULL;
		--resp->len;
		resp->str[resp->len] = 0; /* remove end of line character */

		resp->next = NULL;
		resp->prev = gs->resplast;
		gs->resplast = resp;
		if (resp->prev)
			resp->prev->next = resp;
		else
			gs->respqueue = resp;

		match = apply_filters(gs, resp, cidptr, cidlen);
	}

	if (match == FILTER_CURRENT) {
		if (prefix) {
			resp->len -= preflen + 2;
			memmove(resp->str, resp->str + preflen + 2, resp->len + 1);
		}
		gigaset_logstr(gs->gl, GLOG_DEBUG, "got: ", resp->str);
		retval = resp->len;
	} else {
		gigaset_log(gs->gl, GLOG_DEBUG, "matched: %d\n", match);
		retval = -match;
	}
	if (p) {
		*p = resp->str;
		free_resp(gs, resp, 0);
	} else
		free_resp(gs, resp, 1);

exit:
	if (tempfilter)
		gigaset_delfilter(gs, tempfilter);
	return retval;
}

static int getprefix(struct gstatus *gs, char **p, int timeout)
{
	const char *cmd = "AT^GETPRE";
	ssize_t r;

	if (gs->config != 1)
		return -ERR_MODE;

	r = gigaset_sendcommand(gs, cmd);
	if (r < 0) return r;
	r = gigaset_expectresponse(gs, cmd, strlen(cmd) + 1, NULL, timeout);
	if (r < 0) return r;
	r = gigaset_getresponse(gs, p, timeout);
	if (r < 0) return r;
	r = gigaset_expectresponse(gs, "OK", 2+1, NULL, timeout);
	if (r == -ERR_TIMEOUT && !strcmp(*p, "OK")) {
		**p = 0;
		r = 0;
	}
	if (r < 0) {
		free(*p);
		*p = NULL;
	}
	return r;
}

int gigaset_initconfig (struct gstatus * gs)
{
	//FIXME: retry on error. error checking: echo line, "OK", empty lines
	int r;
	int i;

	free (gs->prefix);
	gs->prefix=NULL;

	//r=gigaset_config(gs,2);
	//if (r<0) return r;

	if (gs->config!=1)
	{
		r=gigaset_config(gs,1);
		if (r<0) return r;
	}
	for (i=0;i<3;++i)
	{
		r=getprefix(gs, &gs->prefix, 30);
		if (r>=0) break;
		gigaset_usleep_unint(1000000);
	}
	if (r<0)
	{
		// FIXME what should we do?
		gigaset_config(gs,0);
		gs->prefix=NULL;
		return r;
	}
	gs->preflen=strlen(gs->prefix);
	return 0;
}

int gigaset_configcommand (struct gstatus * gs, const char * s)
{
	size_t l1,l2;
	char * cmd;
	int r;

	if (gs->config!=1) return -1;
	l1=strlen(gs->prefix);
	l2=strlen(s);
	cmd=malloc(l1+l2+1+2);
	cmd[0]='A';
	cmd[1]='T';
	memcpy (cmd+2, gs->prefix, l1);
	memcpy (cmd+l1+2, s, l2+1);

	r=gigaset_sendcommand(gs,cmd);
	free (cmd);
	return r;
}

int gigaset_quitconfig (struct gstatus * gs)
{
	//FIXME: retry on error?
	int r;

	r = gigaset_configsequence(gs, "O", 0, 0, NULL, 30);

	r=gigaset_config(gs,0);
	if (r<0) return r;

	free(gs->prefix);
	gs->prefix=NULL;

	return 0;
}

int gigaset_reset (struct gstatus * gs)
{
	//FIXME: retry on error?
	ssize_t r;

	r=gigaset_sendcommand(gs,"ATZ");
	if (r<0) return r;
	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 100);
	if (r<0) return r;
	r=gigaset_sendcommand(gs,"AT^SGCI=1");
	if (r<0) return r;
	r=gigaset_expectresponse(gs, "OK", 2+1, NULL, 30);
	if (r<0) return r;
	return 0;
}

/* read from device until the expected response or one of the previously added
 * filter responses turns up for the given call ID
 * Parameters:
 *	gs	status structure
 *	s	expected response (may be NULL)
 *	l	number of characters in s
 *	cid	call ID to accept responses for
 *	p	pointer to a pointer variable which will be set to a malloc'd
 *		copy of the actually received response or NULL on return
 *	timeout	maximum time to wait, in 1/10 sec.
 * Return value:
 *	length of response (>0) if response specified in s received
 *	-(filter return value of response) if other expected response received
 *	error codes: -ERR_TIMEOUT, -ERR_SIGINT, -ERR_MEM, -ERR_READ
 */
ssize_t gigaset_expectcidresponse(struct gstatus *gs, const char *s, size_t l, unsigned cid, char **p, int timeout)
/* if p!=NULL => return string in *p, if response is found (retval>=0) */
{
	if (cid < 1 || cid > CID_MAX)
		return -ERR_PARAMINVALID;
	return gigaset_getfilteredresponse(gs, p, timeout, cid, s, l, NULL, 0);
}

/* read from device until the expected response or one of the previously added
 * filter responses turns up for any call ID or none at all
 * Parameters:
 *	gs	status structure
 *	s	expected response (may be NULL)
 *	l	number of characters in s
 *	p	pointer to a pointer variable which will be set to a malloc'd
 *		copy of the actually received response or NULL on return
 *		(may be NULL)
 *	timeout	maximum time to wait, in 1/10 sec.
 * Return value:
 *	length of response (>0) if response specified in s received
 *	-(filter return value of response) if other expected response received
 *	error codes: -ERR_TIMEOUT, -ERR_SIGINT, -ERR_MEM, -ERR_READ
 */
ssize_t gigaset_expectresponse(struct gstatus *gs, const char *s, size_t l, char **p, int timeout)
/* if p!=NULL => return string in *p, if response is found (retval>=0) */
{
	return gigaset_getfilteredresponse(gs, p, timeout, -1, s, l, NULL, 0);
}

/* read from device until the expected response or one of the previously added
 * filter responses turns up in M10x configuration mode
 * Parameters:
 *	gs	status structure
 *	s	expected response (may be NULL)
 *	l	number of characters in s
 *	p	pointer to a pointer variable which will be set to a malloc'd
 *		copy of the actually received response or NULL on return
 *		(may be NULL)
 *	timeout	maximum time to wait, in 1/10 sec.
 * Return value:
 *	length of response (>0) if response specified in s received
 *	-(filter return value of response) if other expected response received
 *	error codes: -ERR_TIMEOUT, -ERR_SIGINT, -ERR_MEM, -ERR_READ
 */
ssize_t gigaset_expectconfigresponse(struct gstatus *gs, const char *s, size_t l, char **p, int timeout)
/* if p!=NULL => return string in *p, if response is found (retval>=0) */
{
	if (gs->config != 1)
		return -1;
	return gigaset_getfilteredresponse(gs, p, timeout, -1, s, l, gs->prefix, gs->preflen);
}

/* send command in configuration mode and collect response
 * Parameters:
 *	gs	status structure
 *	cmd	command string to send
 *	strings	maximum number of response lines to collect, excluding final 'OK'
 *	testok  if !=0, "OK" terminates the list, even if fewer than 'strings'
 *		lines have been received.
 *	p	pointer to an array of 'strings' pointer variables which will be
 *		set to the received response lines on success
 *		NULL to discard replies
 *	timeout	maximum wait time for each response, in 1/10 s
 * Return value:
 *	<0 on error (error codes -ERR_...)
 *	>=0 number of strings
 */
int gigaset_configsequence(struct gstatus *gs, const char *cmd, int strings,
			   int testok, char **p, int timeout)
{
//FIXME allocate list dynamiyally, allow strings = "infinity"
	ssize_t r;
	char *b;
	char **cur;
	int count;

	if (gs->config != 1)
		return -ERR_MODE;
	if (strings <= 0)
		p = NULL;

	r = gigaset_configcommand(gs, cmd);
	if (r < 0) return r;
	r = gigaset_expectconfigresponse(gs, cmd, strlen(cmd) + 1, NULL, timeout);
	if (r < 0) return r;

	if (p)
		cur = p;
	else
		cur = &b;

	for (count = 0; count < strings; ++count) {
		r = gigaset_getresponse(gs, cur, timeout);
		if (r < 0)
			break;
		if (testok && !strcmp(*cur, "OK")) {
			free(*cur);
			*cur = NULL;
			return count;
		}
		if (p)
			++cur;
		else
			free(b);
	}

	if (r >= 0) {
		r = gigaset_expectresponse(gs, "OK", 2+1, NULL, timeout);
		if (r >= 0)
			return count;
	}

	if (p)
		while (cur != p) {
			free(*--cur);
			*cur = NULL;
		}

	return r;
}

struct resp_filter *gigaset_addfilter(struct gstatus *gs, int retval,
                                      size_t len, const char *str, size_t preflen, const char *prefix)
{
	struct resp_filter *f;
	size_t totlen;

	totlen = len;
	if (prefix)
		totlen += 2 + preflen;

	f = malloc(sizeof *f - 1 + totlen);
	if (!f)
		return NULL;

	f->retval = retval;
	f->status = 0;
	f->len = totlen;
	if (prefix) {
		f->str[0] = 'A';
		f->str[1] = 'T';
		memcpy(f->str + 2, prefix, preflen);
		memcpy(f->str + 2 + preflen, str, len);
	} else {
		memcpy(f->str, str, len);
	}

	f->prev = NULL;
	f->next = gs->filter;
	gs->filter = f;

	if (f->next)
		f->next->prev = f;
	else
		gs->lastfilter = f;

	return f;
}

void gigaset_delfilter(struct gstatus *gs, struct resp_filter *f)
{
	if (!f)
		return;
	if (f->prev)
		f->prev->next = f->next;
	else
		gs->filter = f->next;
	if (f->next)
		f->next->prev = f->prev;
	else
		gs->lastfilter = f->prev;
	free(f);
}


static void gigaset_lograw(struct glog *gl, unsigned level, const char *str)
{
	if (!gl || !(level & gl->verb))
		return;

	if (gl->log)
		fputs(str, gl->log);

	if (!gl->logfunc)
		return;

	gl->logfunc(gl->data, level, str);
}

static const char * const errstr[] = {
	"can't read from device",
	"can't write to device",
	"can't open device",
	"can't close device",
	"can't communicate with device",
	"out of memory",
	"no response from device",
	"invalid response",
	"invalid parameter value",
	NULL,	//ERR_CONNECT
	NULL,	//ERR_LOCK
	NULL,	//ERR_MSGUNEXPECTED
	NULL,	//ERR_CONFIG
	"too long",
	NULL,	//ERR_DATA
	NULL,	//ERR_ABORT
	"checksum error",
	"protocol error",
	"timeout exceeded",
	"no entry found",
	"bad or unsupported firmware version",
	"not implemented",
	"unknown I/O error",
	"interrupted",
	"wrong mode",
};

#define ERRSTR_MIN 1
#define ERRSTR_MAX (ERRSTR_MIN + sizeof errstr / sizeof errstr[0] - 1)

void gigaset_logerr(struct glog *gl, unsigned level, int gig_err, int sys_err, const char *text)
{
	const char *desc = NULL;
	const char *sep;
	char buf[20];

	if (gig_err >= ERRSTR_MIN && gig_err <= ERRSTR_MAX)
		desc = errstr[gig_err - ERRSTR_MIN];
	else if (gl->errorstr && gig_err > ERR_LIBMAX && gig_err <= ERR_MAX)
		desc = gl->errorstr(gig_err);

	if (!desc) {
		desc = buf;
		snprintf(buf, sizeof buf, "error %d", gig_err);
	}

	if (text)
		sep = ": ";
	else {
		sep = "";
		text = "";
	}

	if (ERR_WITH_ERRNO(gig_err) || (gig_err > ERR_LIBMAX && gig_err <= ERR_MAX && sys_err)) {
		gigaset_log(gl, level, "%s%s%s (error %d: %s)\n", text, sep, desc,
		            sys_err, strerror(sys_err));
	} else {
		gigaset_log(gl, level, "%s%s%s\n", text, sep, desc);
	}
}

static const char hexdig[16]="0123456789abcdef";

void gigaset_logbuf(struct glog *gl, unsigned level, const char *text, const unsigned char *data, size_t len)
{
	unsigned char c;
	size_t pos=0;
	char buf[4+1+1 + (2+1)*16 + 1];
	unsigned bufpos;

	fputs (text, gl->log);

	if (!data) {
		gigaset_log(gl, level, "%s%s\n", text, "<NULL>");
		return;
	}

	if (!len) {
		gigaset_log(gl, level, "%s%s\n", text, "<0 bytes>");
		return;
	}

	gigaset_log(gl, level, "%s\n", text);
	sprintf(buf, "%04x: ", 0U);
	bufpos = 4+1+1;
	for (;;) {
		pos++;
		c=*data++;
		buf[bufpos++] = hexdig[(c>>4)&0xf];
		buf[bufpos++] = hexdig[c&0xf];
		bufpos += 2;
		if (pos == len)
			break;
		if (pos & 0xf)
			buf[bufpos++] = ' ';
		else {
			buf[bufpos++] = '\n';
			buf[bufpos++] = 0;
			gigaset_lograw(gl, level, buf);
			sprintf(buf, "%04x: ", (unsigned)pos);
			bufpos = 4+1+1;
		}
	}
	buf[bufpos++] = '\n';
	buf[bufpos++] = 0;
	gigaset_lograw(gl, level, buf);
}

#define addchar(buf, pos, size, bufsize, c) \
	do { \
		if (pos != (bufsize)) { \
			*buf++ = (c); \
			++pos; \
		} \
		++size; \
	} while (0)

static size_t quote_string(char *buf, size_t bufsize, const char *text, const char *data)
{
	int instring;
	unsigned char c;
	size_t pos = 0;
	size_t size = 0;

	while (*text) {
		addchar(buf, pos, size, bufsize, *text);
		++text;
	}

	if (!*data) {
		addchar(buf, pos, size, bufsize, '\'');
		addchar(buf, pos, size, bufsize, '\'');
		addchar(buf, pos, size, bufsize, '\n');
		if (pos != bufsize)
			*buf = 0;
		else if (bufsize)
			*--buf = 0;
		return size;
	}

	instring = 0;
	for (c = *data; c; c = *++data)
		if (c != '\'') {
			if (!instring)
				addchar(buf, pos, size, bufsize, '\'');
			instring = 1;
			if (c == '%' || c == '\\') {
				addchar(buf, pos, size, bufsize, c);
				addchar(buf, pos, size, bufsize, c);
			} else if (isprint(c))
				addchar(buf, pos, size, bufsize, c);
			else {
				addchar(buf, pos, size, bufsize, '\\');
				addchar(buf, pos, size, bufsize, 'x');
				addchar(buf, pos, size, bufsize, hexdig[(c>>4)&0xf]);
				addchar(buf, pos, size, bufsize, hexdig[c&0xf]);
			}
		} else {
			if (instring)
				addchar(buf, pos, size, bufsize, '\'');
			instring = 0;
			addchar(buf, pos, size, bufsize, '\\');
			addchar(buf, pos, size, bufsize, '\'');
		}
	if (instring)
		addchar(buf, pos, size, bufsize, '\'');
	addchar(buf, pos, size, bufsize, '\n');
	if (pos != bufsize)
		*buf = 0;
	else if (bufsize)
		*--buf = 0;
	return size;
}

void gigaset_logstr(struct glog *gl, unsigned level, const char *text, const char *data)
{
	size_t size;
	char buf[80];
	char *str;

	size = quote_string(buf, sizeof buf, text, data);
	if (size >= sizeof buf) {
		++size;
		if (!size)
			--size;
		str = malloc(size);
		if (str) {
			quote_string(str, size, text, data);
			gigaset_lograw(gl, level, str);
			free (str);
			return;
		}
	}
	gigaset_lograw(gl, level, buf);
}

void gigaset_log(struct glog *gl, unsigned level, const char *format, ...)
{
	va_list argp;
	int size;
	char buf[80];
	char *str;

	if (!gl || !(level & gl->verb))
		return;

	if (gl->log) {
		va_start(argp, format);
		vfprintf(gl->log, format, argp);
		va_end(argp);
	}

	if (!gl->logfunc)
		return;

	va_start(argp, format);
	size = vsnprintf(buf, sizeof buf, format, argp);
	va_end(argp);
	if (size >= sizeof buf) {
		if (size != INT_MAX)
			++size;
		str = malloc(size);
		if (str) {
			va_start(argp, format);
			vsnprintf(str, size, format, argp);
			va_end(argp);
			gl->logfunc(gl->data, level, str);
			free(str);
			return;
		}
	}
	gl->logfunc(gl->data, level, buf);
}

void gigaset_mayinterrupt(struct gstatus *gs, int status)
{
	gs->may_interrupt = status;
}

int gigaset_errno(const struct gstatus *gs)
{
	return gs->error;
}

int gigaset_curmode(struct gstatus *gs)
{
	if (gs->cidmode)
		return GMODE_CID;
	if (gs->unimodem)
		return GMODE_UNIMODEM;
	if (gs->config)
		return GMODE_CONFIG;
	return GMODE_UNKNOWN;
}

void gigaset_freestrlist(struct gigstrlist ** list)
{
	struct gigstrlist *cur = *list;
	struct gigstrlist *tmp;

	*list = NULL;

	while (cur) {
		tmp = cur;
		cur = cur->next;
		free(tmp->str);
		free(tmp);
	}
}
