/*
 * Copyright (c) 2003 by Hansjoerg Lipp (hjlipp@web.de)
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <gigaset/interf.h>
#include "interf-private.h"
#include <gigaset/log.h>
#include <gigaset/gigdata.h>

int gigaset_logdel(struct gstatus *gs)
{
	int r;
	int newstyle;

	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	newstyle = gs->ver_fwbase[0] >= 2; /* this is guessed */

	if (newstyle)
		r = gigaset_sendcommand(gs, "AT^SLOGDEL");
	else
		r = gigaset_sendcommand(gs, "AT^SLOG=RESET"); // => ZLOG=0000
	if (r < 0)
		return r;
	r = gigaset_expectresponse(gs, "ZLOG=", 5, NULL, 30);
	if (r < 0)
		return r;
	return 0;
}

/* gigaset_logstatus
 * retrieve current number of journal entries and ID of last entry
 * Parameters:
 *	gs	status structure
 *	pid	receives ID of last entry
 *	pnum	receives number of entries
 * Return value:
 *	0 on success
 *	<0: error code
 */
int gigaset_logstatus(struct gstatus *gs, unsigned *pid, unsigned *pnum)
{
	int l, r;
	char *buf;
	unsigned char dst[6];
	int newstyle;

	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	newstyle = gs->ver_fwbase[0] >= 2; /* this is guessed */

	if (!newstyle)
		return -ERR_VERSION; //FIXME

	r = gigaset_sendcommand(gs, "AT^SLOG");
	if (r < 0)
		return r;
	l = gigaset_expectresponse(gs, "ZLOG=", 5, &buf, 30);
	if (l < 0) {
		free(buf);
		return l;
	}

	r = gigaset_unstuffbuf((char*) &dst, buf+5, l-5, 0x7d, 6);
	free(buf);
	if (r != 6) {
	    if (r == -1)	/* more than 6 result bytes */
		return -ERR_DATA;
	    if (r < 0)		/* unstuff error */
		return r;
	    return -ERR_DATA;	/* less than 6 result bytes */
	}

	if (pid) *pid = gigaset_from_be32(dst);
	if (pnum) *pnum = gigaset_from_be16(dst+4);
	return 0;
}


static const struct eltype {
	int type;
	int size;
} eltypes[] = {
	{ ELTYPE_CBUF,   0 },	/*  0: generic unknown	*/
	{ ELTYPE_STRING, 0 },	/*  1: Caller Number	*/
	{ ELTYPE_STRING, 0 },	/*  2: Caller Name	*/
	{ ELTYPE_STRING, 0 },	/*  3: Called Number	*/
	{ ELTYPE_STRING, 0 },	/*  4: Called Name	*/
	{ ELTYPE_NUMBER, 1 },	/*  5: Internal Number	*/
	{ ELTYPE_STRING, 0 },	/*  6: Internal Name	*/
	{ ELTYPE_NUMBER, 1 },	/*  7: Connection ID	*/
	{ ELTYPE_NUMBER, 4 },	/*  8: Short Message Handle	*/
	{ ELTYPE_NUMBER, 2 },	/*  9: New Status	*/
	{ ELTYPE_NUMBER, 2 },	/* 10: Answering Machine Handle	*/
	{ ELTYPE_CHRG,   0 },	/* 11: Charging Info	*/
	{ ELTYPE_NUMBER, 4 },	/* 12: Call Duration	*/
	{ ELTYPE_NUMBER, 4 },	/* 13: Recording Duration	*/
	{ ELTYPE_STRING, 0 },	/* 14: Controlling Number	*/
	{ ELTYPE_NUMBER, 1 },	/* 15: Message Count	*/
	{ ELTYPE_STRING, 0 },	/* 16: Receiving Number	*/
	{ ELTYPE_NUMBER, 1 }	/* 17: Prefix Length	*/
};

/* gigaset_logread
 * retrieve journal entry
 * Parameters:
 *	gs	status structure
 *	id	ID of entry to retrieve
 *	prec	receives pointer to allocated buffer containing retrieved entry
 * Return value:
 *	0 on success
 *	<0: error code
 */
int gigaset_logread(struct gstatus *gs, uint32_t id, struct glogrec **prec)
{
	int resplen, entlen, i, l, r;
	char *resp;
	const struct eltype *peltype;
	char *buf;
	char *entry;
	struct glogel *pel;
	char cmd[9+8+1];
	char idbuf[4];
	ssize_t s;
	int32_t fcs;
	int newstyle;

	if (!gs->have_ver_fwbase)
		return -ERR_VERSION;

	newstyle = gs->ver_fwbase[0] >= 2; /* this is guessed */

	*prec = NULL;

	memcpy(cmd + 0, "AT^SLOGR=", 9);
	gigaset_to_be32(idbuf, id);
	s = gigaset_stuffbuf(cmd + 9, idbuf, 4, 0x7d, 0x7e, 8);
	cmd[9+s] = 0;

	r = gigaset_sendcommand(gs, cmd);
	if (r < 0)
		return r;

	if (newstyle) {
		resp = "ZLOGR=";
		resplen = 6;
	} else {
		resp = "ZLOG=";
		resplen = 5;
	}
	l = gigaset_expectresponse(gs, resp, resplen, &buf, 30);
	if (l < 0) {
		free(buf);
		return l;
	}
	entry = buf + resplen;
	l -= resplen;

	if (l == 2 && !strncmp(entry, "NO", 2)) {
	    free(buf);
	    return -ERR_NOENT;
	}

	/* check length and checksum */
	if ((fcs = gigaset_fromhex4(entry+l-4)) < 0) {
		free(buf);
		return fcs;
	}
	if ((entlen = gigaset_unstuffbuf(entry, entry, l-4, 0x7d, l-4)) < 0) {
		free(buf);
		return entlen;
	}
	if (!gigaset_verifychecksum(entry, entlen, fcs)) {
		free(buf);
		return -ERR_CRC;
	}
	if (entlen < 13 || gigaset_from_le16(entry+1) != entlen-3) {
		free(buf);
		return -ERR_DATA;
	}
	gigaset_log(gs->gl, GLOG_DEBUG, "Entry %d: %d bytes\n", id, entlen);

	/* count elements, allocate appropriate buffer */
	l = 0;
	for (i = 13; i < entlen; i += 2 + (unsigned char) entry[i+1])
		l++;
	if (i > entlen) {
		/* last element extends beyond end of data */
		free(buf);
		return -ERR_DATA;
	}
	if ((*prec = malloc(sizeof(struct glogrec) + (l-1) * sizeof(struct glogel)))
	    == NULL) {
		free(buf);
		return -ERR_MEM;
	}

	/* decode entry */
	(*prec)->type = (unsigned char) entry[0];
	(*prec)->recid = gigaset_from_le32(entry+3);
	(*prec)->year = (unsigned char) entry[12];
	(*prec)->month = (unsigned char) entry[11];
	(*prec)->day = (unsigned char) entry[10];
	(*prec)->hour = (unsigned char) entry[9];
	(*prec)->min = (unsigned char) entry[8];
	(*prec)->sec = (unsigned char) entry[7];
	(*prec)->nels = l;
	pel = (*prec)->el;
	for (i = 13; i < entlen; i += 2 + (unsigned char) entry[i+1]) {
		pel->type = (unsigned char) entry[i];
		l = (unsigned char) entry[i+1];
		if (pel->type < sizeof eltypes/sizeof eltypes[0]) {
			peltype = &eltypes[pel->type];
		} else {
			peltype = &eltypes[0];
		}
		if (peltype->size != 0 && peltype->size != l) {
			/* unexpected size - fall back to "unknown" */
			peltype = &eltypes[0];
		}
		pel->type |= peltype->type;
		pel->str = NULL;
		switch (peltype->type) {
		case ELTYPE_STRING:
			/* check correct zero termination */
			if (memchr(entry+i+2, '\0', l) != entry+i+l+1)
				goto baddata;
			if ((pel->str = malloc(l)) == NULL) {
				r = -ERR_MEM;
				break;
			}
			memcpy(pel->str, entry+i+2, l);
			break;
		case ELTYPE_NUMBER:
			switch (l) {
			case 1:
				pel->num = (unsigned char) entry[i+2];
				break;
			case 2:
				pel->num = gigaset_from_le16(entry+i+2);
				break;
			case 4:
				pel->num = gigaset_from_le32(entry+i+2);
				break;
			default:
				goto baddata;
			}
			break;
		case ELTYPE_CHRG:
			if (l < 5)
				goto baddata;
			pel->num = gigaset_from_le32(entry+i+3);
			switch (entry[i+2]) {
			case 'U':
				if (l != 5)
					goto baddata;
				break;
			case 'C':
				if (l == 5)
					goto baddata;
				pel->scale = entry[i+7] - 3;
				if (memchr(entry+i+8, '\0', l-6) != entry+i+l+1)
					goto baddata;
				if ((pel->str = malloc(l-6)) == NULL) {
					r = -ERR_MEM;
					break;
				}
				memcpy(pel->str, entry+i+8, l-6);
				break;
			default:
				goto baddata;
			}
			break;
		default:	/* unimplemented eltype code */
		baddata:	/* content doesn't fit eltype */
			pel->type = (pel->type & ~ELTYPE_FORMAT) | ELTYPE_CBUF;
			/* --v-- fall through --v-- */
		case ELTYPE_CBUF:
			pel->num = l;
			if ((pel->str = malloc(l)) == NULL) {
				r = -ERR_MEM;
				break;
			}
			memcpy(pel->str, entry+i+2, l);
		}
		if (r < 0)
			break;
		pel++;
	}
	free(buf);
	if (r < 0) {
		while (pel != (*prec)->el) {
			--pel;
			if (pel->str)
				free(pel->str);
		}
		free(*prec);
		*prec = NULL;
	}
	return r;
}
