/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#ifndef CONFIG_HH
#define CONFIG_HH

#include <QtGui>
#include <limits>
#include "gigasetqt.hh"
extern "C" {
#include <gigaset/conf.h>
}

struct Interval {
	int min, max;
};

class Range {
public:
	Range(int min=std::numeric_limits<int>::min(),
	      int max=std::numeric_limits<int>::max())
	      : totmin(min), totmax(max) {}
	void add(int min, int max);
	void addMin(int min) { add(min, totmax); }
	void addMax(int max) { add(totmin, max); }
	bool parse(const QString &str, int fac1 = 1, int fac2 = 0, int fac3 = 0);
	bool contains(int val) const;
	int first() const { return ilist.isEmpty() ? totmin : ilist.first().min; }
	int last() const { return ilist.isEmpty() ? totmax : ilist.last().max; }
	int isEmpty() const { return ilist.isEmpty(); }
	bool next(int &val) const;
	bool prev(int &val) const;
private:
	bool parseNum(const QString &str, int &val, int deflt, int fac1 = 1, int fac2 = 0, int fac3 = 0);
	bool parsePart(const QString &str, int &val, int &pos, bool &done, int fac);
	int totmin, totmax;
	QList<struct Interval> ilist;
};

class ConfigValidator;

class ConfigItem
{
public:
	enum DataType { DTInvalid, DTByte, DTString };
	enum Type { TRegExp, TRange, THour, TMinute, TYear, TMonth, TDay };
	enum State { SInvalid, SValid, SNone, SIgnore };

	ConfigItem(ConfigItem *parent = 0, const QString &name = QString(), bool node = true, bool templ = false);
	~ConfigItem();

	void appendChild(ConfigItem *child);
	void removeChild(ConfigItem *child);

	ConfigItem *child(int row);
	int childCount() const;
	QVariant data(int column) const;
	QVariant deviceData() const { return device_data; }
	QVariant userData() const { return user_data; }
	bool setData(const QVariant &value);
	int row() const;
	ConfigItem *parent();

	const QString &name() const { return key; }
	bool isNode() const { return is_node; }
	bool isTemplate() const { return is_template; }
	bool isStatic() const { return !is_template && (!parentItem || parentItem->isStatic()); }
	//bool isInstance() const { return parentItem && parentItem->is_template; }
	bool isValid() const;
	bool isReadOnly() const { return read_only; }
	DataType dataType() const { return dtype; }
	State state() const;

	const ConfigItem *buddy(int index) const
	{
		if (index < 0 || index > 2)
			return 0;
		return buddyList[index];
	}
	void setBuddyPath(int index, const QStringList &path)
	{
		if (index < 0 || index > 2)
			return;
		buddyPathList[index] = path;
	}

	const ConfigValidator *validator() const { return val; }
	void setValidator(ConfigValidator *v) { val = v; }

	void setId(int val) { id = val; }
	void setTid(int val) { tid = val; }
	void setTemplateTid(int val) { tplt_tid = val; }
	void setTidDelta(int val) { tplt_tiddelta = val; }
	void setDelta(int val) { tplt_delta = val; }
	void setDataType(DataType val) { dtype = val; }
	void setFWRange(const Range &val) { fw_range = val; }
	void setTemplateRange(const Range &val) { tplt_range = val; }
	void setReadOnly(bool val) { read_only = val; }
	void fillGconfList(struct gconf *&gc, QList<ConfigItem *> &empty, int version, bool data);
	void reset(QList<ConfigItem *> &itemlist);
	bool fillGconf(struct gconf &gc, int version, bool data) const;
	bool readGconf(const struct gconf &gc);
	bool clear();

	ConfigItem *findChild(const QString &key) const;
	//ConfigItem *createChild(ConfigItem *parent, bool node = true, bool templ = false, bool inst = false);

	void debug(const QString &prefix) const;

	void expand();
	bool resolvePaths(ConfigItem *root);
private:
	bool resolvePath(ConfigItem *root, QStringList &Path, ConfigItem *&buddy);
	void cloneChildren(ConfigItem *parent, ConfigItem *tmpl, int index);
	ConfigItem *buddyList[2];
	QStringList buddyPathList[2];

	QList<ConfigItem*> childItems;
	QString key;
	QVariant device_data, user_data;
	ConfigItem *parentItem;

	bool is_node;
	bool is_template;

	DataType dtype;
	int id;
	int tid;
	bool read_only;
	Range fw_range;

	int tplt_delta;
	int tplt_tiddelta;
	int tplt_tid;
	Range tplt_range;
	bool expanded;

	ConfigValidator *val;
	//type
};

class ConfigValidator {
public:
	virtual bool isValid(const ConfigItem &item) const { return true; }
	virtual ~ConfigValidator() {}
};

class CValRegExp: public ConfigValidator {
public:
	CValRegExp(const QRegExp &r): regexp(r) {};
	bool isValid(const ConfigItem &item) const
	{
		const QVariant &value = item.userData();
		return value.type() == QVariant::String && regexp.exactMatch(value.toString());
	}
private:
	const QRegExp regexp;
};

class CValRange: public ConfigValidator {
public:
	CValRange(const Range &r): range(r) {};
	bool isValid(const ConfigItem &item) const
	{
		const QVariant &value = item.userData();
		return value.canConvert(QVariant::Int) && range.contains(value.toInt());
	}
private:
	const Range range;
};

static inline bool check_date(const ConfigItem* day, const ConfigItem* month, const ConfigItem* year)
{
	if (!day || !month || !year)
		return false;
	if (!day->userData().canConvert(QVariant::Int) ||
	    !month->userData().canConvert(QVariant::Int) ||
	    !year->userData().canConvert(QVariant::Int))
		return false;
	if (!QDate(year->userData().toInt(), month->userData().toInt(), day->userData().toInt()).isValid())
		return false;
	return true;
}

static inline bool check_time(const ConfigItem* minute, const ConfigItem* hour)
{
	if (!hour || !minute)
		return false;
	if (!hour->userData().canConvert(QVariant::Int) ||
	    !minute->userData().canConvert(QVariant::Int))
		return false;

	int val;
	val = hour->userData().toInt();
	if (val < 0 || val > 23)
		return false;
	val = minute->userData().toInt();
	if (val < 0 || val > 59)
		return false;
	return true;
}

class CValDay: public ConfigValidator {
public:
	bool isValid(const ConfigItem &item) const
	{
		const ConfigItem* month = item.buddy(0);
		const ConfigItem* year = item.buddy(1);
		return check_date(&item, month, year);
	}
};

class CValMonth: public ConfigValidator {
public:
	bool isValid(const ConfigItem &item) const
	{
		const ConfigItem* day = item.buddy(0);
		const ConfigItem* year = item.buddy(1);
		return check_date(day, &item, year);
	}
};

class CValYear: public ConfigValidator {
public:
	bool isValid(const ConfigItem &item) const
	{
		const ConfigItem* day = item.buddy(0);
		const ConfigItem* month = item.buddy(1);
		return check_date(day, month, &item);
	}
};

class CValMinute: public ConfigValidator {
public:
	bool isValid(const ConfigItem &item) const
	{
		const ConfigItem* hour = item.buddy(0);
		return check_time(&item, hour);
	}
};

class CValHour: public ConfigValidator {
public:
	bool isValid(const ConfigItem &item) const
	{
		const ConfigItem* minute = item.buddy(0);
		return check_time(minute, &item);
	}
};

class ConfigItemModel: public QAbstractItemModel {
	Q_OBJECT
public:
	ConfigItemModel();
	~ConfigItemModel();

	bool readConfigrc(QFile &file);

	int columnCount(const QModelIndex &parent = QModelIndex()) const;
	int rowCount(const QModelIndex &parent = QModelIndex()) const;
	QVariant data(const QModelIndex &index, int role) const;
	Qt::ItemFlags flags(const QModelIndex &index) const;
	QVariant headerData(int section, Qt::Orientation orientation,
	                    int role = Qt::DisplayRole) const;
	QModelIndex index(int row, int column,
	                  const QModelIndex &parent = QModelIndex()) const;
	QModelIndex parent(const QModelIndex &index) const;
	QModelIndex itemindex(ConfigItem* item, int column) const;
	void fillGconfList(struct gconf *&gc, QList<ConfigItem *> &empty, int version, bool data);
	void clear(const QList<ConfigItem *> &itemlist);
	void resetData();
	void readGconfList(struct gconf *&gc);

	bool setData(const QModelIndex &index, const QVariant &value, int role = Qt::EditRole);

private:
	bool setRangeType(const QStringList &args);
	bool setRegExpType(const QStringList &args);
	bool setDateType(ConfigItem::Type type, const QStringList &args);
	bool setTimeType(ConfigItem::Type type, const QStringList &args);
	bool parseType(const QString &str);
	bool readLine(QFile &file, QString &line);
	bool parseKey(QFile &file, QString &line, int start, int &len, unsigned &lno);
	bool parseLine(QFile &file, QString &line, int start, int &len, unsigned &lno);
	bool addKey(const QStringList &keys, bool mult);
	bool addEntry(const QString &lhs, const QString &rhs);
	bool finishEntry();

	bool error;

	//void setupModelData(const QStringList &lines, ConfigItem *parent);
	ConfigItem *rootItem;
	ConfigItem *curItem;
	QList <ConfigValidator*> validators;
	ConfigItemModel(const ConfigItemModel &model); // not implemented
	ConfigItemModel &operator=(const ConfigItemModel &model); // not implemented
};

class Config: public QWidget {
	Q_OBJECT
public:
	Config(GigasetQt &gigaset, QWidget *parent = 0);
public slots:
	void load();
	void store();
	void resetData();
protected:
	void closeEvent(QCloseEvent*);
signals:
	void message(const QString&, int);
private:
	ConfigItemModel m;
	GigasetQt &gs;
	QPushButton *bLoad;
	QPushButton *bStore;
	QPushButton *bReset;
	QTreeView *tv;
};

class ConfigDelegate: public QItemDelegate {
	Q_OBJECT
public:
	ConfigDelegate(QObject *parent = 0);

	QWidget *createEditor(QWidget *parent, const QStyleOptionViewItem &option,
	                      const QModelIndex &index) const;

	void setEditorData(QWidget *editor, const QModelIndex &index) const;
	void setModelData(QWidget *editor, QAbstractItemModel *model,
	                  const QModelIndex &index) const;

	void updateEditorGeometry(QWidget *editor, const QStyleOptionViewItem &option,
	                          const QModelIndex &index) const;
};

#endif
