/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <QtGui>
#include "dial.hh"
//#include "application.hh" //FIXME
#include "gigasetqt.hh"
#include <iostream>
#include <limits>

DialThread::DialThread(GigasetQt &gigaset)
	: abortthread(false), dialing(false), gs(gigaset)
{
}

DialThread::~DialThread()
{
	mutex.lock();
	if (dialing)
		gs.setInterrupt(1);
	abortthread = true;
	condition.wakeOne();
	mutex.unlock();

	wait();
}

int DialThread::dial(unsigned type, const QString to, const QString from)
{
	QMutexLocker locker(&mutex);

	if (dialing)
		return -ERR_BUSY;

	dialing = true;

	numFrom = from;
	numTo = to;
	typeTo = type;

	if (!isRunning()) {
		start();
	} else {
		condition.wakeOne();
	}

	return 0;
}

void DialThread::abort()
{
	mutex.lock();
	if (!dialing) {
		mutex.unlock();
		return;
	}

	gs.setInterrupt(1);
	mutex.unlock();
	// FIXME wait?
}

void DialThread::run()
{
	for (;;) {
		mutex.lock();
		if (abortthread) {
			mutex.unlock();
			return;
		}
		mutex.unlock();

		gs.setInterrupt(0);
		int ret = gs.dial(typeTo, numTo.toLocal8Bit(), numFrom.toLocal8Bit()); //FIXME encoding
		gs.setInterrupt(0);

		mutex.lock();
		dialing = false;
		mutex.unlock();

		emit done(ret);

		mutex.lock();
		condition.wait(&mutex);
		mutex.unlock();
	}
}

DialWidget::DialWidget(GigasetQt &gigaset, QWidget *parent)
	: QWidget(parent), gs(gigaset), dialthread(gigaset)
{
	QGroupBox *groupFrom;
	QGroupBox *groupTo;
	QPushButton *bDial;
	QPushButton *bAbort;
	QFrame * line;

	groupFrom = new QGroupBox(tr("From"), this);
	groupTo = new QGroupBox(tr("To"), this);

	numFrom = new QLineEdit(groupFrom);
	typeTo = new QComboBox(groupTo);
	numTo = new QLineEdit(groupTo);

	typeTo->addItem(tr("internal"));
	typeTo->addItem(tr("external"));

	bDial = new QPushButton(tr("Dial"), this);
	bDial->setDefault(true);
	bDial->setEnabled(gs.isavailable());

	bAbort = new QPushButton(tr("Abort"), this);
	bAbort->setEnabled(false);

	line = new QFrame(this);
	line->setFrameShape(QFrame::HLine);

	QHBoxLayout *toLayout = new QHBoxLayout(groupTo);
	toLayout->addWidget(typeTo);
	toLayout->addWidget(numTo);

	QHBoxLayout *fromLayout = new QHBoxLayout(groupFrom);
	fromLayout->addWidget(numFrom);

	QHBoxLayout *buttonLayout = new QHBoxLayout;
	buttonLayout->addWidget(bDial);
	buttonLayout->addWidget(bAbort);
	buttonLayout->addStretch();

	QVBoxLayout *layout = new QVBoxLayout(this);
	layout->addWidget(groupFrom);
	layout->addWidget(groupTo);
	layout->addWidget(line);
	layout->addLayout(buttonLayout);

	setWindowTitle(tr("Dial"));
	setLayout(layout);

	connect(bDial, SIGNAL(clicked()), SLOT(dial()));
	connect(&gs, SIGNAL(deviceAvailable(bool)), bDial, SLOT(setEnabled(bool))); //FIXME use action?

	connect(bAbort, SIGNAL(clicked()), SLOT(abort()));
	connect(this, SIGNAL(dialing(bool)), bAbort, SLOT(setEnabled(bool))); //FIXME use action?

	connect(&dialthread, SIGNAL(done(int)), this, SLOT(done(int)));
}

void DialWidget::dial()
{
	unsigned type;
	int ret;

	switch (typeTo->currentIndex()) {
	case 0:
		type = GIG_TYPE_INTERNAL;
		break;
	case 1:
		type = GIG_TYPE_EXTERNAL;
		break;
	default:
		return; //FIXME msg
	}

	QString to = numTo->text();
	QString from = numFrom->text();

	if (!gs.claim())
		return; // -ERR_BUSY; //FIXME msg

	ret = gs.cidmode();

	if (ret < 0) {
		gs.release();
		return; //FIXME msg
	}

	emit dialing(true);

	ret = dialthread.dial(type, to, from);

	if (ret < 0)
		done(ret);
}

void DialWidget::done(int error)
{
	//FIXME error code
	gs.release();
	emit dialing(false);
}

void DialWidget::abort()
{
	dialthread.abort();
}

DialWidget::~DialWidget()
{
}

void DialWidget::closeEvent(QCloseEvent *e)
{
	//TODO
	e->accept(); // e->ignore();
}
