/*
 * Copyright (c) 2006 by Hansjoerg Lipp <hjlipp@web.de>
 *
 * =====================================================================
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * =====================================================================
 */

#include <QtGui>
#include "logger.hh"
#include "gigasetqt.hh"
#include <iostream>
//#include <cstdlib>
//#include <limits>

void Logger::callback(unsigned level, const char *str)
{
	//emit newMessage(level, QByteArray(str));
	QByteArray b(str);
	emit newMessage(level, b);
}

extern "C" void loggercallback(void *data, unsigned level, const char *str)
{
	Logger *l = static_cast<Logger *>(data);
	try {
		l->callback(level, str);
	}
	catch (...) {
		std::terminate();
	}
}

Logger::Logger(bool qtHandler)
{
	glog.log = stderr;
	glog.verb = GLOG_ERR|GLOG_WARN|GLOG_INFO|GLOG_STATUS|GLOG_IO|GLOG_DEBUG|GLOG_DUMP|
	            GLOG_QTDBG|GLOG_QTWARN|GLOG_QTCRIT|GLOG_QTFATAL|GLOG_QTUNKNOWN;
	glog.data = this;
	glog.logfunc = loggercallback;
	glog.errorstr = 0;

	if (qtHandler)
		installQtHandler();
}

Logger *Logger::qtData = 0;

Logger::~Logger() {
	if (qtData == this)
		qtData = 0;
}

static void qtCallback(QtMsgType type, const char *msg)
{
	unsigned level;
	bool term = false;

	switch (type) {
	case QtDebugMsg:
		std::cerr << "debug: ";
		level = GLOG_QTDBG;
		break;
	case QtWarningMsg:
		std::cerr << "warning: ";
		level = GLOG_QTWARN;
		break;
	case QtCriticalMsg:
		std::cerr << "critical: ";
		level = GLOG_QTCRIT;
		break;
	case QtFatalMsg:
		std::cerr << "fatal: ";
		level = GLOG_QTFATAL;
		term = true;
		break;
	default:
		std::cerr << "unknown Qt message: ";
		level = GLOG_QTUNKNOWN;
		break;
	}
	std::cerr << msg << std::endl;
	if (term)
		std::terminate();

	Logger *l = Logger::getQtData();
	if (l)
		l->callback(level, msg);
}

void Logger::installQtHandler()
{
	if (qtData) {
		//msg?
		return;
	}
	qtData = this;
	qInstallMsgHandler(qtCallback);
}

QString Logger::levelToString(unsigned level)
{
	if (level < 1U << GLOG_LIBBITS) {
		if (level & GLOG_ERR)
			return tr("error");
		if (level & GLOG_WARN)
			return tr("warning");
		if (level & GLOG_INFO)
			return tr("info");
		if (level & GLOG_STATUS)
			return tr("status");
		if (level & GLOG_IO)
			return tr("i/o");
		if (level & GLOG_DEBUG)
			return tr("debug");
		if (level & GLOG_DUMP)
			return tr("dump");
		return tr("unknown");
	}

	if (level & GLOG_QTFATAL)
		return tr("fatal (qt)");
	if (level & GLOG_QTCRIT)
		return tr("critical (qt)");
	if (level & GLOG_QTWARN)
		return tr("warning (qt)");
	if (level & GLOG_QTDBG)
		return tr("debug (qt)");
	if (level & GLOG_QTUNKNOWN)
		return tr("unknown (qt)");

	return tr("unknown (app)");
}

LogListViewItem::LogListViewItem(QTreeWidget *parent, const QByteArray &msg, unsigned lev)
	: QTreeWidgetItem(parent), message(msg), level(lev)
{
	setText(0, Logger::levelToString(lev));
	setText(1, msg);
}

LogList::LogList(QWidget *parent)
    : QWidget(parent)
{
	tw = new QTreeWidget(this);
	tw->setColumnCount(2);
	QStringList labels;
	labels << tr("level") << tr("message");
	tw->setHeaderLabels(labels);

	tw->setSelectionMode(QAbstractItemView::ExtendedSelection);
	tw->setSelectionBehavior(QAbstractItemView::SelectRows);

	QPushButton *bClear = new QPushButton(tr("&Clear"), this);
	connect(bClear, SIGNAL(clicked()), tw, SLOT(clear()));
	QPushButton *bSave = new QPushButton(tr("&Save"), this);
	connect(bSave, SIGNAL(clicked()), SLOT(save()));

	//FIXME this could be nicer:
	QHBoxLayout *buttonlayout = new QHBoxLayout;
	buttonlayout->addWidget(bSave);
	buttonlayout->addWidget(bClear);
	//buttonlayout->addItem(spc);
	buttonlayout->addStretch();
	QVBoxLayout *layout = new QVBoxLayout;
	layout->addWidget(tw);
	layout->addLayout(buttonlayout);

	setWindowTitle(tr("Log"));

	setLayout(layout);
}

void LogList::newMessage(unsigned int level, QByteArray msg)
{
	while (msg.endsWith('\n'))
		msg.chop(1);
	new LogListViewItem(tw, msg, level);
}

void LogList::save()
{
	QString name = QFileDialog::getSaveFileName(this, tr("Save log")/*, QString(), QString(), 0, 0*/);
	if (name.isNull())
		return;
	QFile file(name);
	if (!file.open(QIODevice::Append|QIODevice::Text|QIODevice::Truncate)) { //FIXME is there something like O_CREAT|O_EXCL?
		qWarning() << "could not open '" << name << "'"; //FIXME msgbox + translation
		return;
	}
	int count = tw->topLevelItemCount();
	for (int i = 0; i < count; ++i) {
		QTreeWidgetItem *item = tw->topLevelItem(i);
		if (!item) {
			//qWarning("log message %d not found", i);
			qWarning() << "log message " << i << " not found";
			continue;
		}
		int len;
		const char *pos;
		const QByteArray levelstr = item->text(0).toUtf8();
		len = levelstr.size();
		pos = levelstr.constData();
		while (len) {
			qint64 num = file.write(pos, len);
			if (num == -1 || num == 0) {
				qWarning() << "error writing to '" << name << "'"; //FIXME msgbox + translation
				return;
			}
			len -= num;
			pos += num;
		}
		if (!file.putChar('\t')) {
			qWarning() << "error writing to '" << name << "'"; //FIXME msgbox + translation
			return;
		}
		const QByteArray msgstr = item->text(1).toUtf8();
		len = msgstr.size();
		pos = msgstr.constData();
		while (len) {
			qint64 num = file.write(pos, len);
			if (num == -1 || num == 0) {
				qWarning() << "error writing to '" << name << "'"; //FIXME msgbox + translation
				return;
			}
			len -= num;
			pos += num;
		}
		if (!file.putChar('\n')) {
			qWarning() << "error writing to '" << name << "'"; //FIXME msgbox + translation
			return;
		}
	}
}

//void LogList::closeEvent(QCloseEvent *e)
//{
//	e->ignore(); //FIXME hide?
//}
